/*

  Copyright (C) 2000, 2001 Silicon Graphics, Inc.  All Rights Reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement 
  or the like.  Any license provided herein, whether implied or 
  otherwise, applies only to this software file.  Patent licenses, if 
  any, provided herein do not apply to combinations of this program with 
  other software, or any other product whatsoever.  

  You should have received a copy of the GNU General Public License along
  with this program; if not, write the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston MA 02111-1307, USA.

  Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pky,
  Mountain View, CA 94043, or:

  http://www.sgi.com

  For further information regarding this notice, see:

  http://oss.sgi.com/projects/GenInfo/NoticeExplan

*/

#if (defined(__MACH__) && defined(__APPLE__)) || defined(__CYGWIN__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
#else
#include <values.h>
#endif
#include "defs.h"
#include "glob.h"
#include "config.h"
#include "wn.h"
#include "wn_util.h"
#include "const.h"

// #include "gnu/MIPS/config.h"
#include "gnu_config.h"
#include "gnu/system.h"
#include "gnu/machmode.h"
#include <setjmp.h>
extern "C" {
#include "gnu/toplev.h"
#include "gnu/flags.h"
#include "gnu/tree.h"
#include "gnu/c-tree.h"
  //extern void warning (char*,...);	// from gnu
extern tree c_strlen_exported (tree);
#include "gnu/function.h"
}

#ifdef GPLUSPLUS_FE
#include "gnu/cp/cp-tree.h"
#endif /* GPLUSPLUS_FE */

#include "ir_reader.h"
#include "tree_symtab.h"
#include "wfe_misc.h"
#include "wfe_decl.h"
#include "wfe_expr.h"
#include "wfe_stmt.h"
#include <cmplrs/rcodes.h>

extern int is_ia32; /* from toplev.h */

extern void dump_ty_idx (TY_IDX);

extern "C" int get_expr_stmts_for_value (void);

struct operator_from_tree_t {
  int      tree_code;
  char*    name;
  char     code;
  int      nkids;
  OPERATOR opr;
} Operator_From_Tree [] = {
  ERROR_MARK,              "error_mark",              'x', 0,  OPERATOR_UNKNOWN,
  IDENTIFIER_NODE,         "identifier_node",         'x', -1, OPERATOR_UNKNOWN,
  OP_IDENTIFIER,           "op_identifier",           'x', 2,  OPERATOR_UNKNOWN,
  TREE_LIST,               "tree_list",               'x', 2,  OPERATOR_UNKNOWN,
  TREE_VEC,                "tree_vec",                'x', 2,  OPERATOR_UNKNOWN,
  BLOCK,                   "block",                   'b', 0,  OPERATOR_UNKNOWN,
  VOID_TYPE,               "void_type",               't', 0,  OPERATOR_UNKNOWN,
  INTEGER_TYPE,            "integer_type",            't', 0,  OPERATOR_UNKNOWN,
  REAL_TYPE,               "real_type",               't', 0,  OPERATOR_UNKNOWN,
  COMPLEX_TYPE,            "complex_type",            't', 0,  OPERATOR_UNKNOWN,
  ENUMERAL_TYPE,           "enumeral_type",           't', 0,  OPERATOR_UNKNOWN,
  BOOLEAN_TYPE,            "boolean_type",            't', 0,  OPERATOR_UNKNOWN,
  CHAR_TYPE,               "char_type",               't', 0,  OPERATOR_UNKNOWN,
  POINTER_TYPE,            "pointer_type",            't', 0,  OPERATOR_UNKNOWN,
  OFFSET_TYPE,             "offset_type",             't', 0,  OPERATOR_UNKNOWN,
  REFERENCE_TYPE,          "reference_type",          't', 0,  OPERATOR_UNKNOWN,
  METHOD_TYPE,             "method_type",             't', 0,  OPERATOR_UNKNOWN,
  FILE_TYPE,               "file_type",               't', 0,  OPERATOR_UNKNOWN,
  ARRAY_TYPE,              "array_type",              't', 0,  OPERATOR_UNKNOWN,
  SET_TYPE,                "set_type",                't', 0,  OPERATOR_UNKNOWN,
  RECORD_TYPE,             "record_type",             't', 0,  OPERATOR_UNKNOWN,
  UNION_TYPE,              "union_type",              't', 0,  OPERATOR_UNKNOWN,
  QUAL_UNION_TYPE,         "qual_union_type",         't', 0,  OPERATOR_UNKNOWN,
  FUNCTION_TYPE,           "function_type",           't', 0,  OPERATOR_UNKNOWN,
  LANG_TYPE,               "lang_type",               't', 0,  OPERATOR_UNKNOWN,
  INTEGER_CST,             "integer_cst",             'c', 2,  OPERATOR_UNKNOWN,
  REAL_CST,                "real_cst",                'c', 3,  OPERATOR_UNKNOWN,
  COMPLEX_CST,             "complex_cst",             'c', 3,  OPERATOR_UNKNOWN,
  STRING_CST,              "string_cst",              'c', 3,  OPERATOR_UNKNOWN,
  FUNCTION_DECL,           "function_decl",           'd', 0,  OPERATOR_UNKNOWN,
  LABEL_DECL,              "label_decl",              'd', 0,  OPERATOR_UNKNOWN,
  CONST_DECL,              "const_decl",              'd', 0,  OPERATOR_UNKNOWN,
  TYPE_DECL,               "type_decl",               'd', 0,  OPERATOR_UNKNOWN,
  VAR_DECL,                "var_decl",                'd', 0,  OPERATOR_UNKNOWN,
  PARM_DECL,               "parm_decl",               'd', 0,  OPERATOR_UNKNOWN,
  RESULT_DECL,             "result_decl",             'd', 0,  OPERATOR_UNKNOWN,
  FIELD_DECL,              "field_decl",              'd', 0,  OPERATOR_UNKNOWN,
  NAMESPACE_DECL,          "namespace_decl",          'd', 0,  OPERATOR_UNKNOWN,
  COMPONENT_REF,           "component_ref",           'r', 2,  OPERATOR_UNKNOWN,
  BIT_FIELD_REF,           "bit_field_ref",           'r', 3,  OPERATOR_UNKNOWN,
  INDIRECT_REF,            "indirect_ref",            'r', 1,  OPERATOR_UNKNOWN,
  BUFFER_REF,              "buffer_ref",              'r', 1,  OPERATOR_UNKNOWN,
  ARRAY_REF,               "array_ref",               'r', 2,  OPERATOR_UNKNOWN,
  CONSTRUCTOR,             "constructor",             'e', 2,  OPERATOR_UNKNOWN,
  COMPOUND_EXPR,           "compound_expr",           'e', 2,  OPERATOR_UNKNOWN,
  MODIFY_EXPR,             "modify_expr",             'e', 2,  OPERATOR_UNKNOWN,
  INIT_EXPR,               "init_expr",               'e', 2,  OPERATOR_UNKNOWN,
  TARGET_EXPR,             "target_expr",             'e', 4,  OPERATOR_UNKNOWN,
  COND_EXPR,               "cond_expr",               'e', 3,  OPERATOR_UNKNOWN,
  BIND_EXPR,               "bind_expr",               'e', 3,  OPERATOR_UNKNOWN,
  CALL_EXPR,               "call_expr",               'e', 3,  OPERATOR_UNKNOWN,
  METHOD_CALL_EXPR,        "method_call_expr",        'e', 4,  OPERATOR_UNKNOWN,
  WITH_CLEANUP_EXPR,       "with_cleanup_expr",       'e', 3,  OPERATOR_UNKNOWN,
  CLEANUP_POINT_EXPR,      "cleanup_point_expr",      'e', 1,  OPERATOR_UNKNOWN,
  PLACEHOLDER_EXPR,        "placeholder_expr",        'x', 0,  OPERATOR_UNKNOWN,
  WITH_RECORD_EXPR,        "with_record_expr",        'e', 2,  OPERATOR_UNKNOWN,
  PLUS_EXPR,               "plus_expr",               '2', 2,  OPR_ADD,
  MINUS_EXPR,              "minus_expr",              '2', 2,  OPR_SUB,
  MULT_EXPR,               "mult_expr",               '2', 2,  OPR_MPY,
  TRUNC_DIV_EXPR,          "trunc_div_expr",          '2', 2,  OPR_DIV,
  CEIL_DIV_EXPR,           "ceil_div_expr",           '2', 2,  OPR_DIV,
  FLOOR_DIV_EXPR,          "floor_div_expr",          '2', 2,  OPERATOR_UNKNOWN,
  ROUND_DIV_EXPR,          "round_div_expr",          '2', 2,  OPERATOR_UNKNOWN,
  TRUNC_MOD_EXPR,          "trunc_mod_expr",          '2', 2,  OPR_REM,
  CEIL_MOD_EXPR,           "ceil_mod_expr",           '2', 2,  OPERATOR_UNKNOWN,
  FLOOR_MOD_EXPR,          "floor_mod_expr",          '2', 2,  OPERATOR_UNKNOWN,
  ROUND_MOD_EXPR,          "round_mod_expr",          '2', 2,  OPERATOR_UNKNOWN,
  RDIV_EXPR,               "rdiv_expr",               '2', 2,  OPR_DIV,
  EXACT_DIV_EXPR,          "exact_div_expr",          '2', 2,  OPR_DIV,
  FIX_TRUNC_EXPR,          "fix_trunc_expr",          '1', 1,  OPERATOR_UNKNOWN,
  FIX_CEIL_EXPR,           "fix_ceil_expr",           '1', 1,  OPERATOR_UNKNOWN,
  FIX_FLOOR_EXPR,          "fix_floor_expr",          '1', 1,  OPERATOR_UNKNOWN,
  FIX_ROUND_EXPR,          "fix_round_expr",          '1', 1,  OPERATOR_UNKNOWN,
  FLOAT_EXPR,              "float_expr",              '1', 1,  OPERATOR_UNKNOWN,
  EXPON_EXPR,              "expon_expr",              '2', 2,  OPERATOR_UNKNOWN,
  NEGATE_EXPR,             "negate_expr",             '1', 1,  OPR_NEG,
  MIN_EXPR,                "min_expr",                '2', 2,  OPR_MIN,
  MAX_EXPR,                "max_expr",                '2', 2,  OPR_MAX,
  ABS_EXPR,                "abs_expr",                '1', 1,  OPR_ABS,
  FFS_EXPR,                "ffs_expr",                '1', 1,  OPERATOR_UNKNOWN,
  LSHIFT_EXPR,             "lshift_expr",             '2', 2,  OPR_SHL,
  RSHIFT_EXPR,             "rshift_expr",             '2', 2,  OPERATOR_UNKNOWN,
  LROTATE_EXPR,            "lrotate_expr",            '2', 2,  OPR_RROTATE,
  RROTATE_EXPR,            "rrotate_expr",            '2', 2,  OPR_RROTATE,
  BIT_IOR_EXPR,            "bit_ior_expr",            '2', 2,  OPR_BIOR,
  BIT_XOR_EXPR,            "bit_xor_expr",            '2', 2,  OPR_BXOR,
  BIT_AND_EXPR,            "bit_and_expr",            '2', 2,  OPR_BAND,
  BIT_ANDTC_EXPR,          "bit_andtc_expr",          '2', 2,  OPERATOR_UNKNOWN,
  BIT_NOT_EXPR,            "bit_not_expr",            '1', 1,  OPR_BNOT,
  TRUTH_ANDIF_EXPR,        "truth_andif_expr",        'e', 2,  OPR_CAND,
  TRUTH_ORIF_EXPR,         "truth_orif_expr",         'e', 2,  OPR_CIOR,
  TRUTH_AND_EXPR,          "truth_and_expr",          'e', 2,  OPR_BAND,
  TRUTH_OR_EXPR,           "truth_or_expr",           'e', 2,  OPR_BIOR,
  TRUTH_XOR_EXPR,          "truth_xor_expr",          'e', 2,  OPR_BXOR,
  TRUTH_NOT_EXPR,          "truth_not_expr",          'e', 1,  OPR_LNOT,
  LT_EXPR,                 "lt_expr",                 '<', 2,  OPR_LT,
  LE_EXPR,                 "le_expr",                 '<', 2,  OPR_LE,
  GT_EXPR,                 "gt_expr",                 '<', 2,  OPR_GT,
  GE_EXPR,                 "ge_expr",                 '<', 2,  OPR_GE,
  EQ_EXPR,                 "eq_expr",                 '<', 2,  OPR_EQ,
  NE_EXPR,                 "ne_expr",                 '<', 2,  OPR_NE,
  UNORDERED_EXPR,          "unordered_expr",          '<', 2,  OPERATOR_UNKNOWN,
  ORDERED_EXPR,            "ordered_expr",            '<', 2,  OPERATOR_UNKNOWN,
  UNLT_EXPR,               "unlt_expr",               '<', 2,  OPERATOR_UNKNOWN,
  UNLE_EXPR,               "unle_expr",               '<', 2,  OPERATOR_UNKNOWN,
  UNGT_EXPR,               "ungt_expr",               '<', 2,  OPERATOR_UNKNOWN,
  UNGE_EXPR,               "unge_expr",               '<', 2,  OPERATOR_UNKNOWN,
  UNEQ_EXPR,               "uneq_expr",               '<', 2,  OPERATOR_UNKNOWN,
  IN_EXPR,                 "in_expr",                 '2', 2,  OPERATOR_UNKNOWN,
  SET_LE_EXPR,             "set_le_expr",             '<', 2,  OPERATOR_UNKNOWN,
  CARD_EXPR,               "card_expr",               '1', 1,  OPERATOR_UNKNOWN,
  RANGE_EXPR,              "range_expr",              '2', 2,  OPERATOR_UNKNOWN,
  CONVERT_EXPR,            "convert_expr",            '1', 1,  OPERATOR_UNKNOWN,
  NOP_EXPR,                "nop_expr",                '1', 1,  OPERATOR_UNKNOWN,
  NON_LVALUE_EXPR,         "non_lvalue_expr",         '1', 1,  OPERATOR_UNKNOWN,
  SAVE_EXPR,               "save_expr",               'e', 3,  OPERATOR_UNKNOWN,
  UNSAVE_EXPR,             "unsave_expr",             'e', 1,  OPERATOR_UNKNOWN,
  RTL_EXPR,                "rtl_expr",                'e', 2,  OPERATOR_UNKNOWN,
  ADDR_EXPR,               "addr_expr",               'e', 1,  OPERATOR_UNKNOWN,
  REFERENCE_EXPR,          "reference_expr",          'e', 1,  OPERATOR_UNKNOWN,
  ENTRY_VALUE_EXPR,        "entry_value_expr",        'e', 1,  OPERATOR_UNKNOWN,
  COMPLEX_EXPR,            "complex_expr",            '2', 2,  OPR_PAIR,
  CONJ_EXPR,               "conj_expr",               '1', 1,  OPERATOR_UNKNOWN,
  REALPART_EXPR,           "realpart_expr",           '1', 1,  OPR_FIRSTPART,
  IMAGPART_EXPR,           "imagpart_expr",           '1', 1,  OPR_SECONDPART,
  PREDECREMENT_EXPR,       "predecrement_expr",       'e', 2,  OPR_SUB,
  PREINCREMENT_EXPR,       "preincrement_expr",       'e', 2,  OPR_ADD,
  POSTDECREMENT_EXPR,      "postdecrement_expr",      'e', 2,  OPR_SUB,
  POSTINCREMENT_EXPR,      "postincrement_expr",      'e', 2,  OPR_ADD,
  VA_ARG_EXPR,             "va_arg_expr",             'e', 1,  OPERATOR_UNKNOWN,
  TRY_CATCH_EXPR,          "try_catch_expr",          'e', 2,  OPERATOR_UNKNOWN,
  TRY_FINALLY_EXPR,        "try_finally_expr",        'e', 2,  OPERATOR_UNKNOWN,
  GOTO_SUBROUTINE_EXPR,    "goto_subroutine",         'e', 2,  OPERATOR_UNKNOWN,
  POPDHC_EXPR,             "popdhc_expr",             's', 0,  OPERATOR_UNKNOWN,
  POPDCC_EXPR,             "popdcc_expr",             's', 0,  OPERATOR_UNKNOWN,
  LABEL_EXPR,              "label_expr",              's', 1,  OPERATOR_UNKNOWN,
  GOTO_EXPR,               "goto_expr",               's', 1,  OPERATOR_UNKNOWN,
  RETURN_EXPR,             "return_expr",             's', 1,  OPERATOR_UNKNOWN,
  EXIT_EXPR,               "exit_expr",               's', 1,  OPERATOR_UNKNOWN,
  LOOP_EXPR,               "loop_expr",               's', 1,  OPERATOR_UNKNOWN,
  LABELED_BLOCK_EXPR,      "labeled_block_expr",      'e', 2,  OPERATOR_UNKNOWN,
  EXIT_BLOCK_EXPR,         "exit_block_expr",         'e', 2,  OPERATOR_UNKNOWN,
  EXPR_WITH_FILE_LOCATION, "expr_with_file_location", 'e', 3,  OPERATOR_UNKNOWN,
  SWITCH_EXPR,             "switch_expr",             'e', 2,  OPERATOR_UNKNOWN,
  LAST_AND_UNUSED_TREE_CODE,"last_and_unused_tree_code",0, 0,  OPERATOR_UNKNOWN,
#ifdef GPLUSPLUS_FE
  OFFSET_REF,              "offset_ref",              'r', 2,  OPERATOR_UNKNOWN,
  NEW_EXPR,                "nw_expr",                 'e', 3,  OPERATOR_UNKNOWN,
  VEC_NEW_EXPR,            "vec_nw_expr",             'e', 3,  OPERATOR_UNKNOWN,
  DELETE_EXPR,             "dl_expr",                 'e', 2,  OPERATOR_UNKNOWN,
  VEC_DELETE_EXPR,         "vec_dl_expr",             'e', 2,  OPERATOR_UNKNOWN,
  SCOPE_REF,               "scope_ref",               'r', 2,  OPERATOR_UNKNOWN,
  MEMBER_REF,              "member_ref",              'r', 2,  OPERATOR_UNKNOWN,
  TYPE_EXPR,               "type_expr",               'e', 1,  OPERATOR_UNKNOWN,
  AGGR_INIT_EXPR,          "aggr_init_expr",          'e', 3,  OPERATOR_UNKNOWN,
  THROW_EXPR,              "throw_expr",              'e', 1,  OPERATOR_UNKNOWN,
  VEC_INIT_EXPR,           "vec_init_expr",           'e', 3,  OPERATOR_UNKNOWN,
  TEMPLATE_DECL,           "template_decl",           'd', 0,  OPERATOR_UNKNOWN,
  TEMPLATE_PARM_INDEX,     "template_parm_index",     'x', 0,  OPERATOR_UNKNOWN,
  TEMPLATE_TYPE_PARM,      "template_type_parm",      't', 0,  OPERATOR_UNKNOWN,
  TEMPLATE_TEMPLATE_PARM,  "template_template_parm",  't', 0,  OPERATOR_UNKNOWN,
  TYPENAME_TYPE,           "typename_type",           't', 0,  OPERATOR_UNKNOWN,
  THUNK_DECL,              "thunk_decl",              'd', 0,  OPERATOR_UNKNOWN,
  USING_DECL,              "using_decl",              'd', 0,  OPERATOR_UNKNOWN,
  DEFAULT_ARG,             "default_arg",             'c', 2,  OPERATOR_UNKNOWN,
  TEMPLATE_ID_EXPR,        "template_id_expr",        'e', 2,  OPERATOR_UNKNOWN,
  CPLUS_BINDING,           "binding",                 'x', 2,  OPERATOR_UNKNOWN,
  OVERLOAD,                "overload",                'x', 1,  OPERATOR_UNKNOWN,
  WRAPPER,                 "wrapper",                 'x', 1,  OPERATOR_UNKNOWN,
  SRCLOC,                  "srcloc",                  'x', 2,  OPERATOR_UNKNOWN,
  LOOKUP_EXPR,             "lookup_expr",             'e', 2,  OPERATOR_UNKNOWN,
  MODOP_EXPR,              "modop_expr",              'e', 3,  OPERATOR_UNKNOWN,
  CAST_EXPR,               "cast_expr",               '1', 1,  OPERATOR_UNKNOWN,
  REINTERPRET_CAST_EXPR,   "reinterpret_cast_expr",   '1', 1,  OPERATOR_UNKNOWN,
  CONST_CAST_EXPR,         "const_cast_expr",         '1', 1,  OPERATOR_UNKNOWN,
  STATIC_CAST_EXPR,        "static_cast_expr",        '1', 1,  OPERATOR_UNKNOWN,
  DYNAMIC_CAST_EXPR,       "dynamic_cast_expr",       '1', 1,  OPERATOR_UNKNOWN,
  SIZEOF_EXPR,             "sizeof_expr",             '1', 1,  OPERATOR_UNKNOWN,
  ALIGNOF_EXPR,            "alignof_expr",            '1', 1,  OPERATOR_UNKNOWN,
  ARROW_EXPR,              "arrow_expr",              'e', 1,  OPERATOR_UNKNOWN,
  DOTSTAR_EXPR,            "dotstar_expr",            'e', 2,  OPERATOR_UNKNOWN,
  TYPEID_EXPR,             "typeid_expr",             'e', 1,  OPERATOR_UNKNOWN,
  EXPR_STMT,               "expr_stmt",               'e', 1,  OPERATOR_UNKNOWN,
  COMPOUND_STMT,           "compound_stmt",           'e', 1,  OPERATOR_UNKNOWN,
  DECL_STMT,               "decl_stmt",               'e', 3,  OPERATOR_UNKNOWN,
  IF_STMT,                 "if_stmt",                 'e', 3,  OPERATOR_UNKNOWN,
  FOR_STMT,                "for_stmt",                'e', 4,  OPERATOR_UNKNOWN,
  WHILE_STMT,              "while_stmt",              'e', 2,  OPERATOR_UNKNOWN,
  DO_STMT,                 "do_stmt",                 'e', 2,  OPERATOR_UNKNOWN,
  RETURN_STMT,             "return_stmt",             'e', 1,  OPERATOR_UNKNOWN,
  BREAK_STMT,              "break_stmt",              'e', 0,  OPERATOR_UNKNOWN,
  CONTINUE_STMT,           "continue_stmt",           'e', 0,  OPERATOR_UNKNOWN,
  SWITCH_STMT,             "switch_stmt",             'e', 2,  OPERATOR_UNKNOWN,
  GOTO_STMT,               "goto_stmt",               'e', 1,  OPERATOR_UNKNOWN,
  ASM_STMT,                "asm_stmt",                'e', 5,  OPERATOR_UNKNOWN,
  CTOR_INITIALIZER,        "ctor_initializer",        'e', 2,  OPERATOR_UNKNOWN,
  CASE_LABEL,              "case_label",              'e', 2,  OPERATOR_UNKNOWN,
  RETURN_INIT,             "return_init",             'e', 2,  OPERATOR_UNKNOWN,
  TRY_BLOCK,               "try_stmt",                'e', 2,  OPERATOR_UNKNOWN,
  HANDLER,                 "catch_stmt",              'e', 2,  OPERATOR_UNKNOWN,
  TAG_DEFN,                "tag_defn",                'e', 0,  OPERATOR_UNKNOWN,
  IDENTITY_CONV,           "identity_conv",           'e', 1,  OPERATOR_UNKNOWN,
  LVALUE_CONV,             "lvalue_conv",             'e', 1,  OPERATOR_UNKNOWN,
  QUAL_CONV,               "qual_conv",               'e', 1,  OPERATOR_UNKNOWN,
  STD_CONV,                "std_conv",                'e', 1,  OPERATOR_UNKNOWN,
  PTR_CONV,                "ptr_conv",                'e', 1,  OPERATOR_UNKNOWN,
  PMEM_CONV,               "pmem_conv",               'e', 1,  OPERATOR_UNKNOWN,
  BASE_CONV,               "base_conv",               'e', 1,  OPERATOR_UNKNOWN,
  REF_BIND,                "ref_bind",                'e', 1,  OPERATOR_UNKNOWN,
  USER_CONV,               "user_conv",               'e', 2,  OPERATOR_UNKNOWN,
  AMBIG_CONV,              "ambig_conv",              'e', 1,  OPERATOR_UNKNOWN,
  RVALUE_CONV,             "rvalue_conv",             'e', 1,  OPERATOR_UNKNOWN,
  LAST_CPLUS_TREE_CODE,    "last_cplus_tree_code",     0,  0,  OPERATOR_UNKNOWN
#endif /* GPLUSPLUSFE */
};


WN *last_ptr_arith_wn = 0;
WN *last_ptr_disp_wn = 0; 
extern BOOL WN_Simp_Fold_LDA;
WN_MAP _upc_struct_map;

// check whether the WHIRL operator has subsumed cvtl in its semantics
// (intended only for integer operations)
bool
Has_Subsumed_Cvtl(OPERATOR opr)
{
  if (OPERATOR_is_load(opr) || OPERATOR_is_leaf(opr))
    return TRUE;
  if (opr == OPR_CVT || opr == OPR_CVTL || opr == OPR_TAS)
    return TRUE;
  if (opr == OPR_EQ || opr == OPR_NE ||
      opr == OPR_GE || opr == OPR_GT ||
      opr == OPR_LE || opr == OPR_LT ||
      opr == OPR_LNOT || opr == OPR_LAND || opr == OPR_LIOR ||
      opr == OPR_CAND || opr == OPR_CIOR)
    return TRUE;
  return FALSE;
}

// Round up an object size to the size it would require in the parameter
// area on the stack.  This is defined to be the difference between its
// start address and the lowest possible starting address of the next parameter.
inline UINT64 Parameter_Size(UINT64 sz)
{
#   if WORDS_BIG_ENDIAN
	return sz;
#   else
	return (sz + UNITS_PER_WORD - 1) & ~(UNITS_PER_WORD - 1);
#   endif
}

inline TYPE_ID
Widen_Mtype (TYPE_ID t)
{
  if (MTYPE_is_m(t))
    return t;
  if (MTYPE_is_void(t) || t == MTYPE_BS) {
    Fail_FmtAssertion ("Widen_Mtype: for MTYPE_V or MTYPE_BS");
    return t;
  }
  if (MTYPE_byte_size(t) >= 4)
    return t;
  return Mtype_TransferSize(MTYPE_I4, t);
}



WN*
WFE_Patch_Field_Addr_Exp(tree exp, WN *wn) {
 

  return wn;
  tree op = exp;
  TY_IDX tmp;
  WN *wn0;
  STRIP_NOPS(op);
  if (TREE_CODE(op) == ADDR_EXPR && TREE_CODE(TREE_OPERAND(op,0)) == COMPONENT_REF 
      && WN_operator(wn) != OPR_ILOAD) {
    tmp = Copy_TY(WN_ty(wn));
    Set_TY_is_shared(tmp);
    
   //  Clear_TY_is_shared(tmp);
    wn = WN_CreateIload(OPR_ILOAD, WN_rtype(wn),
			WN_rtype(wn), 
			0, 
			tmp,
			Make_Pointer_Type(tmp),
			wn, 0);
    
  }
  return wn;
}

// Traverse the tree to see if the address of a variable is being taken

void
WFE_Set_ST_Addr_Saved (WN *wn)
{
  OPERATOR  Operator;
  ST       *st;

  Operator = WN_operator (wn);

  switch ( Operator ) {

    case OPR_LDA:
    case OPR_LDMA:

      st = WN_st (wn);

      if (ST_class(st) == CLASS_VAR || ST_class(st) == CLASS_FUNC)
        Set_ST_addr_saved (st);
      break;

    case OPR_ARRAY:

      WFE_Set_ST_Addr_Saved (WN_kid0 (wn));
      break;

    case OPR_LDID:

      st = WN_st (wn);
      if (ST_pt_to_unique_mem (st))
        Clear_ST_pt_to_unique_mem (st);
      break;

    case OPR_CONST:
    case OPR_ILOAD:
    case OPR_MLOAD:
    case OPR_INTCONST:
    case OPR_INTRINSIC_OP:
    case OPR_CALL:
    case OPR_EQ:
    case OPR_NE:
    case OPR_GT:
    case OPR_GE:
    case OPR_LT:
    case OPR_LE:
    case OPR_ALLOCA:
      break;

    case OPR_EVAL:
    case OPR_TAS:
    case OPR_CVT:
    case OPR_CVTL:
    case OPR_NEG:
    case OPR_ABS:
    case OPR_SQRT:
    case OPR_REALPART:
    case OPR_IMAGPART:
    case OPR_PAREN:
    case OPR_RND:
    case OPR_TRUNC:
    case OPR_CEIL:
    case OPR_FLOOR:
    case OPR_BNOT:
    case OPR_LNOT:
    case OPR_LOWPART:
    case OPR_HIGHPART:
    case OPR_MINPART:
    case OPR_MAXPART:
    case OPR_RECIP:
    case OPR_RSQRT:
    case OPR_PARM:
    case OPR_EXTRACT_BITS:

      WFE_Set_ST_Addr_Saved (WN_kid0(wn));
      break;

    case OPR_CSELECT:

      WFE_Set_ST_Addr_Saved (WN_kid1(wn));
      WFE_Set_ST_Addr_Saved (WN_kid2(wn));
      break;

    case OPR_SELECT:
    case OPR_ADD:
    case OPR_SUB:
    case OPR_MPY:
    case OPR_DIV:
    case OPR_MOD:
    case OPR_REM:
    case OPR_DIVREM:
    case OPR_MAX:
    case OPR_MIN:
    case OPR_MINMAX:
    case OPR_BAND:
    case OPR_BIOR:
    case OPR_BXOR:
    case OPR_BNOR:
    case OPR_LAND:
    case OPR_LIOR:
    case OPR_SHL:
    case OPR_ASHR:
    case OPR_LSHR:
    case OPR_COMPLEX:
    case OPR_HIGHMPY:
    case OPR_RROTATE:
    case OPR_COMPOSE_BITS:

      WFE_Set_ST_Addr_Saved (WN_kid0(wn));
      WFE_Set_ST_Addr_Saved (WN_kid1(wn));
      break;

    case OPR_CAND:
    case OPR_CIOR:

      break;

    case OPR_COMMA:

      WFE_Set_ST_Addr_Saved (WN_kid1(wn));
      break;

    case OPR_RCOMMA:

      WFE_Set_ST_Addr_Saved (WN_kid0(wn));
      break;

    default:

      DevWarn ("WFE_Set_ST_Addr_Saved not implemented");
  }
} /* WFE_Set_ST_Addr_Saved */

typedef struct wfe_bind_expr_t {
  tree  rtl_expr;
  WN   *block;
} WFE_BIND_EXPR;

WFE_BIND_EXPR *wfe_bind_expr_stack       = NULL;
INT32          wfe_bind_expr_stack_last  = -1;
INT32          wfe_bind_expr_stack_max   = 0;

void
WFE_Expand_Start_Stmt_Expr (tree t)
{
  WN *block = WN_CreateBlock ();
  WFE_Stmt_Push (block, wfe_stmk_comma, Get_Srcpos ());
} /* WFE_Start_Stmt_Expr */

void
WFE_Expand_End_Stmt_Expr (tree t)
{
  WN *block = WFE_Stmt_Pop (wfe_stmk_comma);
  ++wfe_bind_expr_stack_last;
  if (wfe_bind_expr_stack_last == wfe_bind_expr_stack_max) {
    if (wfe_bind_expr_stack == NULL) {
      wfe_bind_expr_stack_max = 32;
      wfe_bind_expr_stack     =
        (WFE_BIND_EXPR *) malloc (wfe_bind_expr_stack_max *
                                  sizeof (WFE_BIND_EXPR));
    }
    else {
      wfe_bind_expr_stack_max = wfe_bind_expr_stack_max +
                                (wfe_bind_expr_stack_max >> 1);
      wfe_bind_expr_stack     =
        (WFE_BIND_EXPR *) realloc (wfe_bind_expr_stack,
                                   wfe_bind_expr_stack_max *
                                   sizeof (WFE_BIND_EXPR));
    }
  }
  wfe_bind_expr_stack [wfe_bind_expr_stack_last].rtl_expr = t;
  wfe_bind_expr_stack [wfe_bind_expr_stack_last].block    = block;
} /* WFE_End_Stmt_Expr */

typedef struct wfe_save_expr_t {
  tree  exp;
  ST   *st;
} WFE_SAVE_EXPR;

WFE_SAVE_EXPR *wfe_save_expr_stack      = NULL;
INT32          wfe_save_expr_stack_last = -1;
INT32          wfe_save_expr_stack_max  = 0;

static WN*
WFE_Save_Expr (tree save_exp)
{
  INT32     i;
  tree      exp     = TREE_OPERAND (save_exp, 0);
  TY_IDX    ty_idx  = Get_TY (TREE_TYPE (exp));
  if (TY_is_shared(ty_idx)) {
    //convert it to a non-shared type since the tmp var will live on stack
    //see bug1352. 
    ty_idx = Shared_To_Private_Type(ty_idx);
  }

  TYPE_ID   mtype   = TY_mtype (ty_idx);
  ST       *st;
  WN       *wn;

  for (i = wfe_save_expr_stack_last; i >= 0; i--) {
    if (wfe_save_expr_stack [i].exp == exp) {
      st = wfe_save_expr_stack [i].st;
      FmtAssert (st != 0,
                 ("WFE_Save_Expr: st not yet assigned"));
      wn   = WN_Ldid (mtype, 0, st, ty_idx);
      return wn;
    }
  }
  i = ++wfe_save_expr_stack_last;
  if (i == wfe_save_expr_stack_max) {
    if (wfe_save_expr_stack == NULL) {
      wfe_save_expr_stack_max = 32;
      wfe_save_expr_stack     =
        (WFE_SAVE_EXPR *) malloc (wfe_save_expr_stack_max *
                                  sizeof (WFE_SAVE_EXPR));
    }
    else {
      wfe_save_expr_stack_max = wfe_save_expr_stack_max +
                                (wfe_save_expr_stack_max >> 1);
      wfe_save_expr_stack     =
        (WFE_SAVE_EXPR *) realloc (wfe_save_expr_stack,
                                   wfe_save_expr_stack_max *
                                   sizeof (WFE_SAVE_EXPR));
    }
  }
  wfe_save_expr_stack [i].exp = exp;
  wfe_save_expr_stack [i].st  = 0;
  wn = WFE_Expand_Expr (exp);
  st = Gen_Temp_Symbol (ty_idx, "__save_expr");
  WFE_Set_ST_Addr_Saved (wn);
  wn = WN_Stid (mtype, 0, st, ty_idx, wn);
  WFE_Stmt_Append (wn, Get_Srcpos());
  wfe_save_expr_stack [i].st = st;
  wn = WN_Ldid (mtype, 0, st, ty_idx);
  return wn;
} /* WFE_Save_Expr */
        
static void
WFE_Unsave_Expr (tree save_exp)
{
  INT32     i;
  tree      exp     = TREE_OPERAND (save_exp, 0);
  TY_IDX    ty_idx  = Get_TY (TREE_TYPE (exp));
  TYPE_ID   mtype   = TY_mtype (ty_idx);
  ST       *st;
  WN       *wn;

  for (i = wfe_save_expr_stack_last; i >= 0; i--) {
    if (wfe_save_expr_stack [i].exp == exp) {
      wfe_save_expr_stack [i].exp == NULL;
      wfe_save_expr_stack [i].st  == NULL;
      break;
    }
  }
} /* WFE_Unsave_Expr */

static WN* WFE_Compute_Array_Displ(WN *tree) 
{
  WN *result,*product,*elm_size;
  int i,m;
  int n = WN_num_dim(tree);
  WN_ESIZE element_size = WN_element_size(tree);
  TYPE_ID rtype = Pointer_type;

  result = WN_array_index(tree, n-1);
  
  
  /*
   *  result <- result + index[i] * ( dim[n-1] * dim[n-2] ... dim[i+1] )
   */
  for (i = n-2; i >= 0; i--)
    {
      INT32     m;
      WN        *mpy;
      
      product = WN_array_dim(tree, n-1);
      for (m=n-2; m>i; m--)
	{
        product = WN_Mpy(rtype,
                         product,
			 WN_array_dim(tree, m));
	}
      
      mpy = WN_Mpy(rtype,
                   WN_array_index(tree,i),
                   product);
      result = WN_Add(rtype, result, mpy);
    }
  
 
  /*
   *  result <-  result * elm_size
   */
  

  elm_size = WN_Intconst(rtype, element_size);
  result = WN_Mpy(rtype, result, elm_size);
  return result;
  

}

static WN* WFE_Compute_Field_Addr(WN *wn, tree arg0, tree arg1, UINT32 field_id)
{
  Is_True(WN_operator(wn) == OPR_ARRAY,  (""));
 
  WN *base, *esz, *idx;
  TY_IDX base_idx;
  WN *result;
  base = WN_kid0(wn);
 
  result = WN_Add(Pointer_type, base, WFE_Compute_Array_Displ(wn));  
  return result;

}
/* process the tree doing array indicing and return the WN that performs
 * the address computation; ty_idx returns the high-level array type if it
 * is a DECL node, and the element type if it is an ARRAY_REF node.
 */
static WN *
WFE_Array_Expr(tree exp, 
	       TY_IDX *ty_idx, 
	       TY_IDX component_ty_idx,
	       INT64 component_offset,
	       UINT32 field_id, UINT32 incr_field_id = 0)
{
  WN *wn;
  UINT32 old_fid, old_incr_fid;
  enum tree_code code = TREE_CODE (exp);
  if (code == COMPONENT_REF) {
    TY_IDX ty_idx0;
    tree arg0 = TREE_OPERAND(exp, 0); 
    tree arg1 = TREE_OPERAND(exp, 1); 

    if (component_ty_idx == 0) {
      if(TYPE_SHARED(TREE_TYPE(arg0))) {
	ty_idx0 = Make_Shared_Type(Get_TY(TREE_TYPE(exp)), 1);
      } else
	ty_idx0 = Get_TY(TREE_TYPE(exp));
    }
    else ty_idx0 = component_ty_idx;
    Is_True(! DECL_BIT_FIELD(arg1),
	    ("WFE_Array_Expr: address arithmetic cannot use bitfield addresses"));
    
    INT64 ofst;
    
    ofst = (BITSPERBYTE * Get_Integer_Value(DECL_FIELD_OFFSET(arg1)) +
	    Get_Integer_Value(DECL_FIELD_BIT_OFFSET(arg1)))
      / BITSPERBYTE;
  
    /* COSTIN TOFIX */
    if(field_id != 0 || incr_field_id!= 0) {
      old_fid = field_id;
      old_incr_fid = incr_field_id;
      field_id = incr_field_id = 0;
    }
    wn =  WFE_Array_Expr(arg0, ty_idx, ty_idx0, ofst + component_offset,
			 field_id + DECL_FIELD_ID(arg1)
			 // (TY_kind(component_ty_idx) == KIND_STRUCT ? DECL_FIELD_ID(arg1):0)
			 ,    
			 incr_field_id + // (TY_kind(component_ty_idx) == KIND_STRUCT ? DECL_FIELD_ID(arg1):0)
			 			 DECL_FIELD_ID(arg1)
);
    if(old_fid != 0 || old_incr_fid!= 0) {
      field_id = old_fid;
      incr_field_id = old_incr_fid;
    }
   
    if(TREE_CODE(arg0) == ARRAY_REF && TREE_CODE(TREE_OPERAND(arg0,0)) == COMPONENT_REF) {
      //see bugs912,945,1315,946
      wn = WFE_Compute_Field_Addr(wn, arg0, arg1, field_id);
    }


    return wn;
  }
  else if (code == VAR_DECL || code == PARM_DECL) {
    ST *st = Get_ST (exp);
    ST *base_st = ST_base (st);
    // for VLAs the instead of using the ST use its base st
    // also for the time being do not support VLAs within structs
    if (st != base_st) {
      FmtAssert (component_ty_idx == 0,
                 ("VLAs within struct not currently implemented"));
      wn = WN_Ldid (Pointer_Mtype, 0, base_st, ST_type (base_st));
    }
    else
      wn = WN_Lda (Pointer_Mtype, ST_ofst(st)+component_offset, st, incr_field_id);
    if (component_ty_idx == 0)
      *ty_idx = ST_type(st);
    else {
      *ty_idx = component_ty_idx;
      if (TY_align(ST_type(st)) < TY_align(component_ty_idx))
	Set_TY_align(*ty_idx, TY_align(ST_type(st)));//pick more stringent align
    }
    Is_True(TY_kind(*ty_idx) == KIND_ARRAY,
	    ("WFE_Array_Expr: ARRAY_REF base not of type KIND_ARRAY"));
    return wn;
  }
  else if (code == CONSTRUCTOR) {
    ST *st = WFE_Generate_Temp_For_Initialized_Aggregate (exp, "");
    wn = WN_Lda (Pointer_Mtype, ST_ofst(st)+component_offset, st, field_id);
    if (component_ty_idx == 0)
      *ty_idx = ST_type(st);
    else *ty_idx = component_ty_idx;
    Is_True(TY_kind(*ty_idx) == KIND_ARRAY,
	    ("WFE_Array_Expr: ARRAY_REF base not of type KIND_ARRAY"));
    return wn;
  }
  else if (code == STRING_CST) {
    wn = WFE_Expand_Expr(exp);
    *ty_idx = ST_type (TREE_STRING_ST (exp));
    return wn;
  }
  else if (code == INDIRECT_REF) {
    wn = WFE_Expand_Expr(TREE_OPERAND (exp, 0));
    TY_IDX ty_idx0 = Get_TY(TREE_TYPE(exp));
    if (component_ty_idx == 0)
      *ty_idx = ty_idx0;
    else {
      *ty_idx = component_ty_idx;
      INT node_align = TYPE_ALIGN(TREE_TYPE(exp)) / BITSPERBYTE;
      if (node_align < TY_align(component_ty_idx))
	Set_TY_align(*ty_idx, node_align);//pick more stringent align
    }
    
    // TODO: use ILDA instead
    //old code was generating and add(base,offset) - this is
    // wrong since the ptr arithmetic has the type of the struct base
    //pointer
    WN *wn0;
    TY_IDX spill_ty =  Make_Pointer_Type(*ty_idx);  
 
    wn0 = WN_Create(OPR_TAS, Pointer_Mtype, MTYPE_V,1);
    WN_kid0(wn0) = wn;
    WN_set_ty(wn0, Make_Pointer_Type(MTYPE_To_TY(MTYPE_U1))
	      // spill_ty
	      );
    wn = wn0;
    wn0 = WN_Intconst(Pointer_Mtype, component_offset);
    
    wn = WN_Binary(OPR_ADD, Pointer_Mtype, wn, wn0);

    ST *temp = Gen_Temp_Symbol(spill_ty, "arr_base");
   
    wn0 = WN_Stid(Pointer_Mtype, 0, temp,  spill_ty, wn);
    WFE_Stmt_Append (wn0, Get_Srcpos ());
    
    wn = WN_Ldid(Pointer_Mtype, 0, temp, spill_ty);     
    
    return wn;
  }
  else if (code == CALL_EXPR) {
    wn = WFE_Expand_Expr(exp);
    FmtAssert (WN_opcode (wn) == OPC_MCOMMA,
               ("opcode other than OPC_MCOMMA for call underneath ARRAY_REF"));
    WFE_Stmt_Append (WN_kid0 (wn), Get_Srcpos ());
    ST *st = WN_st (WN_kid1 (wn));
    WN_Delete (WN_kid1 (wn));
    WN_Delete (wn);
    wn = WN_Lda (Pointer_Mtype, ST_ofst(st)+component_offset, st, field_id);
    if (component_ty_idx == 0)
      *ty_idx = ST_type(st);
    else {
      *ty_idx = component_ty_idx;
      if (TY_align(ST_type(st)) < TY_align(component_ty_idx))
	Set_TY_align(*ty_idx, TY_align(ST_type(st)));//pick more stringent align
    }
    Is_True(TY_kind(*ty_idx) == KIND_ARRAY,
	    ("WFE_Array_Expr: ARRAY_REF base not of type KIND_ARRAY"));
    return wn;
  }
  else if (code == ARRAY_REF) { // recursive call
    WN *wn0, *wn1, *wn2;
    TY_IDX ty_idx0 = 0, tmp_idx;

    wn0 = WFE_Array_Expr(TREE_OPERAND (exp, 0), &ty_idx0, component_ty_idx,
			 component_offset, field_id, incr_field_id);
    Is_True(TY_kind(ty_idx0) == KIND_ARRAY,
	    ("WFE_Array_Expr: arg 0 of ARRAY_REF not of type KIND_ARRAY"));
    ARB_HANDLE arb = TY_arb(ty_idx0);
    if (ARB_dimension(arb) == 1 && 
	ARB_first_dimen(arb) && ARB_last_dimen(arb) &&
	ARB_const_lbnd(arb)) {
      if (ARB_const_ubnd(arb)) {
        wn1 = WN_Intconst(MTYPE_I4, ARB_ubnd_val(arb) - ARB_lbnd_val(arb) + 1);
#if 0
	/* Don't remember what it was for ...*/
	if (UPC_TYPE_HAS_THREADS(TREE_TYPE(TREE_OPERAND(exp, 0)))) {
	  /* the "THREADS" symbol must have been processed with the array declaration */
	  WN * threads = WN_Ldid(TY_mtype(ST_type(upc_threads_st)), 0, upc_threads_st, TY_align(ST_type(upc_threads_st)), 0); 
	  wn1 = WN_Mpy(MTYPE_I4, wn1, threads);
	}
#endif
      } else
        wn1 = WN_Intconst(MTYPE_I4, 0);
      wn2 = WFE_Expand_Expr (TREE_OPERAND (exp, 1));
      wn = WN_Ternary(OPR_ARRAY, Pointer_Mtype, wn0, wn1, wn2);
      WN_element_size(wn) = TY_size(Get_TY (TREE_TYPE(exp)));
    }
    else Is_True(FALSE,
		 ("WFE_Array_Expr: only const-bounds 1-dimension arrays handled now"));
    if (component_ty_idx == 0) {
      *ty_idx = TY_etype(ty_idx0);
      if (TY_align(ty_idx0) < TY_align(*ty_idx))
	Set_TY_align(*ty_idx, TY_align(ty_idx0));// pick more stringent align
    }
    else *ty_idx = component_ty_idx;
    return wn;
  }
  else {
    Is_True(FALSE,
	    ("WFE_Array_Expr: unsupported node for base of ARRAY_REF"));
    return NULL;
  }
}

/* rhs_wn is the WN representing the rhs of a MODIFY_EXPR node; this
 * routine processes the lhs of the node and generate the appropriate
 * form of store.
 *
 * In special cases where the RHS of the store is unknown but the
 * statement being expanded is nonetheless semantically a store,
 * rhs_wn can be NULL. This happens, for example, for each output
 * operand of an asm statement. When rhs_wn is NULL, we manufacture an
 * RHS that is an LDID of a PREG specified by rhs_preg_num (generally
 * a negative-numbered PREG). If rhs_st is non-NULL, rhs_preg_num is
 * ignored.  assign_code tells if it is {PRE,POST}{IN,DE}CREMENT_EXPR.
 * Ordinarily, it is MODIFY_EXPR.
 */
WN *
WFE_Lhs_Of_Modify_Expr(tree_code assign_code,
		       tree lhs, 
		       bool need_result,
		       TY_IDX component_ty_idx, 
		       INT64 component_offset,
		       UINT32 field_id,
		       bool is_bit_field,
		       WN *rhs_wn,
		       PREG_NUM rhs_preg_num,
		       bool is_realpart,
		       bool is_imagpart)
{
  WN *wn;
  ST *st;
  bool result_in_temp = FALSE;
  ST *result_preg_st;
  PREG_NUM result_preg;
  PREG_NUM lhs_preg_num = 0;
  enum tree_code code = TREE_CODE (lhs);
  BOOL volt = FALSE;
  if (rhs_wn != NULL) {
    WFE_Set_ST_Addr_Saved (rhs_wn);
  }
  if (code == COMPONENT_REF) {
    INT64 ofst;
    TY_IDX ty_idx0;
    tree arg0 = TREE_OPERAND(lhs, 0);
    tree arg1 = TREE_OPERAND(lhs, 1);
    if (component_ty_idx == 0)
      ty_idx0 = Get_TY(TREE_TYPE(lhs));
    else ty_idx0 = component_ty_idx;
    if (DECL_BIT_FIELD(arg1)) 
      is_bit_field = TRUE;
    if (! is_bit_field)
      
	ofst = (BITSPERBYTE * Get_Integer_Value(DECL_FIELD_OFFSET(arg1)) +
		Get_Integer_Value(DECL_FIELD_BIT_OFFSET(arg1)))
	  / BITSPERBYTE;
   
    else ofst = 0;
    wn = WFE_Lhs_Of_Modify_Expr(assign_code, arg0, need_result, ty_idx0, 
				ofst+component_offset,
			        field_id + DECL_FIELD_ID(arg1), is_bit_field, 
				rhs_wn, rhs_preg_num, is_realpart, is_imagpart);
    return wn;
  }

  if (code == REALPART_EXPR) {
    tree arg0 = TREE_OPERAND(lhs, 0);
    TY_IDX ty_idx0 = Get_TY(TREE_TYPE(arg0));
    wn = WFE_Lhs_Of_Modify_Expr(assign_code, arg0, need_result, ty_idx0,
				component_offset, field_id, is_bit_field,
				rhs_wn, rhs_preg_num, TRUE, FALSE);
    return wn;
  }

  if (code == IMAGPART_EXPR) {
    tree arg0 = TREE_OPERAND(lhs, 0);
    TY_IDX ty_idx0 = Get_TY(TREE_TYPE(arg0));
    wn = WFE_Lhs_Of_Modify_Expr(assign_code, arg0, need_result, ty_idx0,
				component_offset, field_id, is_bit_field,
				rhs_wn, rhs_preg_num, FALSE, TRUE);
    return wn;
  }

  if (code == PARM_DECL || code == VAR_DECL) {
    TY_IDX hi_ty_idx = Get_TY(TREE_TYPE(lhs)); // type associated with field_id
    if (TREE_THIS_VOLATILE(lhs)) {
      Set_TY_is_volatile(hi_ty_idx);
      volt = TRUE;
    }
    TY_IDX desc_ty_idx = component_ty_idx;
    if (desc_ty_idx == 0)
      desc_ty_idx = hi_ty_idx;
    if (TY_is_volatile(desc_ty_idx)) {
      Clear_TY_is_volatile(desc_ty_idx);
      volt = TRUE;
    }
    st = Get_ST (lhs);
    if (ST_assigned_to_dedicated_preg (st)) {
      Set_TY_is_volatile(hi_ty_idx);
      volt = TRUE;
    }
    Is_True(! is_bit_field || field_id <= MAX_FIELD_ID,
	    ("WFE_Lhs_Of_Modify_Expr: field id for bit-field exceeds limit"));

    TYPE_ID rtype = Widen_Mtype(TY_mtype(desc_ty_idx));
    TYPE_ID desc = is_bit_field ? MTYPE_BS : TY_mtype(desc_ty_idx);

    if (rhs_wn == NULL) {
      // Manufacture a negative-PREG RHS for the STID we are about to
      // generate. This feature is used in preparing WHIRL ASM
      // statements.
      // TODO: How to support a bit-field output of non-integral
      // number of bytes?
      if (rtype == MTYPE_M && desc == MTYPE_M) {
        FmtAssert(TY_size(desc_ty_idx) == MTYPE_byte_size(Def_Int_Mtype),
                  ("Size of ASM struct opnd must be equal to register size"));
        desc = rtype = Def_Int_Mtype;
        desc_ty_idx = hi_ty_idx = MTYPE_To_TY(Def_Int_Mtype);
      }
      ST *rhs_st = MTYPE_To_PREG(desc);
      rhs_wn = WN_CreateLdid (OPR_LDID, rtype,
			      desc, rhs_preg_num, rhs_st,
			      desc_ty_idx, 0);
    }
    else {
      WN *result_wn;	// the result wn to be returned
      if (assign_code == MODIFY_EXPR) {
	if (is_realpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     rhs_wn,
			     WN_Unary(OPR_IMAGPART,
				      Mtype_complex_to_real (rtype),
				      WN_CreateLdid(OPR_LDID, rtype, desc,
						    ST_ofst(st) + component_offset,
						    st, hi_ty_idx, field_id)));
	else
	if (is_imagpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     WN_Unary(OPR_REALPART,
				      Mtype_complex_to_real (rtype),
				      WN_CreateLdid(OPR_LDID, rtype, desc,
						    ST_ofst(st) + component_offset,
						    st, hi_ty_idx, field_id)),
			     rhs_wn);
      }
      else {
	if (is_realpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     rhs_wn,
			     WN_Floatconst (Mtype_complex_to_real (rtype), 0.0));
	else
	if (is_imagpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     WN_Floatconst (Mtype_complex_to_real (rtype), 0.0),
			     rhs_wn);
      }

      if (assign_code == PREINCREMENT_EXPR ||
	  assign_code == PREDECREMENT_EXPR) {
        wn = WN_CreateLdid (OPR_LDID, rtype, desc, 
			    ST_ofst(st) + component_offset,
			    st, hi_ty_idx, field_id);
        rhs_wn = WN_Binary(Operator_From_Tree [assign_code].opr,
		           rtype, wn, rhs_wn);
	result_wn = rhs_wn;
      }
      else if (assign_code == POSTINCREMENT_EXPR ||
	       assign_code == POSTDECREMENT_EXPR) {
        result_wn = WN_CreateLdid (OPR_LDID, rtype, desc, 
				   ST_ofst(st) + component_offset,
				   st, hi_ty_idx, field_id);
      }
      else result_wn = rhs_wn;

      if (need_result && 
	  (volt ||
	   assign_code == POSTINCREMENT_EXPR ||
	   assign_code == POSTDECREMENT_EXPR)) { // save result in a preg
        result_in_temp = TRUE;
	TY_IDX idx = Get_TY(TREE_TYPE(lhs));
	if (Type_Is_Shared_Ptr(idx)) {
	  //can't spill into a preg because any communication call
	  //can create a lda(preg) which crashes whirl2c
	  if(TY_kind(idx) == KIND_SCALAR) {
	    desc_ty_idx = MTYPE_To_TY(rtype);
	    result_preg_st = Gen_Temp_Symbol(desc_ty_idx, 
					     Index_To_Str(Save_Str2((char*)".Mspillpost.", (char*)"")));
	    
	  } else { 
	    if(TY_is_shared(idx)) {
	      //shared ptr - bug 2306 
	      if(TY_kind(idx) == KIND_POINTER) {
		idx = Make_Pointer_Type(TY_pointed(idx));
	      } else {
		Fail_FmtAssertion("Postincrement on array variable\n");
	      }		
	    }
	    result_preg_st = Gen_Temp_Symbol(idx, 
					     Index_To_Str(Save_Str2((char*)".Mspillpost.", (char*)"")));
	   
	  }
	  result_preg = 0;
	   
	   
	  
	} else {
	  if (compiling_upc) {
	    //we might as well use a tmp symbol to keep more type information
	    result_preg_st = Gen_Temp_Symbol(desc_ty_idx, Index_To_Str(Save_Str2((char*)".spillpost", "")));
	    result_preg = 0;
	  } else {
	    result_preg_st = MTYPE_To_PREG(rtype);
	    result_preg = Create_Preg(rtype, NULL);
	  }
	}
        wn = WN_Stid(rtype, result_preg, result_preg_st, desc_ty_idx, result_wn,
		     0);
        WFE_Stmt_Append (wn, Get_Srcpos());
        result_wn = WN_Ldid(rtype, result_preg, result_preg_st, desc_ty_idx, 0);
      }

      if (assign_code == POSTINCREMENT_EXPR ||
	  assign_code == POSTDECREMENT_EXPR) {
	//if the type of the target is a pointer, in this case the 
	//rhs_wn holds the size of the element. Need to use the 
	//adjusted size for local pointers
	if(TY_kind(hi_ty_idx) == KIND_POINTER && !Type_Is_Shared_Ptr(hi_ty_idx, TRUE)) {
	  Is_True(WN_operator(rhs_wn) == OPR_INTCONST, ("",""));
	  if(TY_kind(TY_pointed(hi_ty_idx)) != KIND_STRUCT)
	    WN_const_val(rhs_wn) = Adjusted_Type_Size(TY_pointed(hi_ty_idx));
	}
        rhs_wn = WN_Binary(Operator_From_Tree [assign_code].opr,
		           rtype, result_wn, rhs_wn);
      }
      else rhs_wn = result_wn;

      // rhs_wn is now always the right-hand-side of the assignment
    }

    // the assignment
    if (!WFE_Keep_Zero_Length_Structs &&
        desc == MTYPE_M               &&
        TY_size (hi_ty_idx) == 0) {
      // ignore zero length structs
    }
    else {   
      if(Type_Is_Shared_Ptr(hi_ty_idx) && 
	 !Type_Is_Shared_Ptr(WN_ty(rhs_wn)) && 
	 (field_id ? TY_mtype(component_ty_idx) != MTYPE_M :
	  TY_mtype(hi_ty_idx) != MTYPE_M))  {
	//bug 1937 - TAS on scalars confuses the optimizer opt_emit_template.h (l 367)
	if(TY_kind(hi_ty_idx) != KIND_SCALAR) {
	  wn  = WN_Create(OPR_TAS, 
			  field_id ? TY_mtype(component_ty_idx) :
			  TY_mtype(hi_ty_idx), 
			  MTYPE_V, 1);
	  WN_kid0(wn) = rhs_wn;
	  WN_set_ty(wn,field_id ? component_ty_idx : hi_ty_idx);
	  rhs_wn = wn;
	}
      } 
      wn = WN_Stid (desc, ST_ofst(st) + component_offset + lhs_preg_num, st,
		    hi_ty_idx, rhs_wn, field_id);
      WFE_Stmt_Append(wn, Get_Srcpos());
    }
    if (need_result) {
      if (! result_in_temp)
        wn = WN_CreateLdid(OPR_LDID, rtype, desc, 
			   ST_ofst(st) + component_offset, st, hi_ty_idx,
			   field_id);
      else wn = WN_Ldid(rtype, result_preg, result_preg_st, desc_ty_idx, 0);
      if (is_realpart)
	wn = WN_Unary (OPR_REALPART, Mtype_complex_to_real (rtype), wn);
      else
	if (is_imagpart)
	  wn = WN_Unary (OPR_IMAGPART, Mtype_complex_to_real (rtype), wn);
    }
    else wn = NULL;
  }
  else if (code == INDIRECT_REF) {
    WN *addr_wn = WFE_Expand_Expr (TREE_OPERAND (lhs, 0));
    TY_IDX hi_ty_idx = Get_TY(TREE_TYPE(lhs));
    //For structure field accesses we need to change the block size 
    // of the pointer thru which the indirection is made.
    // For this case the addr_wn is a TAS with the right type
    if(WN_operator(addr_wn) == OPR_TAS) {
      TY_IDX tmp_idx = WN_ty(addr_wn);
      if(field_id) {
	if(TY_kind(TY_pointed(tmp_idx)) == KIND_ARRAY) {
	  tmp_idx = Make_Pointer_Type(TY_etype(TY_pointed(tmp_idx)));
	} else if(TY_kind(tmp_idx) == KIND_POINTER) {
	  if(assign_code == PREINCREMENT_EXPR || assign_code == PREDECREMENT_EXPR ||
	     assign_code == POSTINCREMENT_EXPR || assign_code == POSTDECREMENT_EXPR)
	    //see bug 1354 - follow ptr types until pointed-to is struct
	    //the code makes a ptr of hi_ty_idx only in that part - 
	    //otherwise it expects a ptr to a struct
	    while(TY_kind(tmp_idx) == KIND_POINTER)
	      tmp_idx = TY_pointed(tmp_idx);
	} 
	hi_ty_idx = tmp_idx;
      } else {
	Is_True(TY_kind(tmp_idx) == KIND_POINTER, ("",""));
	if (Type_Is_Shared_Ptr(tmp_idx, true)) {
	  if (Get_Type_Block_Size(hi_ty_idx) != Get_Type_Block_Size(tmp_idx)) {
	    Is_True(MTYPE_I1 == TY_mtype(TY_pointed(tmp_idx)), ("",""));
	    if(Get_Type_Block_Size(tmp_idx) == 0) 
	      hi_ty_idx = Make_Shared_Type(MTYPE_To_TY(TY_mtype(hi_ty_idx)), 0);
	    else
	      hi_ty_idx = TY_pointed(tmp_idx);
	  }
	}
      }
    }
    
    if (TREE_THIS_VOLATILE(lhs)) {
      Set_TY_is_volatile(hi_ty_idx);
      volt = TRUE;
    }
    TY_IDX desc_ty_idx = component_ty_idx;
    if (desc_ty_idx == 0)
      desc_ty_idx = hi_ty_idx;
    if (TY_is_volatile(desc_ty_idx)) {
      Clear_TY_is_volatile(desc_ty_idx);
      volt = TRUE;
    }
    Is_True(! is_bit_field || field_id <= MAX_FIELD_ID,
	    ("WFE_Lhs_Of_Modify_Expr: field id for bit-field exceeds limit"));
    if (WN_has_side_effects(addr_wn) &&
	(need_result || 
	 assign_code == PREINCREMENT_EXPR ||
	 assign_code == PREDECREMENT_EXPR ||
	 assign_code == POSTINCREMENT_EXPR ||
	 assign_code == POSTDECREMENT_EXPR)) {
      ST       *preg_st;
      PREG_NUM  preg;
      TY_IDX    address_ty_idx = Get_TY (TREE_TYPE (TREE_OPERAND (lhs, 0)));
      preg_st = MTYPE_To_PREG(Pointer_Mtype);
      preg    = Create_Preg (Pointer_Mtype, NULL);
      wn      = WN_Stid (Pointer_Mtype, preg, preg_st, address_ty_idx, addr_wn);
      WFE_Set_ST_Addr_Saved (addr_wn);
      WFE_Stmt_Append (wn, Get_Srcpos());
      addr_wn = WN_Ldid (Pointer_Mtype, preg, preg_st, address_ty_idx);
    }

    TYPE_ID rtype = Widen_Mtype(TY_mtype(desc_ty_idx));
    TYPE_ID desc = is_bit_field ? MTYPE_BS : TY_mtype(desc_ty_idx);

    if (rhs_wn == NULL) {
      // Manufacture a negative-PREG RHS for the ISTORE we are about to
      // generate. This feature is used in preparing WHIRL ASM
      // statements.
      ST *rhs_st;
      // TODO: How to support a bit-field output of non-integral
      // number of bytes?
      rhs_st = MTYPE_To_PREG(desc);
      // Types are likely to be wrong in the following
      rhs_wn = WN_CreateLdid (OPR_LDID, rtype, desc, rhs_preg_num, rhs_st,
			      desc_ty_idx, 0);
    }
    else {
      WN *result_wn;	// the result wn to be returned

      if (assign_code == MODIFY_EXPR) {
	if (is_realpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     rhs_wn,
			     WN_Unary(OPR_IMAGPART,
				      Mtype_complex_to_real (rtype),
				      WN_CreateIload(OPR_ILOAD, rtype, desc,
						     component_offset,
						     desc_ty_idx,
						     Make_Pointer_Type(hi_ty_idx, FALSE),
						     WN_COPY_Tree (addr_wn),
						     field_id)));
	else
	if (is_imagpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     WN_Unary(OPR_REALPART,
				      Mtype_complex_to_real (rtype),
				      WN_CreateIload(OPR_ILOAD, rtype, desc,
						     component_offset,
						     desc_ty_idx,
						     Make_Pointer_Type(hi_ty_idx, FALSE),
						     WN_COPY_Tree (addr_wn),
						     field_id)),
			     rhs_wn);
      }
      else {
	if (is_realpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     rhs_wn,
			     WN_Floatconst (Mtype_complex_to_real (rtype), 0.0));
	else
	if (is_imagpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     WN_Floatconst (Mtype_complex_to_real (rtype), 0.0),
			     rhs_wn);
      }

      if (assign_code == PREINCREMENT_EXPR ||
	  assign_code == PREDECREMENT_EXPR) {
        wn = WN_CreateIload (OPR_ILOAD, rtype, desc, component_offset,
			     //field_id != 0 ? hi_ty_idx : desc_ty_idx,
			     desc_ty_idx,
			     Make_Pointer_Type(hi_ty_idx, FALSE),
			     //Make_Pointer_Type(field_id != 0 ? desc_ty_idx : hi_ty_idx, FALSE),
			     WN_COPY_Tree (addr_wn),
			     field_id);
        rhs_wn = WN_Binary(Operator_From_Tree [assign_code].opr,
                           rtype, wn, rhs_wn);
        result_wn = rhs_wn;
      }
      else if (assign_code == POSTINCREMENT_EXPR ||
	       assign_code == POSTDECREMENT_EXPR) {
	TY_IDX t_idx;
	t_idx = desc_ty_idx;
	
	if(field_id && Type_Is_Shared_Ptr(hi_ty_idx) && 
	   !Type_Is_Shared_Ptr(desc_ty_idx)) {
	    t_idx = Make_Shared_Type(desc_ty_idx,0);
	} 
	result_wn = WN_CreateIload (OPR_ILOAD, rtype, desc, component_offset,
				    t_idx,
				    Make_Pointer_Type(hi_ty_idx, FALSE),
				    WN_COPY_Tree (addr_wn),
				    field_id);
      }
      else result_wn = rhs_wn;

      if (need_result && 
	  (volt ||
           assign_code == POSTINCREMENT_EXPR ||
           assign_code == POSTDECREMENT_EXPR)) { // save result in a preg
	result_in_temp = TRUE;
        
	result_preg_st = MTYPE_To_PREG(rtype);
	result_preg = Create_Preg(rtype, NULL);
	if (TY_is_shared(desc_ty_idx)) {
	    //bug1559; preg_st should not have shared type
	    desc_ty_idx = Shared_To_Private_Type(desc_ty_idx);
	}
	wn = WN_Stid(rtype, result_preg, result_preg_st, desc_ty_idx, result_wn,
		     0);
        WFE_Stmt_Append (wn, Get_Srcpos());;
        result_wn = WN_Ldid(rtype, result_preg, result_preg_st, desc_ty_idx, 0);
      }

      if (assign_code == POSTINCREMENT_EXPR ||
	  assign_code == POSTDECREMENT_EXPR) {
        rhs_wn = WN_Binary(Operator_From_Tree [assign_code].opr,
                           rtype, result_wn, rhs_wn);
      }
      else rhs_wn = result_wn;

      // rhs_wn is now always the right-hand-side of the assignment
    }

    // the assignment
    if (!WFE_Keep_Zero_Length_Structs &&
        desc == MTYPE_M               &&
        TY_size (hi_ty_idx) == 0) {
      // ignore zero length structs
      if (WN_has_side_effects (addr_wn)) {
	wn = WN_CreateEval (addr_wn);
	WFE_Stmt_Append (wn, Get_Srcpos());
      }
      wn = NULL;
    }
    else {
      TY_IDX hi_ty_idx1 = hi_ty_idx;
      if (TYPE_SHARED(TREE_TYPE(lhs))) {
	if(code == INDIRECT_REF && WN_operator(addr_wn) == OPR_TAS &&  field_id) {
	  Is_True(Type_Is_Shared_Ptr(hi_ty_idx) && 
		  (TY_kind(hi_ty_idx) == KIND_POINTER ||
		   TY_kind(hi_ty_idx) == KIND_ARRAY ) , ("",""));
	  /* bug 1565 - we should enable Is_True */
	  if(TY_kind(hi_ty_idx) != KIND_POINTER &&  TY_kind(hi_ty_idx) != KIND_ARRAY)
	    hi_ty_idx1 =  Make_Pointer_Type (hi_ty_idx, FALSE);
	} else {
	  int bsize =  TY_block_size(hi_ty_idx);
	  hi_ty_idx1 = Make_Shared_Type(hi_ty_idx, bsize, 
				       TY_is_strict(hi_ty_idx) ?  STRICT_CONSISTENCY :
				       NO_CONSISTENCY );
	  hi_ty_idx1 =  Make_Pointer_Type (hi_ty_idx1, FALSE);
	}
      } else 
	hi_ty_idx1 =  Make_Pointer_Type (hi_ty_idx, FALSE);
      

      if (field_id && TY_kind(TY_pointed(hi_ty_idx1)) != KIND_STRUCT) {
	//This can be the case for a mix of array and struct field accesses, e.g.
	//fooo.x[exp].y
	//Since the lhs address will be flattened to the correct offset, we can set field id to be zero. 
	field_id = 0; 
      }
      
      wn = WN_CreateIstore(OPR_ISTORE, MTYPE_V, desc, component_offset, 
			   hi_ty_idx1,
			   rhs_wn, addr_wn, field_id);
      WFE_Stmt_Append(wn, Get_Srcpos());
      if (need_result) {
	
	if(Type_Is_Shared_Ptr(hi_ty_idx) && field_id && !Type_Is_Shared_Ptr(desc_ty_idx)) {
	  // a field of a shared struct. Need to mark the access type as shared.
	  // bug321 
	  //   dbg->real = dbg->imag = 0; 
	 // bug 979 - mark as shared only if it is modify and not pre/postinc
	  if(assign_code == MODIFY_EXPR)
	    desc_ty_idx = Make_Shared_Type(desc_ty_idx,0);
	}
	
	if (! result_in_temp)
          wn = WN_CreateIload(OPR_ILOAD, rtype, desc, component_offset,
			      desc_ty_idx,
			      Make_Pointer_Type (hi_ty_idx, FALSE),
			      WN_COPY_Tree (addr_wn),
			      field_id);
	else wn = WN_Ldid(rtype, result_preg, result_preg_st, desc_ty_idx, 0);
	if (is_realpart)
	  wn = WN_Unary (OPR_REALPART, Mtype_complex_to_real (rtype), wn);
	else
	if (is_imagpart)
	  wn = WN_Unary (OPR_IMAGPART, Mtype_complex_to_real (rtype), wn);
      }
      else wn = NULL;
    }
  }
  else if (code == ARRAY_REF) {
    TY_IDX elem_ty_idx;
    // generate the WHIRL array node
    WN *addr_wn = WFE_Array_Expr(lhs, &elem_ty_idx, 0, 0, 0);
    if (TY_is_volatile(elem_ty_idx))
      volt = TRUE;
    TY_IDX desc_ty_idx = component_ty_idx;
    if (desc_ty_idx == 0)
      desc_ty_idx = Get_TY (TREE_TYPE(lhs));
    if (TY_is_volatile(desc_ty_idx)) {
      Clear_TY_is_volatile(desc_ty_idx);
      volt = TRUE;
    }
    Is_True(! is_bit_field || field_id <= MAX_FIELD_ID,
	    ("WFE_Lhs_Of_Modify_Expr: field id for bit-field exceeds limit"));
    if (WN_has_side_effects(addr_wn) &&
        (need_result ||
         assign_code == PREINCREMENT_EXPR ||
         assign_code == PREDECREMENT_EXPR ||
         assign_code == POSTINCREMENT_EXPR ||
	 assign_code == POSTDECREMENT_EXPR)) {
      ST       *preg_st;
      PREG_NUM  preg;
      TY_IDX    address_ty_idx = Make_Pointer_Type(elem_ty_idx, FALSE);
      preg_st = MTYPE_To_PREG(Pointer_Mtype);
      preg    = Create_Preg (Pointer_Mtype, NULL);
      wn      = WN_Stid (Pointer_Mtype, preg, preg_st, address_ty_idx, addr_wn);
      WFE_Set_ST_Addr_Saved (addr_wn);
      WFE_Stmt_Append (wn, Get_Srcpos());
      addr_wn = WN_Ldid (Pointer_Mtype, preg, preg_st, address_ty_idx);
    }

    TYPE_ID rtype = Widen_Mtype(TY_mtype(desc_ty_idx));
    TYPE_ID desc = is_bit_field ? MTYPE_BS : TY_mtype(desc_ty_idx);

    if (rhs_wn == NULL) {
      // Manufacture a negative-PREG RHS for the ISTORE we are about to
      // generate. This feature is used in preparing WHIRL ASM
      // statements.
      ST *rhs_st;
      // TODO: How to support a bit-field output of non-integral
      // number of bytes?
      rhs_st = MTYPE_To_PREG(desc);
      rhs_wn = WN_CreateLdid (OPR_LDID, rtype, desc, rhs_preg_num, rhs_st,
			      desc_ty_idx, 0);
    }
    else {
      WN *result_wn;    // the result wn to be returned

      if (assign_code == MODIFY_EXPR) {
	if (is_realpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     rhs_wn,
			     WN_Unary(OPR_IMAGPART,
				      Mtype_complex_to_real (rtype),
				      WN_CreateIload(OPR_ILOAD, rtype, desc,
						     component_offset,
						     //field_id != 0 ? elem_ty_idx : desc_ty_idx,
						     desc_ty_idx,
						     Make_Pointer_Type(elem_ty_idx, FALSE),
						     WN_COPY_Tree (addr_wn),
						     field_id)));
	else
	if (is_imagpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     WN_Unary(OPR_REALPART,
				      Mtype_complex_to_real (rtype),
				      WN_CreateIload(OPR_ILOAD, rtype, desc,
						     component_offset,
						     //field_id != 0 ? elem_ty_idx : desc_ty_idx,
						     desc_ty_idx,
						     Make_Pointer_Type(elem_ty_idx, FALSE),
						     WN_COPY_Tree (addr_wn),
						     field_id)),
			     rhs_wn);
      }
      else {
	if (is_realpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     rhs_wn,
			     WN_Floatconst (Mtype_complex_to_real (rtype), 0.0));
	else
	if (is_imagpart)
	  rhs_wn = WN_Binary(OPR_COMPLEX, rtype,
			     WN_Floatconst (Mtype_complex_to_real (rtype), 0.0),
			     rhs_wn);
      }

      if (assign_code == PREINCREMENT_EXPR ||
          assign_code == PREDECREMENT_EXPR) {
        wn = WN_CreateIload (OPR_ILOAD, rtype, desc, component_offset,
                             //field_id != 0 ? elem_ty_idx : desc_ty_idx,
			     desc_ty_idx,
                             Make_Pointer_Type(elem_ty_idx, FALSE),
			     //Make_Pointer_Type(field_id != 0 ? desc_ty_idx : elem_ty_idx, FALSE),
                             WN_COPY_Tree (addr_wn),
                             field_id);
        rhs_wn = WN_Binary(Operator_From_Tree [assign_code].opr,
                           rtype, wn, rhs_wn);
	result_wn = rhs_wn;
      }
      else if (assign_code == POSTINCREMENT_EXPR ||
	       assign_code == POSTDECREMENT_EXPR) {
        result_wn = WN_CreateIload (OPR_ILOAD, rtype, desc, component_offset,
				    //field_id != 0 ? elem_ty_idx : desc_ty_idx,
				    desc_ty_idx,
				    //Make_Pointer_Type (field_id != 0 ? desc_ty_idx : elem_ty_idx, FALSE),
				    Make_Pointer_Type (elem_ty_idx, FALSE),
				    WN_COPY_Tree (addr_wn),
				    field_id);
      }
      else result_wn = rhs_wn;

      if (need_result && 
	  (volt ||
           assign_code == POSTINCREMENT_EXPR ||
	   assign_code == POSTDECREMENT_EXPR)) { // save result in a preg
	  result_in_temp = TRUE;
	  result_preg_st = MTYPE_To_PREG(rtype);
	  result_preg = Create_Preg(rtype, NULL);
	  wn = WN_Stid(rtype, result_preg, result_preg_st, desc_ty_idx, result_wn,
		       0);
	  WFE_Stmt_Append (wn, Get_Srcpos());;
	  result_wn = WN_Ldid(rtype, result_preg, result_preg_st, desc_ty_idx, 0);
      }

      if (assign_code == POSTINCREMENT_EXPR ||
          assign_code == POSTDECREMENT_EXPR) {
        rhs_wn = WN_Binary(Operator_From_Tree [assign_code].opr,
                           rtype, result_wn, rhs_wn);
      }
      else rhs_wn = result_wn;

      // rhs_wn is now always the right-hand-side of the assignment
    }

    // the assignment
    if (!WFE_Keep_Zero_Length_Structs &&
        desc == MTYPE_M               &&
        TY_size (elem_ty_idx) == 0) {
      // ignore zero length structs
      if (WN_has_side_effects (addr_wn)) {
        wn = WN_CreateEval (addr_wn);
        WFE_Stmt_Append (wn, Get_Srcpos());
      }
      wn = NULL;
    }
    else {
      wn = WN_CreateIstore(OPR_ISTORE, MTYPE_V, desc, component_offset, 
			   Make_Pointer_Type(elem_ty_idx, FALSE), rhs_wn,
			   addr_wn, field_id);
      WFE_Stmt_Append(wn, Get_Srcpos());
      if (need_result) {
        if (! result_in_temp)
	  wn = WN_CreateIload (OPR_ILOAD, rtype, desc, component_offset,
			       //field_id != 0 ? elem_ty_idx : desc_ty_idx,
			       desc_ty_idx,
			       //Make_Pointer_Type (field_id != 0 ? desc_ty_idx : elem_ty_idx, FALSE),
                               Make_Pointer_Type (elem_ty_idx, FALSE),
			       WN_COPY_Tree (addr_wn),
			       field_id);
	else wn = WN_Ldid(rtype, result_preg, result_preg_st, desc_ty_idx, 0);
	if (is_realpart)
	  wn = WN_Unary (OPR_REALPART, Mtype_complex_to_real (rtype), wn);
	else
	if (is_imagpart)
	  wn = WN_Unary (OPR_IMAGPART, Mtype_complex_to_real (rtype), wn);
      }
      else wn = NULL;
    }
  }
  else Fail_FmtAssertion ("WFE_Lhs_Of_Modify_Expr: unhandled tree node in LHS of MODIFY_EXPR");
  return wn;
}

/* ============================================================================
 *
 * WFE_Expand_Expr_With_Sequence_Point
 *
 * This routine is invoked instead of WN_Expand_Expr to handle the
 * following expression nodes
 *
 *   both operands of && and ||
 *   all three operands of conditional ?
 *   controlling expression of if
 *   controlling expression of switch
 *   controlling expression of while
 *   statement expression
 *
 * In order to generate WHIRL for an expression with side effects,
 * we would like to move operations such as calls, pre increment/decrement
 * into a comma operator, and operations such as post increment/decrement
 * into a rcomma operator.
 *
 * Sequence points related to function call and return are not handled
 * here as we cannot generate RCOMMA nodes in these cases.
 *
 * ============================================================================
 */

WN*
WFE_Expand_Expr_With_Sequence_Point (tree exp, TYPE_ID mtype)
{
  WN *wn;

  if (mtype == MTYPE_V)
    wn = WFE_Expand_Expr (exp, FALSE);

  else {

    WN *comma_block      = WN_CreateBlock ();

    WFE_Stmt_Push (comma_block, wfe_stmk_comma, Get_Srcpos ());
    wn = WFE_Expand_Expr (exp);
    WFE_Stmt_Pop (wfe_stmk_comma);
    if (WN_first (comma_block)) {
      if (wn)
	wn = WN_CreateComma (OPR_COMMA, Mtype_comparison (mtype), MTYPE_V,
			     comma_block, wn);
      else {
	DevWarn("WFE_Expand_Expr_With_Sequence_Point: no wn for COMMA");
	WFE_Stmt_Append (comma_block, Get_Srcpos());
      }
    }
    else
      WN_Delete (comma_block);
  }

  return wn;
} /* WFE_Expand_Expr_With_Sequence_Point */

static void
emit_barrier (bool type, tree list, INT32 k)
{
  INT32  i;
  WN    *wn = WN_CreateBarrier (type, k);

  for (i = 0; i < k; i++) {
    tree exp = TREE_VALUE (list);
    ST *st   = Get_ST (exp);
    WN_kid (wn, i) = WN_Lda (Pointer_Mtype, 0, st,
                             Make_Pointer_Type (ST_type (st), FALSE));
    list = TREE_CHAIN (list);
  }

  WFE_Stmt_Append (wn, Get_Srcpos());
} /* emit_barrier */

static WN *
emit_builtin_lock_test_and_set (tree exp, INT32 k)
{
  WN        *wn;
  WN        *arg_wn;
  WN        *ikids [2];
  TYPE_ID    obj_mtype;
  TY_IDX     arg_ty_idx;
  TYPE_ID    arg_mtype;
  tree       list = TREE_OPERAND (exp, 1);
  OPCODE     opc;
  INTRINSIC  iopc;

  obj_mtype  = TY_mtype (TY_pointed (Get_TY(TREE_TYPE(TREE_VALUE(list)))));
  arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
  arg_mtype  = TY_mtype (arg_ty_idx);
  arg_wn     = WFE_Expand_Expr (TREE_VALUE (list));
  arg_wn     = WN_CreateParm (arg_mtype, arg_wn, arg_ty_idx, WN_PARM_BY_VALUE);
  ikids [0]  = arg_wn;
  list       = TREE_CHAIN (list);
  arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
  arg_mtype  = TY_mtype (arg_ty_idx);
  arg_wn     = WFE_Expand_Expr (TREE_VALUE (list));
  arg_wn     = WN_CreateParm (arg_mtype, arg_wn, arg_ty_idx, WN_PARM_BY_VALUE);
  ikids [1]  = arg_wn;
  list       = TREE_CHAIN (list);

  if (obj_mtype == MTYPE_I4) {
    opc  = OPC_I4INTRINSIC_CALL;
    iopc = INTRN_LOCK_TEST_AND_SET_I4;
  }
  else
  if (obj_mtype == MTYPE_U4) {
    opc  = OPC_U4INTRINSIC_CALL;
    iopc = INTRN_LOCK_TEST_AND_SET_I4;
  }
  else
  if (obj_mtype == MTYPE_I8) {
    opc  = OPC_I8INTRINSIC_CALL;
    iopc = INTRN_LOCK_TEST_AND_SET_I8;
  }
  else
  if (obj_mtype == MTYPE_U8) {
    opc  = OPC_U8INTRINSIC_CALL;
    iopc = INTRN_LOCK_TEST_AND_SET_I8;
  }
  else {
    Fail_FmtAssertion ("unknown object type in __builtin_lock_test_and_set");
    opc  = OPCODE_UNKNOWN;
    iopc = INTRINSIC_NONE;
  }

  wn = WN_Create_Intrinsic (opc, iopc, 2, ikids);
  WFE_Stmt_Append (wn, Get_Srcpos());

  ST       *preg_st = MTYPE_To_PREG(obj_mtype);
  TY_IDX    preg_ty_idx = Be_Type_Tbl(obj_mtype);
  PREG_NUM  preg = Create_Preg (obj_mtype, NULL);

  wn = WN_Ldid (obj_mtype, -1, Return_Val_Preg, preg_ty_idx);
  wn = WN_Stid (obj_mtype, preg, preg_st, preg_ty_idx, wn),
  WFE_Stmt_Append (wn, Get_Srcpos());

  emit_barrier (FALSE, list, k);

  wn = WN_Ldid (obj_mtype, preg, preg_st, preg_ty_idx);

  return wn;
} /* emit_builtin_lock_test_and_set */

static void
emit_builtin_lock_release (tree exp, INT32 k)
{
  WN        *wn;
  WN        *arg_wn;
  WN        *ikids [1];
  TYPE_ID    obj_mtype;
  TY_IDX     arg_ty_idx;
  TYPE_ID    arg_mtype;
  tree       list = TREE_OPERAND (exp, 1);
  OPCODE     opc;
  INTRINSIC  iopc;

  obj_mtype  = TY_mtype (TY_pointed (Get_TY(TREE_TYPE(TREE_VALUE(list)))));
  arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
  arg_mtype  = TY_mtype (arg_ty_idx);
  arg_wn     = WFE_Expand_Expr (TREE_VALUE (list));
  arg_wn     = WN_CreateParm (arg_mtype, arg_wn, arg_ty_idx, WN_PARM_BY_VALUE);
  ikids [0]  = arg_wn;
  list       = TREE_CHAIN (list);

  emit_barrier (TRUE, list, k);

  opc = OPC_VINTRINSIC_CALL;
  if (obj_mtype == MTYPE_I4)
    iopc = INTRN_LOCK_RELEASE_I4;
  else
  if (obj_mtype == MTYPE_U4)
    iopc = INTRN_LOCK_RELEASE_I4;
  else
  if (obj_mtype == MTYPE_I8)
    iopc = INTRN_LOCK_RELEASE_I8;
  else
  if (obj_mtype == MTYPE_U8)
    iopc = INTRN_LOCK_RELEASE_I8;
  else {
    Fail_FmtAssertion ("unknown object type in __builtin_lock_test_and_set");
    opc  = OPCODE_UNKNOWN;
    iopc = INTRINSIC_NONE;
  }

  wn = WN_Create_Intrinsic (opc, iopc, 1, ikids);
  WFE_Stmt_Append (wn, Get_Srcpos());
} /* emit_builtin_lock_release */

static WN *
emit_builtin_compare_and_swap (tree exp, INT32 k)
{
  WN        *wn;
  WN        *arg_wn;
  WN        *ikids [3];
  TYPE_ID    obj_mtype;
  TY_IDX     arg_ty_idx;
  TYPE_ID    arg_mtype;
  tree       list = TREE_OPERAND (exp, 1);
  OPCODE     opc;
  INTRINSIC  iopc;

  obj_mtype  = TY_mtype (TY_pointed (Get_TY(TREE_TYPE(TREE_VALUE(list)))));
  arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
  arg_mtype  = TY_mtype (arg_ty_idx);
  arg_wn     = WFE_Expand_Expr (TREE_VALUE (list));
  arg_wn     = WN_CreateParm (arg_mtype, arg_wn, arg_ty_idx, WN_PARM_BY_VALUE);
  ikids [0]  = arg_wn;
  list       = TREE_CHAIN (list);
  arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
  arg_mtype  = TY_mtype (arg_ty_idx);
  arg_wn     = WFE_Expand_Expr (TREE_VALUE (list));
  arg_wn     = WN_CreateParm (arg_mtype, arg_wn, arg_ty_idx, WN_PARM_BY_VALUE);
  ikids [1]  = arg_wn;
  list       = TREE_CHAIN (list);
  arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
  arg_mtype  = TY_mtype (arg_ty_idx);
  arg_wn     = WFE_Expand_Expr (TREE_VALUE (list));
  arg_wn     = WN_CreateParm (arg_mtype, arg_wn, arg_ty_idx, WN_PARM_BY_VALUE);
  ikids [2]  = arg_wn;
  list       = TREE_CHAIN (list);

  emit_barrier (TRUE, list, k);

  opc = OPC_I4INTRINSIC_CALL;
  if (obj_mtype == MTYPE_I4)
    iopc = INTRN_COMPARE_AND_SWAP_I4;
  else
  if (obj_mtype == MTYPE_U4)
    iopc = INTRN_COMPARE_AND_SWAP_I4;
  else
  if (obj_mtype == MTYPE_I8)
    iopc = INTRN_COMPARE_AND_SWAP_I8;
  else
  if (obj_mtype == MTYPE_U8)
    iopc = INTRN_COMPARE_AND_SWAP_I8;
  else {
    Fail_FmtAssertion ("unknown object type in __builtin_lock_test_and_set");
    opc  = OPCODE_UNKNOWN;
    iopc = INTRINSIC_NONE;
  }

  wn = WN_Create_Intrinsic (opc, iopc, 3, ikids);
  WFE_Stmt_Append (wn, Get_Srcpos());

  ST       *preg_st = MTYPE_To_PREG(MTYPE_I4);
  TY_IDX    preg_ty_idx = Be_Type_Tbl(MTYPE_I4);
  PREG_NUM  preg = Create_Preg (MTYPE_I4, NULL);

  wn = WN_Ldid (MTYPE_I4, -1, Return_Val_Preg, preg_ty_idx);
  wn = WN_Stid (MTYPE_I4, preg, preg_st, preg_ty_idx, wn),
  WFE_Stmt_Append (wn, Get_Srcpos());

  emit_barrier (FALSE, list, k);

  wn = WN_Ldid (MTYPE_I4, preg, preg_st, preg_ty_idx);

  return wn;
} /* emit_builtin_compare_and_swap */

//Find the correpsonding intrinsics to the upc library functions, so that 
//we can identify them easily later while in the optimizer
//TODO:  allocator functions are currently not converted into intrinsics,
//we probably should make them so in the future.
static INTRINSIC 
find_upc_intrinsic(const char * name) {

  if(strcmp(name, "upc_threadof") == 0) {
    return INTRN_THREADOF_S; 
  } else if (strcmp(name, "upc_phaseof") == 0) {
    return INTRN_PHASEOF_S;
  } else if (strcmp(name, "upc_addrfield") == 0) {
    return INTRN_ADDROF_S;
  } else if (strcmp(name, "upc_affinitysize") == 0) {
    return INTRN_AFF_SIZE;
  } else if (strcmp(name, "upc_lock") == 0) {
    return INTRN_LOCK;
  } else if (strcmp(name, "upc_lock_attempt") == 0) {
    return INTRN_LOCK_ATTEMPT;
  } else if (strcmp(name, "upc_unlock") == 0) {
    return INTRN_UNLOCK;
  } else if (strcmp(name, "upc_memget") == 0) {
    return INTRN_UPC_MEMGET;
  } else if (strcmp(name, "upc_memput") == 0) {
    return INTRN_UPC_MEMPUT;
  } else if (strcmp(name, "upc_memcpy") == 0) {
    return INTRN_UPC_MEMCPY;
  } else if (strcmp(name, "upc_memset") == 0) {
    return INTRN_UPC_MEMSET;
  }

  return INTRINSIC_NONE;
    
}

static void
emit_builtin_synchronize (tree exp, INT32 k)
{
  WN *wn;
  tree list = TREE_OPERAND (exp, 1);
  emit_barrier (TRUE,  list, k);
  wn = WN_Create_Intrinsic (OPC_VINTRINSIC_CALL, INTRN_SYNCHRONIZE, 0, NULL);
  WFE_Stmt_Append (wn, Get_Srcpos());
  emit_barrier (FALSE, list, k);
} /* emit_builtin_synchronize */

static char *
get_string_pointer (WN *wn)
{
  char *ptr = NULL;

  if (WN_operator (wn) == OPR_LDA) {
    ST *st = WN_st (wn);
    if (ST_class (st) == CLASS_CONST) {
      TCON tcon = Tcon_Table [ST_tcon (st)];
      if (TCON_ty (tcon) == MTYPE_STRING)
        ptr = ((char *) Targ_String_Address (tcon)) + WN_offset (wn);
    }
  }

  return ptr;
} /* get_string_pointer */

/* expand gnu expr tree into symtab & whirl */
WN *
WFE_Expand_Expr (tree exp, 
		 bool need_result,
		 TY_IDX nop_ty_idx, 
		 TY_IDX component_ty_idx, 
		 INT64 component_offset,
		 UINT16 field_id,
		 bool is_bit_field)
{
  enum tree_code code = TREE_CODE (exp);
  WN *wn, *wn0, *wn1, *wn2;
  ST *st;
  TY_IDX ty_idx;
  TY_IDX desc_ty_idx;
  tree arg0, arg1, arg2;

  wn = NULL;

#ifdef WFE_DEBUG
  fprintf (stderr,
           "{( WFE_Expand_Expr: %s\n", Operator_From_Tree [code].name); // ")}"
#endif /* WFE_DEBUG */

  switch (code)
    {
    // leaves
    case ADDR_EXPR:
      {
	arg0 = TREE_OPERAND (exp, 0);
	enum tree_code code0 = TREE_CODE (arg0);
	switch (code0) {
	  case VAR_DECL:
	  case PARM_DECL:
	  case FUNCTION_DECL:
	    {
	      st = Get_ST (arg0);
	      ty_idx = (code0 == FUNCTION_DECL) ? ST_pu_type(st) : ST_type (st);
              // for VLAs, use the base_st instead of st
              if (code0 == VAR_DECL &&
                  st != ST_base(st)) {
                FmtAssert (ST_ofst (st) == 0,
                           ("VLA within struct not currently implemented"));
                wn = WN_Ldid (Pointer_Mtype, 0, ST_base(st), ST_type(ST_base(st)));
              }
              else
	      if (!WFE_Keep_Zero_Length_Structs &&
                  code0 == PARM_DECL            &&
		  TY_mtype (ty_idx) == MTYPE_M  &&
		  TY_size (ty_idx) == 0) {
		// taking address of zero length struct passed as parameter
		DevWarn ("taking address of zero length struct %s at line %d",
			 ST_name (st), lineno);
		wn = WN_Intconst (Pointer_Mtype, 0);
	      }
	      else {
	        wn = WN_Lda (Pointer_Mtype, ST_ofst(st), st);
	      }
	    }
	    break;

	  case STRING_CST:
	    {
              TCON tcon;
              tcon = Host_To_Targ_String (MTYPE_STRING,
                                          TREE_STRING_POINTER(arg0),
                                          TREE_STRING_LENGTH(arg0));
              ty_idx = Get_TY(TREE_TYPE(arg0));
              st = New_Const_Sym (Enter_tcon (tcon), ty_idx);
	      wn = WN_Lda (Pointer_Mtype, ST_ofst(st), st);
	      TREE_STRING_ST (arg0) = st;
	    }
	    break;

	  case CONSTRUCTOR:
	    {
		st = WFE_Generate_Temp_For_Initialized_Aggregate (arg0, "");
	        wn = WN_Lda (Pointer_Mtype, ST_ofst(st), st);
	    }
	    break;

          case LABEL_DECL:
            {
              DevWarn ("taking address of a label at line %d", lineno);
              LABEL_IDX label_idx = WFE_Get_LABEL (arg0, FALSE);
              FmtAssert (arg0->decl.symtab_idx == CURRENT_SYMTAB,
                         ("line %d: taking address of a label not defined in current function currently not implemented", lineno));
              wn = WN_LdaLabel (Pointer_Mtype, label_idx);
	      Set_LABEL_addr_saved (label_idx);
              Set_PU_no_inline (Get_Current_PU ());
            }
            break;

	  case COMPONENT_REF:
	    {
	      ST *st;
	      UINT tmp_field_id = 0;
	      TY_IDX spill_ty;
	      wn = WFE_Expand_Expr (arg0);
              ty_idx = Get_TY(TREE_TYPE(arg0));
	      if (WN_operator (wn) == OPR_TAS) {
		// This is sufficient for bug2962, but has otherwise been untested.
                wn = WN_kid0(wn);
	        // Falls through to the OPT_LDID or OPR_ILOAD
		// Should we be patching their type?
	      }
	      if (WN_operator (wn) == OPR_LDID) {
		if (Type_Is_Shared_Ptr(WN_ty(wn))) {
		  
		  // we have here a load of a struct field of a
		  // shared struct 
		  spill_ty = WN_ty(wn);
		  //spill the expr that computes the struct addr
		  // otherwise can't tell in the backend which part
		  // contributes to the ptr arith and which part
		  // contributes to the offset
		  
		  //If the type of the field is T, the type of the temp
		  // is shared [] T ;
		  spill_ty = FLD_type(FLD_get_to_field(WN_ty(wn), 
						       WN_field_id(wn), tmp_field_id));
		  switch(TY_kind(spill_ty)) {
		  case KIND_SCALAR:
		  case KIND_STRUCT:
		    spill_ty = Make_Pointer_Type(Make_Shared_Type(spill_ty, 0));
		    break;
		    //on pointer and array fields usually a load is followed
		    // by ptr arithmetic.
		    // Since the displacements are already lowered, i.e. idx*el_size
		    // cast to shared [] char 
		  case KIND_POINTER:
		    spill_ty = // Make_Shared_Type(spill_ty, 0)
		      Make_Pointer_Type(Make_Shared_Type(MTYPE_To_TY(MTYPE_I1), 0));
		    break;
		  case KIND_ARRAY:
		    spill_ty = Make_Pointer_Type(Make_Shared_Type(// TY_etype(spill_ty)
								  MTYPE_To_TY(MTYPE_I1), 0));
		    break;
		  default:
		    Is_True(0,("",""));
		  }

		  
		  st =  Gen_Temp_Symbol(spill_ty,
					    Index_To_Str(Save_Str2((char*)".Mspillstruct.",
								   (char*)"")));
		  wn = WN_Lda (Pointer_Mtype, WN_offset (wn), WN_st (wn));
		  wn2 = WN_CreateBlock();
		  wn1  = WN_Create(OPR_TAS, TY_mtype(spill_ty), MTYPE_V, 1);
		  
		  WN_kid0(wn1) = wn;
		  WN_set_ty(wn1,spill_ty);
		  WN_INSERT_BlockLast (wn2, WN_Stid (TY_mtype(spill_ty), 0, st, spill_ty, wn1));
		  wn0 = WN_CreateComma (OPR_COMMA, TY_mtype(spill_ty), MTYPE_V, wn2,
					WN_Ldid(TY_mtype(spill_ty), 0, st, spill_ty));
		  wn1  = WN_Create(OPR_TAS, TY_mtype(spill_ty), MTYPE_V, 1);
		  WN_kid0(wn1) = wn0;
		  WN_set_ty(wn1,spill_ty);
		  wn = wn1;
	
		}  else {
		  WN_set_operator (wn, OPR_LDA);
		  WN_set_desc (wn, MTYPE_V);
		}	
	      } else
	      if (WN_operator (wn) == OPR_ILOAD) {
		BOOL is_array = TRUE;
		BOOL keep_iload = FALSE;
		wn0 = WN_kid0 (wn);
		spill_ty = WN_ty(wn);
		TY_IDX struct_ty = TY_pointed(WN_load_addr_ty(wn));
		if (Type_Is_Shared_Ptr(spill_ty) || Type_Is_Shared_Ptr(WN_load_addr_ty(wn))) {
		  //spill the expr that computes the struct addr
		  // otherwise can't tell in the backend which part
		  // contributes to the ptr arith and which part
		  // contributes to the offset
		  
		  //If the type of the field is T, the type of the temp
		  // is shared [] T ;
		  
		  tmp_field_id = 0;
		  spill_ty = FLD_type(FLD_get_to_field(struct_ty, WN_field_id(wn), tmp_field_id));
		  if (!Type_Is_Shared_Ptr(spill_ty)) {
		    //this is the case where we access the field of a shared struct
		    // i.e. shared struct { ....; int a; ..}
		    if(TY_kind(spill_ty) == KIND_ARRAY) {
		      is_array = TRUE;
		      keep_iload = FALSE;
		    }
		      spill_ty = MTYPE_To_TY(MTYPE_I1);		    
		  }
		  
		  switch(TY_kind(spill_ty)) {
		  case KIND_SCALAR:
		  case KIND_STRUCT:
		    spill_ty = Make_Pointer_Type(Make_Shared_Type(spill_ty, 0));
		    break;
		    case KIND_POINTER:
		      spill_ty = Make_Shared_Type(spill_ty, 0);
		      break;
		  case KIND_ARRAY:
		    keep_iload = FALSE;
		    spill_ty = Make_Pointer_Type(Make_Shared_Type(TY_etype(spill_ty),0));
		    break;
		  default:
		    Is_True(0,("",""));
		  }

		  wn1  = WN_Create(OPR_TAS, TY_mtype(spill_ty), MTYPE_V, 1);
		  WN_kid0(wn1) = wn0;
		  WN_set_ty(wn1,spill_ty);
		  WN_kid0(wn) = wn1;
		  wn0 = wn1;
		}
		//this is the only place where Adjust_Field_Offset
		//is used in the FE. Don't know if I can get rid of it.
		wn1 = WN_Intconst (Pointer_Mtype,  WN_offset(wn));
	        wn0  = WN_Binary (OPR_ADD, Pointer_Mtype, wn0, wn1);
		if(Type_Is_Shared_Ptr(spill_ty) && spill_ty != WN_ty(wn0)) {
		  wn1  = WN_Create(OPR_TAS, TY_mtype(spill_ty), MTYPE_V, 1);
		  WN_kid0(wn1) = wn0;
		  WN_set_ty(wn1,spill_ty);
		  wn0 = wn1;
		}  //else wn = wn0;

		//have to preserve the iload here.
		if (keep_iload) {
		  WN_kid0(wn) = wn0;
		  WN_offset(wn) = 0;
		} else wn = wn0;
		//should we reset the field_id too?
		//WN_field_id(wn) = 0;
		   
		// if(is_array) {
// 		  //need to spill the ptr since we can't rely on 
// 		  // the type of the iload in the back-end. 
// 		  //over there the code gets the type from the field id.
// 		  st =  Gen_Temp_Symbol(spill_ty,
// 					Index_To_Str(Save_Str2((char*)".Mspillstruct.",
// 							       (char*)"")));
// 		  wn2 = WN_CreateBlock();
// 		  WN_INSERT_BlockLast (wn2, WN_Stid (TY_mtype(spill_ty), 0, st, spill_ty, wn));
// 		  wn = WN_CreateComma (OPR_COMMA, TY_mtype(spill_ty), MTYPE_V, wn2,
// 					WN_Ldid(TY_mtype(spill_ty), 0, st, spill_ty));	 
// 		}

	      }
	      else
	        Fail_FmtAssertion ("WFE_Expand_Expr: ADDR_EXPR has COMPONENT_REF with unhandled %s",
				   OPERATOR_name(WN_operator(wn)));
	    }

	    break;

	  default:
	    {
	      Fail_FmtAssertion ("WFE_Expand_Expr: ADDR_EXPR has unhandled %s",
				 Operator_From_Tree [code0].name);
	    }
	    break;
	}
      }
      break;

    case FUNCTION_DECL:
      {
	 st = Get_ST (exp);
	 ty_idx = ST_type (st);
	 wn = WN_Lda (Pointer_Mtype, ST_ofst(st), st);
      }
      break;

    case BIND_EXPR:
      {
        INT32    i;
        WN      *block;
        TYPE_ID  mtype;
        tree     t;

	DevWarn ("Encountered BIND_EXPR at line %d", lineno);

        for (i = wfe_bind_expr_stack_last; i >= 0; --i) {

          if (wfe_bind_expr_stack [i].rtl_expr == TREE_OPERAND (exp, 1)) {

            block = wfe_bind_expr_stack [i].block;
            t     = wfe_bind_expr_stack [i].rtl_expr;
            wfe_bind_expr_stack [i] = wfe_bind_expr_stack [wfe_bind_expr_stack_last];
            --wfe_bind_expr_stack_last;
            break;
          }
        }

        FmtAssert (i >= 0,
                   ("BIND_EXPR: did not find tree"));
	ty_idx = Get_TY (TREE_TYPE(t));
        mtype  = TY_mtype (ty_idx);
	if (mtype == MTYPE_V) {
	  WFE_Stmt_Append (block, Get_Srcpos ());
          break;
	}
	else {
          wn1 = WN_last (block);
          if (wn1 && WN_operator (wn1) == OPR_EVAL) {
            wn1 = WN_COPY_Tree (WN_kid0 (wn1));
            wn0 = block;
            WN_DELETE_FromBlock (wn0, WN_last (wn0));
	    WFE_Stmt_Append (wn0, Get_Srcpos ());
	    if (nop_ty_idx == 0 && component_ty_idx == 0) {
	      wn = wn1;
              break;
	    }
            if (WN_operator (wn1) == OPR_LDID)
              st = WN_st (wn1);
            else {
              st = Gen_Temp_Symbol (ty_idx, "__bind_expr");
              WFE_Set_ST_Addr_Saved (wn1);
              wn0 = WN_Stid (mtype, 0, st, ty_idx, wn1);
              WFE_Stmt_Append (wn0, Get_Srcpos ());
            }
          }
          else {
	    WFE_Stmt_Append (block, Get_Srcpos ());
            break;
          }
	}
      }
      /*FALLTHRU*/

    case CONSTRUCTOR:
    case PARM_DECL: // for formal parms
    case VAR_DECL:
      {
        PREG_NUM preg_num = 0;
	desc_ty_idx = component_ty_idx;
	TY_IDX hi_ty_idx = Get_TY (TREE_TYPE(exp));
	if (desc_ty_idx == 0)
	  desc_ty_idx = hi_ty_idx;

        //wei: ty_idx should just be desc_ty_idx, unless it happens to be void *,
        //in which case we see if there's a surrounding cast whose type (nop_ty_idx) we can use
        //see bug383 (missing casts between two int types with same size)
        ty_idx = desc_ty_idx;
	if (nop_ty_idx != 0 && TY_kind(ty_idx) == KIND_POINTER) {
	  TY_IDX pointed = TY_pointed(ty_idx);
	  if (TY_kind(pointed) == KIND_VOID && !TY_is_shared(pointed)) {
            ty_idx = nop_ty_idx;
          }
	  if (TY_kind(pointed) == KIND_ARRAY && !TY_is_shared(pointed)) {
	    //For LDID of arrays, use the type of the nop_ty_idx instead
	    ty_idx = nop_ty_idx;
	  }
        }

	/*

	if (! MTYPE_is_integral(TY_mtype(desc_ty_idx)))
	  ty_idx = desc_ty_idx;
	else {
	  ty_idx = nop_ty_idx;
	  if (ty_idx == 0) 
	    ty_idx = desc_ty_idx;
	  else if(TY_is_shared(desc_ty_idx) && !TY_is_shared(ty_idx))
	    ty_idx = desc_ty_idx;
	  else if (TY_is_logical(nop_ty_idx) && TY_kind(desc_ty_idx) == KIND_SCALAR) {
	    //use desc_ty_idx to avoid missing a user supplied cast (see bug383)
	    ty_idx = desc_ty_idx;
	  }
	}
	*/

	UINT cvtl_size = 0; // if non-zero, need to generate CVTL with this size
	if (! is_bit_field) {
	  if (TY_size(desc_ty_idx) > TY_size(ty_idx)) {
	    cvtl_size = TY_size(ty_idx) * 8;
	    ty_idx = desc_ty_idx;
	  }
	}
	else {
	  if (TY_size(desc_ty_idx) > TY_size(ty_idx)) 
	    ty_idx = desc_ty_idx;
	}

        TYPE_ID rtype = Widen_Mtype(TY_mtype(ty_idx));
        TYPE_ID desc = TY_mtype(desc_ty_idx);
        if (MTYPE_is_integral(desc)) {
          if (MTYPE_signed(rtype) != MTYPE_signed(desc)) {
            if (MTYPE_size_min(rtype) > MTYPE_size_min(desc) ||
		is_bit_field)
              rtype = Mtype_TransferSign(desc, rtype);
            else desc = Mtype_TransferSign(rtype, desc);
          }
        }

	if (TREE_THIS_VOLATILE(exp))
	  Set_TY_is_volatile(ty_idx);

	if (code == PARM_DECL || code == VAR_DECL) {
	  st = Get_ST (exp);
          if (ST_assigned_to_dedicated_preg (st))
	    Set_TY_is_volatile(ty_idx);
        }
	else
	if (code == CONSTRUCTOR) {
	  DevWarn ("Encountered CONSTRUCTOR at line %d", lineno);
	  st = WFE_Generate_Temp_For_Initialized_Aggregate (exp, "");
	}

	Is_True(! is_bit_field || field_id <= MAX_FIELD_ID,
		("WFE_Expand_Expr: field id for bit-field exceeds limit"));
	wn = WN_CreateLdid (OPR_LDID, rtype,
			    is_bit_field ? MTYPE_BS : desc,
			    ST_ofst(st)+component_offset+preg_num, st,
			    field_id != 0 ? hi_ty_idx : ty_idx, field_id);
	if (cvtl_size != 0) {
	  wn = WN_CreateCvtl(OPR_CVTL, rtype, MTYPE_V, cvtl_size, wn);
	}
      }
      break;

    case INTEGER_CST:
      {
	ty_idx = Get_TY (TREE_TYPE(exp));
	TYPE_ID mtyp = TY_mtype(ty_idx);
	mtyp = (mtyp == MTYPE_V) ? MTYPE_I4 : Widen_Mtype(mtyp);
	wn = WN_Intconst(mtyp, Get_Integer_Value(exp));
      }
      break;

    case REAL_CST:
      {
	TCON tcon;
	ty_idx = Get_TY (TREE_TYPE(exp));
#if defined(TARG_IA32) && !defined(REAL_ARITHMETIC)
	tcon = Host_To_Targ_Float (TY_mtype (ty_idx), TREE_REAL_CST(exp));
#else
	REAL_VALUE_TYPE real = TREE_REAL_CST(exp);
	int rval;
	int rbuf [4];
	switch (TY_mtype (ty_idx)) {
	case MTYPE_F4:
	  REAL_VALUE_TO_TARGET_SINGLE (real, rval);
	  //tcon = Host_To_Targ_Float_4 (MTYPE_F4, *(float *) &rval);
	  tcon = Host_To_Targ_Float_4 (MTYPE_F4, (float) real);
	  break;
	case MTYPE_F8:
	case MTYPE_FQ:
	  REAL_VALUE_TO_TARGET_DOUBLE (real, rbuf);
	  //tcon = Host_To_Targ_Float (MTYPE_F8, *(double *) &rbuf);
	  tcon = Host_To_Targ_Float (MTYPE_F8, real);
	  break;
	default:
	  FmtAssert(FALSE, ("WFE_Expand_Expr unexpected float size"));
	  break;
	}

#endif
	st = New_Const_Sym (Enter_tcon (tcon), ty_idx);
	wn = WN_CreateConst (OPR_CONST, TY_mtype (ty_idx), MTYPE_V, st);
      }
      break;

    case COMPLEX_CST:
      {
	TCON tcon;
	ty_idx = Get_TY (TREE_TYPE(exp));
#if defined(TARG_IA32) && !defined(REAL_ARITHMETIC)
        tcon = Host_To_Targ_Complex (TY_mtype (ty_idx),
				     TREE_REAL_CST(TREE_REALPART(exp)),
				     TREE_REAL_CST(TREE_IMAGPART(exp)));
#else
	REAL_VALUE_TYPE real = TREE_REAL_CST(TREE_REALPART(exp));
	REAL_VALUE_TYPE imag = TREE_REAL_CST(TREE_IMAGPART(exp));
        int rval;
	int ival;
	int rbuf [4];
	int ibuf [4];
	switch (TY_mtype (ty_idx)) {
	  case MTYPE_C4:
	    REAL_VALUE_TO_TARGET_SINGLE (real, rval);
	    REAL_VALUE_TO_TARGET_SINGLE (imag, ival);
	    tcon = Host_To_Targ_Complex_4 (MTYPE_C4,
					   *(float *) &rval,
					   *(float *) &ival);
	    break;
	  case MTYPE_C8:
	    REAL_VALUE_TO_TARGET_DOUBLE (real, rbuf);
	    REAL_VALUE_TO_TARGET_DOUBLE (imag, ibuf);
	    tcon = Host_To_Targ_Complex (MTYPE_C8,
					 *(double *) &rbuf,
					 *(double *) &ibuf);
	    break;
	  default:
	    FmtAssert(FALSE, ("WFE_Expand_Expr unexpected float size"));
	    break;
	}
#endif
	st = New_Const_Sym (Enter_tcon (tcon), ty_idx);
	wn = WN_CreateConst (OPR_CONST, TY_mtype (ty_idx), MTYPE_V, st);
      }
      break;

    // this should occur only if string is a statement expression
    case STRING_CST:
      {
	TCON tcon;
	tcon = Host_To_Targ_String (MTYPE_STRING,
				    TREE_STRING_POINTER(exp),
				    TREE_STRING_LENGTH(exp));
	ty_idx = Get_TY(TREE_TYPE(exp));
	st = New_Const_Sym (Enter_tcon (tcon), ty_idx);
	wn = WN_Lda (Pointer_Mtype, ST_ofst(st), st);
	TREE_STRING_ST (exp) = st;
      }
      break;

    // unary ops
    case BIT_NOT_EXPR:
    case ABS_EXPR:
    case NEGATE_EXPR:
    case TRUTH_NOT_EXPR:
    case REALPART_EXPR:
    case IMAGPART_EXPR:
      {
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
        wn  = WN_Unary (Operator_From_Tree [code].opr,
                        Widen_Mtype(TY_mtype(Get_TY(TREE_TYPE(exp)))), wn0);
      }
      break;

    case CONJ_EXPR:
      {
	ty_idx = Get_TY (TREE_TYPE(exp));
        TYPE_ID complex_mtype = TY_mtype(ty_idx);
        TYPE_ID float_mtype   = Mtype_complex_to_real (complex_mtype);
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
	if (WN_has_side_effects (wn0)) {
	  ST       *preg_st;
	  PREG_NUM  preg;
	  preg_st = MTYPE_To_PREG(complex_mtype);
	  preg    = Create_Preg (complex_mtype, NULL);
	  wn0     = WN_Stid (complex_mtype, preg, preg_st, ty_idx, wn0);
	  WFE_Stmt_Append (wn0, Get_Srcpos());
	  wn0 = WN_Ldid (complex_mtype, preg, preg_st, ty_idx);
	}
        wn = WN_Binary (OPR_COMPLEX, complex_mtype,
			WN_Unary (OPR_REALPART, float_mtype, wn0),
			WN_Unary (OPR_NEG, float_mtype,
				  WN_Unary (OPR_REALPART, float_mtype, wn0)));
      }
      break;

    case NOP_EXPR:
      {
	ty_idx = Get_TY (TREE_TYPE(exp));
        TYPE_ID mtyp = TY_mtype(ty_idx);

	TY_IDX l_idx, r_idx;
	r_idx = Get_TY(TREE_TYPE(TREE_OPERAND(exp,0)));
	if(TY_kind(ty_idx) == KIND_POINTER && TY_kind(r_idx) == KIND_POINTER &&
	   (Type_Is_Shared_Ptr(ty_idx) || Type_Is_Shared_Ptr(r_idx))) {
	  ty_idx = 0;
	}

	if(TY_kind(ty_idx) == KIND_SCALAR && Type_Is_Shared_Ptr(ty_idx) &&
	   !Type_Is_Shared_Ptr(r_idx))
	  ty_idx = 0;

	tree lt, rt;
	int blockl = -1, blockr = -1 ;

	// do not pass struct type down because will cause rtype of MTYPE_M
        wn = WFE_Expand_Expr (TREE_OPERAND (exp, 0), TRUE, 
			      (mtyp == MTYPE_M) ? 0 : ty_idx);

	if (mtyp == MTYPE_M || compiling_upc) {
          lt = TREE_TYPE(exp);
          rt = TREE_TYPE(TREE_OPERAND(exp,0));
          if(compiling_upc &&
             TREE_CODE(lt) == POINTER_TYPE &&
             TREE_CODE(rt) == POINTER_TYPE ) {
            TY_IDX lti,rti;
            lti = Get_TY(lt);
            rti = Get_TY(rt);
	    
            if((TREE_SHARED(TREE_TYPE(lt)) && TREE_SHARED(TREE_TYPE(rt))) ||
               TREE_SHARED(TREE_TYPE(rt)) ) {
	      lt = TREE_TYPE(lt);
              rt = TREE_TYPE(rt);
              lti = Get_TY(lt);
              rti = Get_TY(rt);
              INTRINSIC iop;
	      
	      //it's hard to tell here whether the TAS is needed or changes behavior
	      // it's needed for ptr arithmetic   : localp = globalp+offset
	      //it's not needed for struct accesses : localp = globals->ptr;
	      //The only way differentiate is by looking at wn and see if it
	      //is an iload with a field_id or if it is a TAS with the correct type
	     
	      TY_IDX temp = TY_pointed(WN_Get_Ref_TY(wn));
	      if((WN_operator(wn) != OPR_ILOAD && WN_operator(wn) != OPR_TAS) 
		 || (WN_operator(wn) == OPR_ILOAD && WN_field_id(wn) == 0) 
		 || /* this is for user level cast */
		 (TY_is_shared(lti) != TY_is_shared(TY_pointed(WN_ty(wn))))
		 || (TY_is_shared(lti) == TY_is_shared(temp) &&
		     TY_size(lti) != TY_size(temp))
		 || (TY_is_shared(lti) == TY_is_shared(temp) &&
		     TY_block_size(lti) != TY_block_size(temp))
		 )
		  {
		    
		    WN *wn0  = WN_Create(OPR_TAS, mtyp, MTYPE_V, 1);
		    WN_kid0(wn0) = wn;
		    wn = wn0;
		    WN_set_ty(wn, Get_TY(TREE_TYPE(exp)));
		  }

              goto sc;
            }  else if(TY_mtype(TY_pointed(lti)) != TY_mtype(TY_pointed(rti)) &&
                       TY_kind(TY_pointed(rti)) == KIND_VOID) {
              //user level cast - need to spill the intermediate result since
              // inserting a TAS will be lost in the back-end
              TY_IDX ltit;
              if(TY_is_shared(lti))
                ltit = Shared_To_Private_Type(lti);
              else
                ltit = lti;
              ST *temp = Gen_Temp_Symbol(ltit, "_casttmp");
              WN *wn0 = WN_Create(OPR_TAS, Pointer_Mtype, MTYPE_V,1);
              WN_kid0(wn0) = wn;
              WN_set_ty(wn0, ltit);
              wn0 = WN_Stid(Pointer_Mtype, 0, temp,  ltit, wn0);
              WFE_Stmt_Append (wn0, Get_Srcpos ());
              wn = WN_Ldid(Pointer_Mtype, 0, temp, ltit);
	    } else if((lti & 0xff) != (rti & 0xff) || 
		      (((TY_pointed(lti)&0xe0) != (TY_pointed(rti)&0xe0)) &&
		       !TY_is_const(TY_pointed(lti)) || TY_is_const(TY_pointed(rti))) ) {
	      //bug670. bug1804 - we should not generate TAS for T*  to const T*
	      
	      WN *wn0  = WN_Create(OPR_TAS, mtyp, MTYPE_V, 1);
	      WN_kid0(wn0) = wn;
	      wn = wn0;
	      WN_set_ty(wn, Get_TY(TREE_TYPE(exp)));
	    }
	  } else 
	    goto cont;
	sc:	  
	  break;
	cont:
	 
	  if(mtyp == MTYPE_M)
	    break;
	}

	if (MTYPE_is_integral(mtyp) && MTYPE_is_integral(WN_rtype(wn))) {
	  if (MTYPE_size_min(mtyp) < MTYPE_size_min(WN_rtype(wn))) {
	    if (MTYPE_size_min(mtyp) != 32)
	      wn = WN_CreateCvtl(OPR_CVTL, Widen_Mtype(mtyp), MTYPE_V,
			         MTYPE_size_min(mtyp), wn);
	    else {
	      /* save a cast if we can*/
	      if (compiling_upc && WN_operator(wn) == OPR_CVTL && 
		  MTYPE_bit_size(mtyp) == WN_cvtl_bits(wn) &&
		  ((MTYPE_is_signed(WN_rtype(wn)) && MTYPE_is_signed(mtyp)) ||
		   (MTYPE_is_unsigned(WN_rtype(wn)) && MTYPE_is_unsigned(mtyp)))) {
		/* handle a cast to integral type explicitly thru TAS instead of CVT,
		   to preserve their type names (e.g., long instead of int32_t) (see bug383) */
		if (compiling_upc && TY_is_logical(ty_idx)) {
		  wn = WN_Tas(mtyp, ty_idx, wn);
		}
	      } else {
		wn = WN_Cvt(WN_rtype(wn), mtyp, wn);
	      }
	    }
	  }
	  else {
	    TY_IDX ty_idx0 = Get_TY(TREE_TYPE(TREE_OPERAND (exp, 0)));
	    TYPE_ID mtyp0 = TY_mtype(ty_idx0);

	    if (MTYPE_size_min(mtyp) > MTYPE_size_min(mtyp0) &&
		! Has_Subsumed_Cvtl(WN_operator(wn)))
	      wn = WN_CreateCvtl(OPR_CVTL, Widen_Mtype(mtyp0), MTYPE_V,
				 MTYPE_size_min(mtyp0), wn);

	    if (compiling_upc) {
	      //only generate downcasts and casts between signed/unsigned,
	      //to avoid unnecessary CVTs hurting the quality of the generated code
	      if (MTYPE_size_min(WN_rtype(wn)) > MTYPE_size_min(mtyp) ||
		  (MTYPE_signed(WN_rtype(wn)) ^ MTYPE_signed(mtyp))) {
		wn = WN_Cvt(WN_rtype(wn), mtyp, wn);
	      }
	    } else if (MTYPE_size_min(mtyp) > MTYPE_size_min(WN_rtype(wn))) {
	      wn = WN_Cvt(WN_rtype(wn), mtyp, wn);
	    } else { // same size
	      if (mtyp != WN_rtype(wn)) 
		wn = WN_Cvt(WN_rtype(wn), mtyp, wn);
	    }
	  }
	}
	else {
	  if (mtyp != WN_rtype(wn)) 
	    wn = WN_Cvt(WN_rtype(wn), mtyp, wn);
	}
      }
      break;

    case COMPONENT_REF:
      {
	INT64 ofst;
	arg0 = TREE_OPERAND (exp, 0);
	arg1 = TREE_OPERAND (exp, 1);
	ty_idx = Get_TY (TREE_TYPE(exp));
	TY_IDX struct_ty = Get_TY(TREE_TYPE(arg0));


	TY_IDX  shared_field_idx = 0;
	if (component_ty_idx == 0) {
	  //fields of a shared struct do not get mark as shared, so we need to check if 
	  //the enclosing struct is shared
	 
	  if (!TY_is_shared(ty_idx) && TY_is_shared(struct_ty)) {
	    ty_idx = Make_Shared_Type(ty_idx,  0, 
 				      TY_is_strict(struct_ty) ? STRICT_CONSISTENCY : NO_CONSISTENCY);

	  } else if(!TY_is_shared(struct_ty) && Type_Is_Shared_Ptr(ty_idx, TRUE)) {
	    //field of struct is shared ptr
	    shared_field_idx = ty_idx;
	  }
	}  else { 
	  
	  if(TY_is_shared(struct_ty) && !TY_is_shared(component_ty_idx)) {
	    ty_idx = component_ty_idx;
	  } else 
	    ty_idx = component_ty_idx;
	}
	if (DECL_BIT_FIELD(arg1)) 
	  is_bit_field = TRUE;

	if (! is_bit_field && 
	    component_ty_idx == 0) {  // only for top-level COMPONENT_REF
          // if size does not agree with ty_idx, fix ty_idx
          tree sizenode = DECL_SIZE(arg1);
          if (TREE_CODE(sizenode) == INTEGER_CST) {
	    TYPE_ID c_mtyp = TY_mtype(ty_idx);
	    UINT64 bsize = Get_Integer_Value(sizenode);
	    if (MTYPE_size_min(c_mtyp) > bsize) {
	      FmtAssert(MTYPE_is_integral(c_mtyp), 
	        ("COMPONENT_REF: integer type expected at inconsistent field size"));
	      c_mtyp = Mtype_AlignmentClass(bsize >> 3, MTYPE_type_class(c_mtyp));
	      ty_idx = MTYPE_To_TY(c_mtyp);
	    }
	  }
        }

	if (! is_bit_field) {
	 
	    ofst = (BITSPERBYTE * Get_Integer_Value(DECL_FIELD_OFFSET(arg1)) +
		    Get_Integer_Value(DECL_FIELD_BIT_OFFSET(arg1)))
	      / BITSPERBYTE;

	}
	else ofst = 0;
        wn = WFE_Expand_Expr (arg0, TRUE, nop_ty_idx, ty_idx, ofst+component_offset,
			      field_id + DECL_FIELD_ID(arg1), is_bit_field);
	
	if(shared_field_idx) {
	  wn1  = WN_Create(OPR_TAS, TY_mtype(shared_field_idx), MTYPE_V, 1);
	  WN_kid0(wn1) = wn;
	  WN_set_ty(wn1,shared_field_idx);
	  wn = wn1;
	}
	
      }
      break;

    case INDIRECT_REF:
      {
	UINT xtra_BE_ofst = 0; 	// only needed for big-endian target
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));

	TY_IDX hi_ty_idx = Get_TY(TREE_TYPE(exp));
       
	//patch things up for struct accesses -
	//whirl has correct types (see comment for OPR_ADD)
	TY_IDX patch_idx;
	if(WN_operator(wn0) == OPR_TAS) {
	  patch_idx =  WN_Get_Ref_TY(wn0);
	  if(TY_kind(patch_idx) == KIND_POINTER )
	    hi_ty_idx = TY_pointed(patch_idx);
	}
	
	
	desc_ty_idx = component_ty_idx;
	if (desc_ty_idx == 0)
	  desc_ty_idx = hi_ty_idx;

	
        if (! MTYPE_is_integral(TY_mtype(desc_ty_idx)) || TY_is_shared(desc_ty_idx)) {
	  /* for shared variable, also don't perform integer promotion, since
	     This means we'll load the wrong size for char and short */
	  ty_idx = desc_ty_idx;
	} else {
	  ty_idx = nop_ty_idx;
	  if (ty_idx == 0) 
	    ty_idx = desc_ty_idx;
	}


	if (! is_bit_field) {
	  if (TY_size(desc_ty_idx) > TY_size(ty_idx)) {
	    //bug1477: I have no idea what it is supposed to do,
	    //but it's causing trouble for bug1477 with a wrong offset
	    if (!compiling_upc && Target_Byte_Sex == BIG_ENDIAN)
	      xtra_BE_ofst = TY_size(desc_ty_idx) - TY_size(ty_idx);
	    desc_ty_idx = ty_idx;
	  }
	}
	else {
	  if (TY_size(desc_ty_idx) > TY_size(ty_idx)) 
	    ty_idx = desc_ty_idx;
	}

	TYPE_ID rtype = Widen_Mtype(TY_mtype(ty_idx));
	TYPE_ID desc = TY_mtype(desc_ty_idx);
	if (MTYPE_is_integral(desc)) {
	  if (MTYPE_signed(rtype) != MTYPE_signed(desc)) {
	    if (MTYPE_size_min(rtype) > MTYPE_size_min(desc) ||
		is_bit_field)
	      rtype = Mtype_TransferSign(desc, rtype);
	    else desc = Mtype_TransferSign(rtype, desc);
	  }
	}

	if (TREE_THIS_VOLATILE(exp))
	  Set_TY_is_volatile(hi_ty_idx);

	Is_True(! is_bit_field || field_id <= MAX_FIELD_ID,
		("WFE_Expand_Expr: field id for bit-field exceeds limit"));

        if (!WFE_Keep_Zero_Length_Structs &&
            rtype == MTYPE_M              &&
            TY_size (hi_ty_idx) == 0) {
	  if (WN_has_side_effects (wn0)) {
	    wn = WN_CreateEval (wn0);
	    WFE_Stmt_Append (wn, Get_Srcpos());
	  }
	  wn = NULL;
        }
        else {
	  // special case indexing into a constant string
	  if (WN_operator (wn0) == OPR_LDA          &&
	      ST_class (WN_st (wn0)) == CLASS_CONST &&
	      is_bit_field == FALSE                 &&
	      field_id == 0) {
            st = WN_st (wn0);
	    TCON tcon = Tcon_Table [ST_tcon (st)];
	    if (TCON_ty (tcon) == MTYPE_STRING &&
                TY_size (Be_Type_Tbl (desc)) == 1) {
	      mUINT32 len = Targ_String_Length (tcon);
	      mUINT64 offset = component_offset + xtra_BE_ofst + WN_offset (wn0);
	      if (offset <= len    &&
		  desc == MTYPE_U1 &&
		  (rtype == MTYPE_U4 || rtype == MTYPE_U8)) {
		unsigned char *cp = (unsigned char *) Targ_String_Address (tcon);
		unsigned long long val = cp [offset];
		wn = WN_Intconst (rtype, val);
		break;
	      }
	      else
	      if (offset <= len    &&
		  desc == MTYPE_I1 &&
		  (rtype == MTYPE_I4 || rtype == MTYPE_I8)) {
		signed char *cp = (signed char *) Targ_String_Address (tcon);
		signed long long val = cp [offset];
		wn = WN_Intconst (rtype, val);
		break;
	      }
	    }
	  }

	  TY_IDX access_ty = WN_ty(wn0);
	  
	  if(Type_Is_Shared_Ptr(access_ty) &&  
	     !(TY_kind(access_ty) == KIND_POINTER && TY_is_shared(access_ty) && 
	       !TY_is_shared(TY_pointed(access_ty)))) {
	   
	    hi_ty_idx = Make_Shared_Type(hi_ty_idx,
 					 Get_Type_Block_Size(access_ty), 
 					 TY_is_strict(access_ty) ? STRICT_CONSISTENCY : NO_CONSISTENCY);
	    Is_True(TY_kind(access_ty) == KIND_POINTER ||
		    TY_kind(access_ty) == KIND_STRUCT, ("",""));
	   
	    if(!Type_Is_Shared_Ptr(ty_idx) || 
	       Get_Type_Block_Size(access_ty) != Get_Type_Block_Size(ty_idx)) {
	      
	      // for a.s and a.s[i] we get here with
	      // access_ty = shared [] char
	      // ty_idx = typeof(a)
	      //for a[i].s[j] or a[i].s
	      // we get there with access_ty = ptr to typeof(a)
	      // and ty_idx = typeof(a)
	      // Is_True(Get_Type_Block_Size(access_ty) == 0 && 
// 		      TY_mtype(TY_pointed(access_ty)) == MTYPE_I1, ("",""));
	      
	      if((TY_kind(ty_idx) == KIND_POINTER || TY_kind(ty_idx) == KIND_ARRAY) &&
		 Get_Type_Block_Size(access_ty) == 0 && 
		 TY_mtype(TY_pointed(access_ty)) == MTYPE_I1)
		ty_idx = TY_pointed(access_ty);
	      
	      if(TY_kind(ty_idx) == KIND_ARRAY) {
		//don't pass array type down since 
		//the lowering code takes the size of the type 
	
		  ty_idx = Make_Pointer_Type(TY_etype(ty_idx));
	      }
	    }
	    
	  }
	  
	  wn = WN_CreateIload(OPR_ILOAD, rtype,
			      is_bit_field ? MTYPE_BS : desc, 
			      component_offset+xtra_BE_ofst, 
			      ty_idx,
			      Make_Pointer_Type (hi_ty_idx, FALSE),
			      wn0, field_id);
	}
      }
      break;

    case CONVERT_EXPR:
    case FLOAT_EXPR:
      {
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
	ty_idx = Get_TY (TREE_TYPE(exp));
	TY_IDX kid_idx = 0;
	TYPE_ID mtyp = TY_mtype(ty_idx);
	if (mtyp == MTYPE_V)
	  wn = wn0;
	else {
	  mtyp = Widen_Mtype(TY_mtype(ty_idx));
	  if (mtyp == WN_rtype(wn0) || mtyp == MTYPE_V)
	    wn = wn0;
	  else {
	    if(OPERATOR_is_load(WN_operator(wn0)) || WN_operator(wn0) == OPR_TAS || 
	       WN_operator(wn0) == OPR_LDA) {
	      switch(WN_operator(wn0)) {
	      case OPR_LDID:
	      case OPR_TAS:
		if(Type_Is_Shared_Ptr(WN_ty(wn0), TRUE)) {
		  kid_idx = WN_ty(wn0);
		}
		break;
	      case OPR_ILOAD:
		if(Type_Is_Shared_Ptr(WN_ty(wn0), TRUE)) {
		  kid_idx = WN_ty(wn0);
		}
		break;
	      case OPR_MLOAD:
		break;
	      case OPR_LDA:
		if(Type_Is_Shared_Ptr(WN_ty(wn0), TRUE)) {
		  kid_idx = WN_ty(wn0);
		}
		break;
	      } 
	    }
	    if(kid_idx == 0) {
	      //wei: Do not use CVT for pointer arithmetic, since it could confuse the 
	      //backend and resulting in everything cast to char *
	      if (TY_kind(ty_idx) == KIND_POINTER && MTYPE_is_integral(WN_rtype(wn0))) {
		wn = wn0;
	      } else {
		wn = WN_Cvt(WN_rtype(wn0), mtyp, wn0);
	      }
	    } else {
	      wn  = WN_Create(OPR_TAS, TY_mtype(kid_idx), MTYPE_V, 1);
	      WN_kid0(wn) = wn0;
	      WN_set_ty(wn,kid_idx);
	    }
	    // The following opcodes are not valid for MIPS
	    if (WN_opcode(wn) == OPC_I4U4CVT ||
	        WN_opcode(wn) == OPC_U4I4CVT ||
	        WN_opcode(wn) == OPC_I8U8CVT ||
	        WN_opcode(wn) == OPC_U8I8CVT) {
	      wn = WN_kid0 (wn);
	    }
	  }
	}
      }
      break;

    case FIX_TRUNC_EXPR:
      {
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
	ty_idx = Get_TY (TREE_TYPE(exp));
	TYPE_ID mtyp = Widen_Mtype(TY_mtype(ty_idx));
	wn = WN_Trunc(WN_rtype(wn0), mtyp, wn0);
      }
      break;
      
    case MINUS_EXPR: {
      //special case for subtraction of shared pointers
      // the tree is always minus ( nop(convert( arg1) , nop(convert(arg1)) )
      tree arg1 = TREE_OPERAND(exp, 1);
      tree arg0 = TREE_OPERAND(exp, 0);
      if (TREE_CODE(arg1) == NOP_EXPR && TREE_CODE(arg0) == NOP_EXPR) {
	arg0 = TREE_OPERAND(arg0, 0);
	arg1 = TREE_OPERAND(arg1,0);
	if (TREE_CODE(arg1) == CONVERT_EXPR && TREE_CODE(arg0) == CONVERT_EXPR) {
	  arg0 = TREE_OPERAND(arg0,0);
	  arg1 = TREE_OPERAND(arg1,0);
	  if (Type_Is_Shared_Ptr(Get_TY(TREE_TYPE(arg0)), TRUE) && 
	      Type_Is_Shared_Ptr(Get_TY( TREE_TYPE(arg1)), TRUE) ) {
	    // shared ptr - shared ptr
	    TYPE_ID etype = TY_mtype(Get_TY(TREE_TYPE(exp)));
	    WN *wn0 = WFE_Expand_Expr (arg0);
	    WN *wn1 = WFE_Expand_Expr (arg1);
	    return WN_Binary (Operator_From_Tree [code].opr, Widen_Mtype(etype), wn0, wn1);
	    
	  }
	}
      }
    }  // fall thru for regular subtractions

    // binary ops
    case PLUS_EXPR:
    case MULT_EXPR:
    case MAX_EXPR:
    case MIN_EXPR:
    case LSHIFT_EXPR:
    case BIT_AND_EXPR:
    case BIT_IOR_EXPR:
    case BIT_XOR_EXPR:
    case TRUNC_DIV_EXPR:
    case TRUNC_MOD_EXPR:
    case RDIV_EXPR:
    case EXACT_DIV_EXPR:
    case TRUTH_AND_EXPR:
    case TRUTH_OR_EXPR:
    case TRUTH_XOR_EXPR:
    case COMPLEX_EXPR:
    case CEIL_DIV_EXPR:
      {

	TYPE_ID etype = TY_mtype(Get_TY(TREE_TYPE(exp)));
	tree arg1 = TREE_OPERAND(exp, 1);
	tree arg0 = TREE_OPERAND(exp, 0);
	/*  for p+i+j where p is local struct*, the gcc part
	    creates the ptr arithmetic trees with the original 
	    struct size. Need to go through all the MULT(disp, size)
	    expressions and adjust the size accordingly - NOT SURE what
	    will this do the the optimizer */
	  
	
	if(TREE_CODE(TREE_TYPE(arg0)) == POINTER_TYPE && 
	   (TREE_CODE(TREE_TYPE(TREE_TYPE(arg0))) == RECORD_TYPE ||
	    TREE_CODE(TREE_TYPE(TREE_TYPE(arg0))) == UNION_TYPE) ) {
	  TY_IDX pt_idx = Get_TY(TREE_TYPE(arg0));
	  if(!Type_Is_Shared_Ptr(pt_idx, TRUE)) {
	    if(TREE_CODE(arg1) == MULT_EXPR) {
	      arg1 = TREE_OPERAND(arg1, 1); 
	      if(TREE_CODE(arg1) == INTEGER_CST) {
		int size = Get_Integer_Value(arg1);
		int asize, fsize;
		pt_idx = TY_pointed(pt_idx);
		asize = TY_adjusted_size(pt_idx);
		fsize = Get_Integer_Value(TYPE_SIZE(TREE_TYPE(TREE_TYPE(arg0))))/8;
		asize = (size/fsize)*asize;
		
		TREE_INT_CST_LOW (arg1) = asize;
		TREE_INT_CST_HIGH (arg1) = 0;

	      }
	    }
	  }
	}
	
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
        wn1 = WFE_Expand_Expr (TREE_OPERAND (exp, 1));

	BOOL fold_lda;
	BOOL shared_args = WN_Type_Is_Shared_Ptr(wn0) || WN_Type_Is_Shared_Ptr(wn1);
	
	if(WN_Type_Is_Shared_Ptr(wn0) && TY_kind(WN_ty(wn0)) == KIND_POINTER && 
	   TY_kind(TY_pointed(WN_ty(wn0))) == KIND_VOID) {
	   error_with_file_and_line(input_filename, lineno, "Pointer arithmetic on shared void*");
	} 
	
	if(WN_Type_Is_Shared_Ptr(wn1) && TY_kind(WN_ty(wn1)) == KIND_POINTER && 
	   TY_kind(TY_pointed(WN_ty(wn1))) == KIND_VOID) {
	   error_with_file_and_line(input_filename, lineno, "Pointer arithmetic on shared void*");
	} 
	
	if(shared_args) {
	  if(WN_operator(wn0) == OPR_LDID && 
	     WN_operator(wn1) == OPR_INTCONST &&
	     TREE_CODE(TREE_OPERAND(exp,0)) == NOP_EXPR) {
	    fold_lda =  WN_Simp_Fold_LDA;
	    WN_Simp_Fold_LDA = FALSE;
	  } else 
	    fold_lda = WN_Simp_Fold_LDA;
	}
        wn  = WN_Binary (Operator_From_Tree [code].opr, Widen_Mtype(etype), wn0, wn1);

	//&s->a[i] gets expanded into add(add(ldid,off),mpy(i,sz)) 
	//spill the inner add  to make sure that const folding does not move the offset
	// away from the base pointer
	if((WN_operator(wn0) == OPR_LDID ||
	    (WN_operator(wn0) == OPR_TAS && WN_operator(WN_kid0(wn0)) == OPR_LDID)
	    ) && 
	   WN_operator(wn1) == OPR_INTCONST &&
	   TREE_CODE(TREE_OPERAND(exp,0)) == NOP_EXPR) 
	  {
	    TY_IDX op_ty = WN_ty(wn0);
	    if(TY_kind(op_ty) == KIND_POINTER && !Type_Is_Shared_Ptr(op_ty,1) && 
	       TY_kind(TY_pointed(op_ty)) == KIND_STRUCT &&
	       TY_size(TY_pointed(op_ty)) > WN_const_val(wn1)) {
	
	      wn1  = WN_Create(OPR_TAS, Pointer_Mtype, MTYPE_V, 1);
	      WN_kid0(wn1) = wn0;
	      WN_set_ty(wn1, Make_Pointer_Type(MTYPE_To_TY(MTYPE_I1)));
	      WN_kid0(wn) = wn1;

	      TY_IDX spill_ty = Get_TY(TREE_TYPE(TREE_OPERAND(exp,0)));
	      ST *st =  Gen_Temp_Symbol(spill_ty,
					Index_To_Str(Save_Str2((char*)".Mspillstruct.",
							       (char*)"")));
	      wn2 = WN_CreateBlock();
	      wn1 = wn;
	      WN_INSERT_BlockLast (wn2, WN_Stid (TY_mtype(spill_ty), 0, st, spill_ty, wn1));
	      wn = WN_CreateComma (OPR_COMMA, TY_mtype(spill_ty), MTYPE_V, wn2,
				    WN_Ldid(TY_mtype(spill_ty), 0, st, spill_ty));
	    }
	    
	  }
	
	
	
	if (shared_args)
	  WN_Simp_Fold_LDA =  fold_lda;



	if ((MTYPE_is_integral(etype)) &&
	    (Widen_Mtype(etype) != etype) &&
	    (TY_size (Get_TY(TREE_TYPE(exp))) < 32) &&
	    (code == PLUS_EXPR || code == MINUS_EXPR || code == MULT_EXPR))
	  wn = WN_CreateCvtl(OPR_CVTL, Widen_Mtype(etype), MTYPE_V,
			     TY_size (Get_TY(TREE_TYPE(exp))) * 8, wn);
	if(compiling_upc &&
	   (TYPE_SHARED(TREE_TYPE(exp)) ||
	    (POINTER_TYPE_P(TREE_TYPE(exp)) && TYPE_SHARED(TREE_TYPE(TREE_TYPE(exp)))) ) )  {
	  
	  if(TYPE_SHARED(TREE_TYPE(exp))) {
	    if (TREE_CODE(TREE_TYPE(exp)) == POINTER_TYPE) {
	      // create Tas
	      tree el_type = TREE_TYPE(TREE_TYPE(exp));
	      if (TYPE_SHARED(el_type)) {
		wn0  = WN_Create(OPR_TAS, Pointer_Mtype, MTYPE_V, 1);
		WN_kid0(wn0) = wn;
		wn = wn0;
		TY_IDX el_idx = Make_Shared_Type(TYPE_TY_IDX(el_type), 
						 TYPE_BLOCK_SIZE(el_type) ? 
						 Get_Integer_Value(TYPE_BLOCK_SIZE(el_type)) : 1,
						 TYPE_STRICT(el_type) ? STRICT_CONSISTENCY : NO_CONSISTENCY);
		WN_set_ty(wn, Make_Pointer_Type(el_idx, FALSE));
	      }
	    }
	  } else {
	    
	    // mark the arithmetic node with the type of 
	    // the pointer - need to know the exact type in
	    //order to decide what operation it expands to in the
	    // the back-end.
	    
	    // Exception to the rule - for struct field access,
	    // the gnu part has the type of the field marked as shared
	    // while the whirl code contains the correct types.
	    //It's ambiguous as hell, and the only way to tell 
	    //is to check kid0/kid1 for field_ids and change the
	    //cast. Even if the type is ptr to local the cast
	    // is needed to patch things further up in indirect_ref
	   
	    //create TAS
	    wn0  = WN_Create(OPR_TAS, Pointer_Mtype, MTYPE_V, 1);
	    WN_kid0(wn0) = wn;
	   
	    tree el_type = TREE_TYPE(TREE_TYPE(exp));
	    TY_IDX el_idx = Get_TY(el_type);
	    int bsize;
	    CONSISTENCY_class  consistency;
	    
	    
	    WN *kid;
	    TY_IDX kid_idx;

	    for(int i = 0; i < WN_kid_count(wn); i++) { 
	      kid = WN_kid(wn,i);
	      kid_idx = (kid && WN_Has_Valid_Type(kid)) ? WN_Get_Ref_TY(kid) : 0;
	      if(kid_idx && WN_operator(kid) != OPR_TAS &&  WN_field_id(kid) && 
		 !Type_Is_Shared_Ptr(kid_idx) && TY_kind(kid_idx) != KIND_SCALAR ) {
		if(TY_kind(kid_idx) == KIND_ARRAY)
		  kid_idx = Make_Pointer_Type(TY_etype(kid_idx));
		WN_set_ty(wn0, kid_idx);
		return wn0;
		
	      }
	    }
	    
	    
	    
	     wn = wn0;

	    if(!TY_is_shared(el_idx)) {
	      //this happens only for the last structure field access
	      //in a chain of field accesses 
	      // need to fake here the block size for the inserted type
	      // cast, otherwise the backend can't retrieve that information
	      TY_IDX ptr_ty = WN_ty(WN_kid0(wn0));
	      bsize = 0;
	      if(!ptr_ty) 
		if(WN_operator(WN_kid0(wn0)) == OPR_ADD) {
		  ptr_ty = WN_ty(WN_kid0(WN_kid0(wn0)));
		  if(!ptr_ty || !Type_Is_Shared_Ptr(ptr_ty, TRUE)) {
		    wn0 = WN_kid1(WN_kid0(wn0));
		    ptr_ty = WN_ty(wn0);
		  }
		}
	      
	      if(ptr_ty && Type_Is_Shared_Ptr(ptr_ty)) {
		el_idx = TY_pointed(ptr_ty);
		bsize = Get_Type_Block_Size(ptr_ty);
		consistency = TY_is_strict(ptr_ty) ? STRICT_CONSISTENCY : NO_CONSISTENCY;
	      }
	    } else {
	      bsize =  Get_Type_Block_Size(el_idx);
	      consistency = (TY_is_strict(el_idx) ? STRICT_CONSISTENCY : NO_CONSISTENCY);
	    }
	    
	    
	    el_idx = Make_Shared_Type(el_idx, bsize, consistency);
	    WN_set_ty(wn, Make_Pointer_Type(el_idx, FALSE));
	  }	 
	}
      }
      //For vectorization we need the ARA analysis in Walk_Loop_Dependence
      //That code always expects  the base ptr in kid0
      if(WN_operator(wn) == OPR_ADD && WN_operator(WN_kid1(wn)) == OPR_LDID &&
	 (TY_kind(WN_ty(WN_kid1(wn))) == KIND_POINTER || TY_kind(WN_ty(WN_kid1(wn))) == KIND_ARRAY )) {
	WN *temp = WN_kid1(wn);
	WN_kid1(wn) = WN_kid0(wn);
	WN_kid0(wn) = temp;
	
      }
      break;

    case LROTATE_EXPR:
      {
	ty_idx = Get_TY(TREE_TYPE(exp));
	TYPE_ID mtype = TY_mtype (ty_idx);
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
        wn1 = WFE_Expand_Expr (TREE_OPERAND (exp, 1));
	wn1 = WN_Binary (OPR_SUB, Widen_Mtype (mtype),
			 WN_Intconst (Widen_Mtype (mtype),
				      TY_size (ty_idx) * 8),
			 wn1);
	wn  = WN_Rrotate (TY_mtype(Get_TY(TREE_TYPE(exp))), wn0, wn1);
      }
      break;

    case RROTATE_EXPR:
      {
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
        wn1 = WFE_Expand_Expr (TREE_OPERAND (exp, 1));
	wn  = WN_Rrotate (TY_mtype(Get_TY(TREE_TYPE(exp))), wn0, wn1);
      }
      break;

    case RSHIFT_EXPR:
      {
	TYPE_ID mtyp = Widen_Mtype(TY_mtype(Get_TY(TREE_TYPE(exp))));
        wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
        wn1 = WFE_Expand_Expr (TREE_OPERAND (exp, 1));
        wn  = WN_Binary (MTYPE_signed(mtyp) ? OPR_ASHR : OPR_LSHR,
                         mtyp, wn0, wn1);
      }
      break;

    case TRUTH_ANDIF_EXPR:
    case TRUTH_ORIF_EXPR:
      {
        wn0 = WFE_Expand_Expr_With_Sequence_Point (TREE_OPERAND (exp, 0),
						   Boolean_type);
        wn1 = WFE_Expand_Expr_With_Sequence_Point (TREE_OPERAND (exp, 1),
						   Boolean_type);
        wn  = WN_Binary (Operator_From_Tree [code].opr,
                         Boolean_type, wn0, wn1);
        if (Boolean_type != MTYPE_B &&
	    Widen_Mtype(TY_mtype(Get_TY(TREE_TYPE(exp)))) != Boolean_type)
	  wn = WN_Cvt (Boolean_type, Widen_Mtype(TY_mtype(Get_TY(TREE_TYPE(exp)))), wn);
      }
      break;

    case LT_EXPR:
    case LE_EXPR:
    case GT_EXPR:
    case GE_EXPR:
    case EQ_EXPR:
    case NE_EXPR:
      {
	tree k0 = TREE_OPERAND (exp, 0);
	tree k1 = TREE_OPERAND (exp, 1);
	ty_idx = Get_TY (TREE_TYPE(exp));
        TYPE_ID mtyp = TY_mtype(ty_idx);
	TY_IDX ty_idx0 = Get_TY(TREE_TYPE(k0));
	TYPE_ID mtyp0 = TY_mtype(ty_idx0);
	TY_IDX ty_idx1 = Get_TY(TREE_TYPE(k1));
	TYPE_ID mtyp1 = TY_mtype(ty_idx1);
	tree tmp;
	TY_IDX tty;

	//if (code == EQ_EXPR || code == NE_EXPR) {
	  //when comparing 2 shared ptrs with the same elem type but
	  //different block sizes the GNU front-end casts  one of them
	  // to the other. Strip any NOPs since we get in trouble is
	  // we compare shared with pshared and it happens that their
	  // sizes (in the backend)  don't match.
	  
	  if((TREE_CODE(k0) == NOP_EXPR && TREE_CODE(TREE_OPERAND(k0,0)) == INTEGER_CST) ||
	     TREE_CODE(k0) == INTEGER_CST) {
	    /* could be null in k0 */
	    if(TREE_CODE(k1) == NOP_EXPR && TREE_CODE(TREE_TYPE(k1)) == POINTER_TYPE && 
	       TREE_CODE(TREE_TYPE(TREE_TYPE(k1))) == VOID_TYPE) {
	      tmp = TREE_OPERAND(k1,0);
	      tty = Get_TY(TREE_TYPE(tmp));
	      if(Type_Is_Shared_Ptr(tty) && TY_kind(tty) != KIND_SCALAR) {
		k1 = tmp;
		ty_idx1 = tty; 
	      }
	    }
	  }

	  
	  if((TREE_CODE(k1) == NOP_EXPR && TREE_CODE(TREE_OPERAND(k1,0)) == INTEGER_CST) ||
	     TREE_CODE(k1) == INTEGER_CST) {
	    /* could be null in k0 */
	    if(TREE_CODE(k0) == NOP_EXPR && TREE_CODE(TREE_TYPE(k0)) == POINTER_TYPE && 
	       TREE_CODE(TREE_TYPE(TREE_TYPE(k0))) == VOID_TYPE) {
	      tmp = TREE_OPERAND(k0,0);
	      tty = Get_TY(TREE_TYPE(tmp));
	      if(Type_Is_Shared_Ptr(tty) && TY_kind(tty) != KIND_SCALAR) {
		k0 = tmp;
		ty_idx1 = tty; 
	      }
	    }
	  }


	  
	  /*
	    when comparing shared ptr to shared ptr 
	    the gnu front-end converts both to void* 
	    bug 594,740
	  */
	  /* bug 2541 - it seems that the assumption for bugs 594, 740 was too conservative
	     just strip any casts when the GCC tree types match for both children  */
	  if(TREE_CODE(k0) == NOP_EXPR && TREE_CODE(k1) == NOP_EXPR &&
	     TREE_TYPE(k0) == TREE_TYPE(k1) /* && TREE_CODE(TREE_TYPE(k0)) == POINTER_TYPE &&
					       TREE_CODE(TREE_TYPE(TREE_TYPE(k0))) == VOID_TYPE*/) {
	    ty_idx0 = Get_TY(TREE_TYPE(TREE_OPERAND(k0,0)));
	    ty_idx1 = Get_TY(TREE_TYPE(TREE_OPERAND(k1,0)));
	    if(Type_Is_Shared_Ptr(ty_idx0) && Type_Is_Shared_Ptr(ty_idx1)) {
	      k0 = TREE_OPERAND(k0, 0);
	      k1 = TREE_OPERAND(k1, 0);
	    }
	  }

	  /* When comparing a shared pointer to NULL,
	     the front end casts the pointer into a (void*).
	     Get rid of the cast here so the backend can handle it correctly
	     See bug319.
	   */
	  
	  for (tmp = k0; TREE_CODE(tmp) == NOP_EXPR; tmp = TREE_OPERAND(tmp,0), ty_idx0 = Get_TY(TREE_TYPE(tmp))) {
	    if (TY_kind(ty_idx0) != KIND_POINTER ||
		TY_kind(TY_pointed(ty_idx0)) != KIND_VOID ||
		/* explicit cast to void of shared ptr */
		(TY_kind(ty_idx0) == KIND_POINTER 
		 && TY_kind(TY_pointed(ty_idx0)) == KIND_VOID &&
		 TREE_CODE(TREE_OPERAND(tmp,0)) != INTEGER_CST)
		) {
	      break;
	    }
	  }
	  if (Type_Is_Shared_Ptr(ty_idx0, TRUE)) {
	    k0 = tmp;
	  }
	  for (tmp = k1; TREE_CODE(tmp) == NOP_EXPR; tmp = TREE_OPERAND(tmp,0),	ty_idx1 = Get_TY(TREE_TYPE(tmp))) {
	    if (TY_kind(ty_idx1) != KIND_POINTER ||
		TY_kind(TY_pointed(ty_idx1)) != KIND_VOID ||
		/* explicit cast to void of shared ptr */
		(TY_kind(ty_idx1) == KIND_POINTER 
		 && TY_kind(TY_pointed(ty_idx1)) == KIND_VOID &&
		 TREE_CODE(TREE_OPERAND(tmp,0)) != INTEGER_CST)
		) {
	      break;
	    }
	  }
	  if (Type_Is_Shared_Ptr(ty_idx1, TRUE)) {
	    k1 = tmp;
	  }

	  //} 
	  
	wn0 = WFE_Expand_Expr (k0);
	wn1 = WFE_Expand_Expr (k1);

	
	if(WN_operator(wn0) == OPR_COMMA && WN_operator(WN_kid1(wn0)) == OPR_LDID && 
	   WN_class(WN_kid1(wn0)) == CLASS_PREG) {
	  WN *twn;
	  twn = WN_kid1(wn0);
	  ST *tst = Gen_Temp_Symbol(WN_ty(twn), "_spilleq");
	  wn0 = WN_Stid(TY_mtype(WN_ty(twn)),0, tst, WN_ty(twn), wn0);
	  WFE_Stmt_Append (wn0, Get_Srcpos ());
	  wn0 = WN_Ldid(TY_mtype(WN_ty(twn)), 0, tst, WN_ty(twn));
	}

	if(WN_operator(wn1) == OPR_COMMA && WN_operator(WN_kid1(wn1)) == OPR_LDID &&
	   WN_class(WN_kid1(wn1)) == CLASS_PREG) {
	  WN *twn;
	  twn = WN_kid1(wn1);
	  ST *tst = Gen_Temp_Symbol(WN_ty(twn), "_spilleq");
	  wn1 = WN_Stid(TY_mtype(WN_ty(twn)),0, tst, WN_ty(twn), wn1);
	  WFE_Stmt_Append (wn1, Get_Srcpos ());
	  wn1 = WN_Ldid(TY_mtype(WN_ty(twn)), 0, tst, WN_ty(twn));
	}
	// check if conversion is needed
	TYPE_ID exp_mtyp = Widen_Mtype(mtyp0);

	if (MTYPE_size_min(mtyp1) > MTYPE_size_min(mtyp0) &&
	    ! Has_Subsumed_Cvtl(WN_operator(wn0))) {
	  wn0 = WN_CreateCvtl(OPR_CVTL, Widen_Mtype(mtyp0), MTYPE_V,
			      MTYPE_size_min(mtyp0), wn0);
	  exp_mtyp = Widen_Mtype(mtyp1);
	}
	if (MTYPE_size_min(mtyp0) > MTYPE_size_min(mtyp1) &&
	    ! Has_Subsumed_Cvtl(WN_operator(wn1)))
	  wn1 = WN_CreateCvtl(OPR_CVTL, Widen_Mtype(mtyp1), MTYPE_V,
			      MTYPE_size_min(mtyp1), wn1);

	wn = WN_CreateExp2(Operator_From_Tree [code].opr, MTYPE_B,
			   exp_mtyp, wn0, wn1);

      }
      break;

    case COND_EXPR:
      {
	ty_idx = Get_TY (TREE_TYPE(exp));
	wn0 = WFE_Expand_Expr_With_Sequence_Point (TREE_OPERAND (exp, 0),
						   Boolean_type);
	if (TY_mtype (ty_idx) == MTYPE_V) {
	  WN *then_block = WN_CreateBlock ();
	  WN *else_block = WN_CreateBlock ();
	  WN *if_stmt    = WN_CreateIf (wn0, then_block, else_block);
	  WFE_Stmt_Append (if_stmt, Get_Srcpos());
	  WFE_Stmt_Push (then_block, wfe_stmk_if_then, Get_Srcpos());
	  wn1 = WFE_Expand_Expr (TREE_OPERAND (exp, 1), FALSE);
	  if (wn1) {
	    wn1 = WN_CreateEval (wn1);
	    WFE_Stmt_Append (wn1, Get_Srcpos());
	  }
	  WFE_Stmt_Pop (wfe_stmk_if_then);
	  WFE_Stmt_Push (else_block, wfe_stmk_if_else, Get_Srcpos());
	  wn2 = WFE_Expand_Expr (TREE_OPERAND (exp, 2), FALSE);
	  if (wn2) {
	    wn2 = WN_CreateEval (wn2);
	    WFE_Stmt_Append (wn2, Get_Srcpos());
	  }
	  WFE_Stmt_Pop (wfe_stmk_if_else);
        }
	else {
	  wn1 = WFE_Expand_Expr_With_Sequence_Point (TREE_OPERAND (exp, 1),
						     TY_mtype (ty_idx));
	  wn2 = WFE_Expand_Expr_With_Sequence_Point (TREE_OPERAND (exp, 2),
						     TY_mtype (ty_idx));
	  wn  = WN_CreateExp3 (OPR_CSELECT, Mtype_comparison (TY_mtype (ty_idx)),
			   MTYPE_V, wn0, wn1, wn2);
	  if (WN_operator(wn1) == OPR_LDA && WN_operator(wn2) == OPR_LDA) {
	    if (ST_sym_class(WN_st(wn1)) == CLASS_CONST &&
		TCON_ty(STC_val(WN_st(wn1))) == MTYPE_STRING &&
		ST_sym_class(WN_st(wn2)) == CLASS_CONST &&
		TCON_ty(STC_val(WN_st(wn2))) == MTYPE_STRING) {
	      //both arg1 and arg2 are string constants, and the result type should be const char *
	      //see bug959
	      if (TY_kind(ty_idx) == KIND_POINTER) {
		TY_IDX pointed = TY_pointed(ty_idx);
		Set_TY_is_const(pointed);
		ty_idx = Make_Pointer_Type(pointed);
	      }
	    }
	  }

	  if (TY_kind(ty_idx) != KIND_SCALAR) {
	    //if ?: involves shared scalar operands, its resulting type should be non-shared
	    WN_set_ty(wn, ty_idx);
	  }

	  Set_PU_has_very_high_whirl (Get_Current_PU ());
        }
      }
      break;

    case MODIFY_EXPR:
    case PREDECREMENT_EXPR:
    case PREINCREMENT_EXPR:
    case POSTDECREMENT_EXPR:
    case POSTINCREMENT_EXPR:
      {
        wn1 = WFE_Expand_Expr (TREE_OPERAND (exp, 1)); // r.h.s.
	wn1 = WFE_Patch_Field_Addr_Exp(TREE_OPERAND(exp, 1), wn1);
	wn  = WFE_Lhs_Of_Modify_Expr(code, TREE_OPERAND (exp, 0), need_result, 
				     0, 0, 0, FALSE, wn1, 0, FALSE, FALSE);
      }
      break;

    // ternary ops

    case BIT_FIELD_REF:
      {
        wn = WFE_Expand_Expr (TREE_OPERAND (exp, 0), TRUE, nop_ty_idx, 
			      component_ty_idx, component_offset,
			      field_id, FALSE);
	ty_idx = Get_TY (TREE_TYPE(exp));
	TYPE_ID rtype = TY_mtype(ty_idx);
	UINT siz = TY_size(ty_idx);
	TYPE_ID desc;
	if (siz <= 8) {
	  if (MTYPE_signed(rtype))
	    desc = Mtype_AlignmentClass(siz, MTYPE_CLASS_INTEGER);
	  else desc = Mtype_AlignmentClass(siz, MTYPE_CLASS_UNSIGNED_INTEGER);
	  rtype = Widen_Mtype(desc);
	}
	else desc = rtype;
	WN_set_rtype(wn, rtype);
	WN_set_desc(wn, desc);
	INT bofst = Get_Integer_Value(TREE_OPERAND(exp, 2));
	INT bsiz =Get_Integer_Value(TREE_OPERAND(exp, 1));
	if ((bsiz & 7) == 0 &&	// field size multiple of bytes
	    MTYPE_size_min(desc) % bsiz == 0 && // accessed loc multiple of bsiz
	    bofst % bsiz == 0) {		// bofst multiple of bsiz
	  // not really a bit-field extraction!
	  if (MTYPE_signed(rtype))
	    WN_set_desc(wn, Mtype_AlignmentClass(bsiz >> 3, MTYPE_CLASS_INTEGER));
	  else WN_set_desc(wn, Mtype_AlignmentClass(bsiz >> 3, MTYPE_CLASS_UNSIGNED_INTEGER));
	  WN_load_offset(wn) = WN_load_offset(wn) + (bofst >> 3);

	  if (WN_operator(wn) == OPR_ILOAD && WN_load_offset(wn)) {
	    /* WEI: sometimes a component_ref can be lowered into a bit_field_ref in the front end.
	     * In this case we lose information about its type and instead treat it as a (char *) */
	    WN_set_load_addr_ty(wn, Make_Pointer_Type(MTYPE_To_TY(MTYPE_I1)));
	  }
	} else {
	  if (WN_operator(wn) == OPR_LDID)
	    WN_set_operator(wn, OPR_LDBITS);
	  else WN_set_operator(wn, OPR_ILDBITS);
	  WN_set_bit_offset_size(wn, bofst, bsiz);
	}
	if (MTYPE_byte_size (WN_desc(wn)) != TY_size(WN_ty(wn)))
	  // the container is smaller than the entire struct
	  WN_set_ty (wn, MTYPE_To_TY (WN_desc(wn)));
      }
      break;

    // n-ary ops

    case ARRAY_REF:
      {
	UINT xtra_BE_ofst = 0; 	// only needed for big-endian target
	TY_IDX elem_ty_idx;
	// generate the WHIRL array node
        wn0 = WFE_Array_Expr(exp, &elem_ty_idx, 0, 0, 0);

	// generate the iload node
	TY_IDX hi_ty_idx = Get_TY (TREE_TYPE(exp));
	desc_ty_idx = component_ty_idx;
	if (desc_ty_idx == 0)
          desc_ty_idx = hi_ty_idx;

        if (! MTYPE_is_integral(TY_mtype(desc_ty_idx)))
	  ty_idx = desc_ty_idx;
	else {
	  /* We should cast the expression to the nop_ty_idx with a cvt,
	     instead of using it directly as the type for the iload
	     (see bug374)
	   */
	  ty_idx = desc_ty_idx;
	}

	if (! is_bit_field) {
	  if (TY_size(desc_ty_idx) > TY_size(ty_idx)) {
	    if (Target_Byte_Sex == BIG_ENDIAN)
	      xtra_BE_ofst = TY_size(desc_ty_idx) - TY_size(ty_idx);
	    desc_ty_idx = ty_idx;
	  }
	}
        else {
          if (TY_size(desc_ty_idx) > TY_size(ty_idx))
            ty_idx = desc_ty_idx;
        }

        TYPE_ID rtype = Widen_Mtype(TY_mtype(ty_idx));
        TYPE_ID desc = TY_mtype(desc_ty_idx);
        if (MTYPE_is_integral(desc)) {
          if (MTYPE_signed(rtype) != MTYPE_signed(desc)) {
            if (MTYPE_size_min(rtype) > MTYPE_size_min(desc) ||
		is_bit_field)
              rtype = Mtype_TransferSign(desc, rtype);
            else desc = Mtype_TransferSign(rtype, desc);
          }
        }

	Is_True(! is_bit_field || field_id <= MAX_FIELD_ID,
		("WFE_Expand_Expr: field id for bit-field exceeds limit"));
	// if the elem_type is shared , the ty_idx we get from the
	// tree lacks the shared qualifier.
	if(TY_is_shared(elem_ty_idx) && !TY_is_shared(ty_idx))
	  ty_idx = Make_Shared_Type(ty_idx, TY_block_size(elem_ty_idx));

	wn = WN_CreateIload(OPR_ILOAD, rtype,
			    is_bit_field ? MTYPE_BS : desc, 
			    component_offset+xtra_BE_ofst,
			    //field_id != 0 ? hi_ty_idx : ty_idx,
			    ty_idx,
			    //Make_Pointer_Type (field_id != 0 ? ty_idx : elem_ty_idx, FALSE),
			    Make_Pointer_Type (hi_ty_idx, FALSE),
			    wn0, field_id);
	//fdump_tree(stderr, wn);

	if (nop_ty_idx != 0 && TY_mtype(nop_ty_idx) != TY_mtype(ty_idx)) {
	  wn = WN_Cvt(TY_mtype(nop_ty_idx), TY_mtype(ty_idx), wn);
	}
      }
      break;

    case CALL_EXPR:
      {
	tree arglist = TREE_OPERAND (exp, 1);
        TYPE_ID ret_mtype;
        WN *call_wn;
        WN *arg_wn;
	TY_IDX  arg_ty_idx;
        TYPE_ID arg_mtype;
        INT num_args = 0;
        INT i;
	tree list;
	char* name;
	INTRINSIC iopc = INTRINSIC_NONE;

	tree fun_decl;
	arg0 = TREE_OPERAND (exp, 0);

	enum tree_code code0 = TREE_CODE (arg0);
	if (LANG_Ansi_Setjmp_On == FALSE) {
	  if (current_function_calls_setjmp)
	    Set_PU_calls_setjmp (Get_Current_PU ());
	  if (current_function_calls_longjmp)
	    Set_PU_calls_longjmp (Get_Current_PU ());
	}
	for (list = TREE_OPERAND (exp, 1); list; list = TREE_CHAIN (list)) {
	  if (TREE_CODE(TREE_VALUE(list)) == ERROR_MARK)
	    exit (RC_USER_ERROR);
          arg_ty_idx = MTYPE_To_TY(MTYPE_I4); 
	    //Get_TY(TREE_TYPE(TREE_VALUE(list)));
          if (!WFE_Keep_Zero_Length_Structs    &&
              TY_mtype (arg_ty_idx) == MTYPE_M &&
              TY_size (arg_ty_idx) == 0) {
            // zero length struct parameter
          }
          else
            num_args++;
        }

	//set return type
        ty_idx = Get_TY(TREE_TYPE(exp));
        if (need_result) {
          if (!WFE_Keep_Zero_Length_Structs  &&
              TY_mtype (ty_idx) == MTYPE_M   &&
              TY_size (ty_idx) == 0) {
            // zero length struct return
            ret_mtype = MTYPE_V;
          } else {
            ret_mtype = TY_mtype (ty_idx);
	  }
        } else {
          ret_mtype = MTYPE_V;
	}

        st = NULL;
        if (code0 == ADDR_EXPR                  &&
            TREE_CODE (TREE_OPERAND (arg0, 0))) {
	  tree func = TREE_OPERAND (arg0, 0);
	  BOOL intrinsic_op = FALSE;
          BOOL whirl_generated = FALSE;
	 
          
	  if (DECL_BUILT_IN (func)) {

            switch (DECL_FUNCTION_CODE (func)) {

	      case END_BUILTINS:
		break;

	      case BUILT_IN_STDARG_START:
	      {
		arg1 = TREE_VALUE (arglist);
		arg2 = TREE_VALUE (TREE_CHAIN (arglist));
		WN *arg_wn = WFE_Expand_Expr (arg1);
                TY_IDX arg_ty_idx = Get_TY (TREE_TYPE (arg1));
		while (TREE_CODE (arg2) == NOP_EXPR
		       || TREE_CODE (arg2) == CONVERT_EXPR
		       || TREE_CODE (arg2) == NON_LVALUE_EXPR
		       || TREE_CODE (arg2) == INDIRECT_REF)
		  arg2 = TREE_OPERAND (arg2, 0);
		ST *st2 = Get_ST (arg2);
		wn = WN_Lda (Pointer_Mtype, 
                             ((TY_size (ST_type (st2)) + 7) & (-8)),
                             st2);
		if (WN_operator (arg_wn) == OPR_LDA) {
			wn = WN_Stid (Pointer_Mtype, WN_offset (arg_wn),
				      WN_st (arg_wn), arg_ty_idx, wn);
		}
		else {
			wn = WN_CreateIstore (OPR_ISTORE, MTYPE_V,
					      Pointer_Mtype, 0, arg_ty_idx,
					      arg_wn, 0);
		}

		WFE_Stmt_Append (wn, Get_Srcpos());
		whirl_generated = TRUE;
		wn = NULL;
		break;
	      }

	      case BUILT_IN_VARARGS_START:
	      {
		arg1 = TREE_VALUE (arglist);
		WN *arg_wn = WFE_Expand_Expr (arg1);
		wn = WN_Lda (Pointer_Mtype, 0, WFE_Vararg_Start_ST);
		wn = WN_Stid (Pointer_Mtype, WN_offset (arg_wn),
			      WN_st (arg_wn), arg_ty_idx, wn);

		WFE_Stmt_Append (wn, Get_Srcpos());
		whirl_generated = TRUE;
		wn = NULL;
		break;
	      }

	      case BUILT_IN_VA_COPY:
	      {
		arg1 = TREE_VALUE (arglist);
		arg2 = TREE_VALUE (TREE_CHAIN (arglist));
                TY_IDX arg_ty_idx = Get_TY (TREE_TYPE (arg1));
		WN *src  = WN_CreateParm (Pointer_Mtype, WFE_Expand_Expr (arg1),
					  arg_ty_idx, WN_PARM_BY_VALUE);
		WN *dst  = WN_CreateParm (Pointer_Mtype, WFE_Expand_Expr (arg2),
					  arg_ty_idx, WN_PARM_BY_VALUE);
		WN *size = WN_CreateParm (MTYPE_I4,
					  WN_Intconst(MTYPE_I4,TY_size(arg_ty_idx)),
					  Be_Type_Tbl(MTYPE_I4), WN_PARM_BY_VALUE);
		wn = WN_Create (OPR_INTRINSIC_CALL, ret_mtype, MTYPE_V, 3);
		WN_intrinsic (wn) = INTRN_MEMCPY;
		WN_kid0 (wn) = dst;
		WN_kid1 (wn) = src;
		WN_kid2 (wn) = size;
		WFE_Stmt_Append (wn, Get_Srcpos());
		whirl_generated = TRUE;
		wn = NULL;
	      }

	      case BUILT_IN_VA_END:
	      {
		whirl_generated = TRUE;
		break;
	      }

	      case BUILT_IN_NEXT_ARG:
	      {
                tree last_parm = tree_last (DECL_ARGUMENTS (current_function_decl));
		while (TREE_CODE (last_parm) == NOP_EXPR
		       || TREE_CODE (last_parm) == CONVERT_EXPR
		       || TREE_CODE (last_parm) == NON_LVALUE_EXPR
		       || TREE_CODE (last_parm) == INDIRECT_REF)
		  last_parm = TREE_OPERAND (last_parm, 0);
		ST *st = Get_ST (last_parm);
		arg_wn = WN_Lda (Pointer_Mtype, ST_ofst(st), st);
		wn = WN_Binary (OPR_ADD, Pointer_Mtype, arg_wn,
				WN_Intconst (Pointer_Mtype,
					     Parameter_Size(ST_size(st))));
                whirl_generated = TRUE;
		break;
	      }

              case BUILT_IN_ALLOCA:
		Set_PU_has_alloca (Get_Current_PU ());
		Set_PU_has_user_alloca (Get_Current_PU ());
                arg_wn = WFE_Expand_Expr (TREE_VALUE (TREE_OPERAND (exp, 1)));
	        wn = WN_CreateAlloca (arg_wn);
                whirl_generated = TRUE;
                break;

              case BUILT_IN_MEMCPY:
		iopc = INTRN_MEMCPY;
                break;

              case BUILT_IN_MEMCMP:
		iopc = INTRN_MEMCMP;
                break;

              case BUILT_IN_MEMSET:
		iopc = INTRN_MEMSET;
                break;

              case BUILT_IN_STRCPY:
		iopc = INTRN_STRCPY;
                break;

              case BUILT_IN_STRCMP:
		if (arglist == 0
		    /* Arg could be non-pointer if user redeclared this fcn wrong.  */
		    || TREE_CODE (TREE_TYPE (TREE_VALUE (arglist))) != POINTER_TYPE
		    || TREE_CHAIN (arglist) == 0
		    || TREE_CODE (TREE_TYPE (TREE_VALUE (TREE_CHAIN (arglist)))) != POINTER_TYPE)
		  break;
		else {
		  arg1 = TREE_VALUE (arglist);
		  arg2 = TREE_VALUE (TREE_CHAIN (arglist));
		  tree len1 = c_strlen_exported (arg1);
		  if (len1) {
		    tree len2 = c_strlen_exported (arg2);
		    if (len2) {
		      char *ptr1 = get_string_pointer (WFE_Expand_Expr (arg1));
		      char *ptr2 = get_string_pointer (WFE_Expand_Expr (arg2));
		      if (ptr1 && ptr2) {
			wn = WN_Intconst (MTYPE_I4,
					  strcmp (ptr1, ptr2));
			whirl_generated = TRUE;
			break;
		      }
		    }
		  }
		  iopc = INTRN_STRCMP;
//		  intrinsic_op = TRUE;
		}
                break;

              case BUILT_IN_STRLEN:
		if (arglist == 0
		/* Arg could be non-pointer if user redeclared this fcn wrong.  */
		   || TREE_CODE (TREE_TYPE (TREE_VALUE (arglist))) != POINTER_TYPE)
		  break;
		else {
		  tree src = TREE_VALUE (arglist);
		  tree len = c_strlen_exported (src);
		  if (len) {
		    wn = WFE_Expand_Expr (len);
		    whirl_generated = TRUE;
		  }
		  else {
		    iopc = INTRN_STRLEN;
//		    intrinsic_op = TRUE;
		  }
		}
                break;

              case BUILT_IN_FSQRT:
                arg_wn = WFE_Expand_Expr (TREE_VALUE (TREE_OPERAND (exp, 1)));
                wn = WN_CreateExp1 (OPR_SQRT, ret_mtype, MTYPE_V, arg_wn);
                whirl_generated = TRUE;
                break;

              case BUILT_IN_SIN:
		     if (ret_mtype == MTYPE_F4) iopc = INTRN_F4SIN;
                else if (ret_mtype == MTYPE_F8) iopc = INTRN_F8SIN;
                else Fail_FmtAssertion ("unexpected mtype for intrinsic 'sin'");
		intrinsic_op = TRUE;
                break;

              case BUILT_IN_COS:
		     if (ret_mtype == MTYPE_F4) iopc = INTRN_F4COS;
                else if (ret_mtype == MTYPE_F8) iopc = INTRN_F8COS;
                else Fail_FmtAssertion ("unexpected mtype for intrinsic 'cos'");
		intrinsic_op = TRUE;
                break;

              case BUILT_IN_CONSTANT_P:
              {
		DevWarn ("Encountered BUILT_IN_CONSTANT_P: at line %d\n",
                         lineno);
                tree arg = TREE_VALUE (TREE_OPERAND (exp, 1));
                STRIP_NOPS (arg);
                if (really_constant_p (arg)
                    || (TREE_CODE (arg) == ADDR_EXPR
                        && TREE_CODE (TREE_OPERAND (arg, 0)) == STRING_CST))
                  wn = WN_Intconst (MTYPE_I4, 1);

                else
                  wn = WN_Intconst (MTYPE_I4, 0);
//                wn = WFE_Expand_Expr (TREE_VALUE (TREE_OPERAND (exp, 1)));
                whirl_generated = TRUE;
                break;
              }

              case BUILT_IN_LOCK_TEST_AND_SET:
                wn = emit_builtin_lock_test_and_set (exp, num_args-2);
                whirl_generated = TRUE;
                break;

              case BUILT_IN_LOCK_RELEASE:
                emit_builtin_lock_release (exp, num_args-1);
                whirl_generated = TRUE;
                break;

              case BUILT_IN_COMPARE_AND_SWAP:
                wn = emit_builtin_compare_and_swap (exp, num_args-3);
                whirl_generated = TRUE;
                break;

              case BUILT_IN_SYNCHRONIZE:
                emit_builtin_synchronize (exp, num_args);
                whirl_generated = TRUE;
                break;

              case BUILT_IN_RETURN_ADDRESS:
                i = Get_Integer_Value (TREE_VALUE (TREE_OPERAND (exp, 1)));
		if (i > 0) {
			// currently don't handle levels > 0,
			// which requires iterating thru call-stack
			// and finding $ra in fixed place.
			warning("non-zero levels not supported for builtin_return_address");
			wn = WN_Intconst(Pointer_Mtype, 0);
		}
		else {
			st = WFE_Get_Return_Address_ST (i);
			wn = WN_Ldid (Pointer_Mtype, 0, st, ST_type (st));
		}
                whirl_generated = TRUE;
		break;

	      case BUILT_IN_EXPECT:
	      {
		// __builtin_expect has two arguments
		// the first argument is the value to be returned
		// the second value is the expected value for branch prediction

		list = TREE_OPERAND (exp, 1);
		wn   = WFE_Expand_Expr (TREE_VALUE (list));
		whirl_generated = TRUE;
		break;
	      }

	      case BUILT_IN_FFS:
		iopc = INTRN_I4FFS;
		intrinsic_op = TRUE;
                if (ret_mtype == MTYPE_V)
                  ret_mtype = MTYPE_I4;
		break;

	      case BUILT_IN_ROUND_F2LL:
                arg_wn = WFE_Expand_Expr (TREE_VALUE (TREE_OPERAND (exp, 1)));
                wn = WN_CreateExp1 (OPR_RND, MTYPE_I8, MTYPE_F4, arg_wn);
                whirl_generated = TRUE;
                break;

	      case BUILT_IN_ROUND_D2LL:
                arg_wn = WFE_Expand_Expr (TREE_VALUE (TREE_OPERAND (exp, 1)));
                wn = WN_CreateExp1 (OPR_RND, MTYPE_I8, MTYPE_F8, arg_wn);
                whirl_generated = TRUE;
                break;

	      case BUILT_IN_ROUND_ED2LL:
                arg_wn = WFE_Expand_Expr (TREE_VALUE (TREE_OPERAND (exp, 1)));
                wn = WN_CreateExp1 (OPR_RND, MTYPE_I8, MTYPE_F10, arg_wn);
                whirl_generated = TRUE;
                break;

	      case BUILT_IN_CAST_I2F:
		list = TREE_OPERAND (exp, 1);
		wn = WFE_Expand_Expr (TREE_VALUE (list));
		wn = WN_Tas (MTYPE_F4, Be_Type_Tbl(MTYPE_I4), wn);
		whirl_generated = TRUE;
		break;

	      case BUILT_IN_CAST_F2I:
		list = TREE_OPERAND (exp, 1);
		wn = WFE_Expand_Expr (TREE_VALUE (list));
		wn = WN_Tas (MTYPE_I4, Be_Type_Tbl(MTYPE_F4), wn);
		whirl_generated = TRUE;
		break;

	      case BUILT_IN_CAST_LL2D:
		list = TREE_OPERAND (exp, 1);
		wn = WFE_Expand_Expr (TREE_VALUE (list));
		wn = WN_Tas (MTYPE_F8, Be_Type_Tbl(MTYPE_I8), wn);
		whirl_generated = TRUE;
		break;

	      case BUILT_IN_CAST_D2LL:
		list = TREE_OPERAND (exp, 1);
		wn = WFE_Expand_Expr (TREE_VALUE (list));
		wn = WN_Tas (MTYPE_I8, Be_Type_Tbl(MTYPE_F8), wn);
		whirl_generated = TRUE;
		break;

	      case BUILT_IN_GETF_EXP:
		iopc = INTRN_GETF_EXP;
		intrinsic_op = TRUE;
		break;

	      case BUILT_IN_GETF_SIG:
		iopc = INTRN_GETF_SIG;
		intrinsic_op = TRUE;
		break;

	      case BUILT_IN_SETF_EXP:
		iopc = INTRN_SETF_EXP;
		intrinsic_op = TRUE;
		break;

	      case BUILT_IN_SETF_SIG:
		iopc = INTRN_SETF_SIG;
		intrinsic_op = TRUE;
		break;

	      case BUILT_IN_FMERGE_NS:
		iopc = INTRN_FMERGE_NS;
		intrinsic_op = TRUE;
		break;

	      case BUILT_IN_FMERGE_S:
		iopc = INTRN_FMERGE_S;
		intrinsic_op = TRUE;
		break;

	      case BUILT_IN_FMERGE_SE:
		iopc = INTRN_FMERGE_SE;
		intrinsic_op = TRUE;
		break;
	      default:
		DevWarn ("Encountered BUILT_IN: %d at line %d\n",
			 DECL_FUNCTION_CODE (func), lineno);
		break;
            }
	  }

          if (whirl_generated) {
            break;
          }
	

	  fun_decl = TREE_OPERAND(arg0, 0);
	  name = IDENTIFIER_POINTER(DECL_NAME(fun_decl));

	  iopc = find_upc_intrinsic(name);
	  if (strcmp(name, "upc_resetphase") == 0) {
	    /* resetphase needs special treatment, since for phaseless pointers it should be a nop */
	    list = TREE_OPERAND(exp, 1);
	    
	    arg_ty_idx = Get_TY(TREE_TYPE(TREE_OPERAND(TREE_VALUE(list), 0)));
	    bool is_pshared = (TY_kind(TY_pointed(arg_ty_idx)) != KIND_VOID) && Get_Type_Block_Size(arg_ty_idx) <= 1;
	    if (is_pshared) {
	      return WFE_Expand_Expr (TREE_VALUE (list));
	    } else {
	      iopc = INTRN_S_RESET;
	    }
	  } 

	  if (intrinsic_op) {
	    WN *ikids [5];
	    for (i = 0, list = TREE_OPERAND (exp, 1);
		 list;
		 i++, list = TREE_CHAIN (list)) {
              arg_wn     = WFE_Expand_Expr (TREE_VALUE (list));
	      arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
	      arg_mtype  = TY_mtype(arg_ty_idx);
              arg_wn     = WN_CreateParm (Mtype_comparison (arg_mtype), arg_wn,
					  arg_ty_idx, WN_PARM_BY_VALUE);
	      ikids [i]  = arg_wn;
	    }
	    wn = WN_Create_Intrinsic (OPR_INTRINSIC_OP, ret_mtype, MTYPE_V,
				      iopc, num_args, ikids);
	    break;
	  }

	  if (iopc) {
            call_wn = WN_Create (OPR_INTRINSIC_CALL, ret_mtype, MTYPE_V, num_args);
	    WN_intrinsic (call_wn) = iopc;
	  }
	  else {
            call_wn = WN_Create (OPR_CALL, ret_mtype, MTYPE_V, num_args);
            ST *st2 = DECL_ST2 (TREE_OPERAND (arg0, 0));
            if (Opt_Level > 0 && st2) {
              WN_st_idx (call_wn) = ST_st_idx (st2);
            }
            else {
              st = Get_ST (TREE_OPERAND (arg0, 0));
              WN_st_idx (call_wn) = ST_st_idx (st);
            }
	  }
        }

        else {
	  num_args++;
          call_wn = WN_Create (OPR_ICALL, ret_mtype, MTYPE_V, num_args);
	  WN_kid(call_wn, num_args-1) = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
	  WN_set_ty (call_wn, TY_pointed(Get_TY(TREE_TYPE (TREE_OPERAND (exp, 0)))));
	}

        WN_Set_Linenum (call_wn, Get_Srcpos());
	WN_Set_Call_Default_Flags (call_wn);

        if (st) {
          tree func = TREE_OPERAND (arg0, 0);
          if (DECL_INLINE (func)) {
            wfe_invoke_inliner = TRUE;
          }
        }

        i = 0;
	for (list = TREE_OPERAND (exp, 1);
	     list;
	     list = TREE_CHAIN (list)) {
	  tree arg = TREE_VALUE(list);
	  if (TREE_CODE(arg) == NOP_EXPR && 
	      (iopc == INTRN_THREADOF_S || iopc == INTRN_PHASEOF_S ||
	       iopc == INTRN_ADDROF_S || iopc == INTRN_S_RESET)) {
	    // 	    WEI: use the child instead, otherwise we get the cast to (shared void *)
 	    arg = TREE_OPERAND(arg, 0);
 	  }
	  arg_wn     = WFE_Expand_Expr (arg);
	  arg_ty_idx = Get_TY(TREE_TYPE(arg));
	  arg_mtype  = TY_mtype(arg_ty_idx);
          if (!WFE_Keep_Zero_Length_Structs    &&
              TY_mtype (arg_ty_idx) == MTYPE_M &&
              TY_size (arg_ty_idx) == 0) {
            // zero length struct parameter
	    if (arg_wn && WN_has_side_effects (arg_wn)) {
	      arg_wn = WN_CreateEval (arg_wn);
	      WFE_Stmt_Append (arg_wn, Get_Srcpos());
            }
          } 
          else {
	    // gcc allows non-struct actual to correspond to a struct formal;
	    // fix mtype of parm node so as not to confuse back-end
	    if (arg_mtype == MTYPE_M)
	      arg_mtype = WN_rtype(arg_wn);
	    if (st && WN_operator(arg_wn) == OPR_INTCONST && WN_const_val(arg_wn) == 0) {
	      //C99 allows 0 to be treated as a null pointer, so we may need to add a cast here
	      //depending on the formal parameter type
	      //see bug576
	      TYLIST_IDX params = TY_parms(ST_pu_type(st));
	      TY_IDX parm_ty = 0;
	      int j = 0;
	      for (;Tylist_Table[params] != 0; j++, params = TYLIST_next(params)) {
		if (j == i) {
		  parm_ty = TYLIST_item(Tylist_Table[params]);
		  break;
		}
	      }
	      if (parm_ty != 0 && TY_kind(parm_ty) == KIND_POINTER) {
		arg_ty_idx = parm_ty;
	      }
	    }

            arg_wn = WN_CreateParm (Mtype_comparison (arg_mtype), arg_wn,
				    arg_ty_idx, WN_PARM_BY_VALUE);
            WN_kid (call_wn, i++) = arg_wn;
          }
        }
	
	//for UPC library calls determine here what is the correct intrinsic version (based on phase(less))
	if(iopc == INTRN_THREADOF_S || iopc == INTRN_PHASEOF_S || iopc ==INTRN_ADDROF_S || iopc == INTRN_S_RESET) {
	  TY_IDX act_ty = WN_ty(WN_kid0(arg_wn));
	  if (act_ty == 0) {
	    act_ty = WN_ty(arg_wn);
	    if (!Type_Is_Shared_Ptr(act_ty, true)) {
	      //explicitly set the type to shared void * (see bug1222b)
	      act_ty = Make_Pointer_Type(Make_Shared_Type(MTYPE_To_TY(MTYPE_V), 1));
	    }
	  }
	  bool is_pshared = (TY_kind(TY_pointed(act_ty)) != KIND_VOID) && Get_Type_Block_Size(TY_pointed(act_ty)) <= 1;
	  switch(iopc) {
	  case INTRN_THREADOF_S:
	    if(is_pshared) {
	      iopc = INTRN_THREADOF_P;
	    }
	    break;
	  case INTRN_PHASEOF_S:
	    if(is_pshared) {
	      iopc = INTRN_PHASEOF_P; 
	    }
	    break;
	  case INTRN_ADDROF_S:
	    if(is_pshared) {
	      iopc = INTRN_ADDROF_P;
	    }
	    break;
	  }

	  WN_intrinsic(call_wn) = iopc;
	  //WEI: need to set the type of parm node, to avoid a unnecessary cast later
	  WN_set_ty(arg_wn, act_ty);
	}

	//identify the allocator functions to help the alias analysis
	if (st) {
	    if (strcmp(ST_name(st), "malloc") == 0 ||
		strcmp(ST_name(st), "alloca") == 0 ||
		strcmp(ST_name(st), "calloc") == 0) {
		WN_Set_Call_Does_Mem_Alloc(call_wn);
	    }
	    if (strcmp(ST_name(st), "free") == 0) {
		WN_Set_Call_Does_Mem_Free(call_wn);
	    }
	    if (compiling_upc) {
	      if (strcmp(ST_name(st), "upc_alloc") == 0 ||
		  strcmp(ST_name(st), "upc_local_alloc") == 0 ||
		  strcmp(ST_name(st), "upc_global_alloc") == 0 ||
		  strcmp(ST_name(st), "upc_all_alloc") == 0 ||
		  strcmp(ST_name(st), "upc_global_lock_alloc") == 0 ||
		  strcmp(ST_name(st), "upc_all_lock_alloc") == 0) {
		WN_Set_Call_Does_Mem_Alloc(call_wn);
	      }
	      if (strcmp(ST_name(st), "upc_free") == 0 ||
		  strcmp(ST_name(st), "upc_lock_free") == 0) {
		WN_Set_Call_Does_Mem_Free(call_wn);
	      }
	    }
	}

        if (ret_mtype == MTYPE_V) {
	  WFE_Stmt_Append (call_wn, Get_Srcpos());
        }
	else {
	    wn0 = WN_CreateBlock ();
	    WN_INSERT_BlockLast (wn0, call_wn);
	    
	    wn1 = WN_Ldid (ret_mtype, -1, Return_Val_Preg, ty_idx);
	    //wei: In addition to structs and shared pointers, 
	    //we need to spill if the return value is a floating point, or does not fit
	    //in the register, so that we could take the address of the return value as the argument
	    //to upcr_put_{p}shared
	    bool spill = Compile_Upc && (ret_mtype >= MTYPE_F4 && ret_mtype <= MTYPE_F16) ||
		(is_ia32 && MTYPE_byte_size(ret_mtype) > 4); 
	  
	  if (ret_mtype == MTYPE_M) { // copy the -1 preg to a temp area
	    TY_IDX ret_ty_idx = ty_idx;
	    if (Aggregate_Alignment > 0 &&
		Aggregate_Alignment > TY_align (ret_ty_idx))
	      Set_TY_align (ret_ty_idx, Aggregate_Alignment);
            if (TY_align (ret_ty_idx) < MTYPE_align_best(Spill_Int_Mtype))
              Set_TY_align (ret_ty_idx, MTYPE_align_best(Spill_Int_Mtype));
	    ST *ret_st = Gen_Temp_Symbol(ret_ty_idx, 
		  st ? Index_To_Str(Save_Str2((char*) ".Mreturn.",
					      ST_name(ST_st_idx(st))))
		     : (char*) ".Mreturn.");
	    
	    wn1  = WN_CreateComma (OPR_COMMA, WN_rtype (wn1), MTYPE_V,
                                wn0, wn1);
	    wn1 = WN_Stid (ret_mtype, 0, ret_st, ty_idx, wn1);
	    wn0 =  WN_CreateBlock ();
            WN_INSERT_BlockLast (wn0, wn1);

	    // ritual for determining the right mtypes to be used in the LDID
            UINT xtra_BE_ofst = 0;  // only needed for big-endian target
            desc_ty_idx = component_ty_idx;
            if (desc_ty_idx == 0)
              desc_ty_idx = Get_TY (TREE_TYPE(exp));
              
            if (! MTYPE_is_integral(TY_mtype(desc_ty_idx)))
              ty_idx = desc_ty_idx;
            else { 
              ty_idx = nop_ty_idx;
              if (ty_idx == 0 || Type_Is_Shared_Ptr(desc_ty_idx))
                ty_idx = desc_ty_idx;
            }

	    if (! is_bit_field) {
              if (TY_size(desc_ty_idx) > TY_size(ty_idx)) {
                if (Target_Byte_Sex == BIG_ENDIAN)
                  xtra_BE_ofst = TY_size(desc_ty_idx) - TY_size(ty_idx);
                desc_ty_idx = ty_idx;
	      }
            }
	    else {
	      if (TY_size(desc_ty_idx) > TY_size(ty_idx))
		ty_idx = desc_ty_idx;
	    }

	    TYPE_ID rtype = Widen_Mtype(TY_mtype(ty_idx));
	    TYPE_ID desc = TY_mtype(desc_ty_idx);
	    if (MTYPE_is_integral(desc)) {
	      if (MTYPE_signed(rtype) != MTYPE_signed(desc)) {
		if (MTYPE_size_min(rtype) > MTYPE_size_min(desc) ||
		    is_bit_field)
		  rtype = Mtype_TransferSign(desc, rtype);
		else desc = Mtype_TransferSign(rtype, desc);
	      }
	    }

            Is_True(! is_bit_field || field_id <= MAX_FIELD_ID,
                    ("WFE_Expand_Expr: field id for bit-field exceeds limit"));
  
	    wn1 = WN_CreateLdid(OPR_LDID, TY_mtype(ret_ty_idx)/*rtype*/,
			        is_bit_field ? MTYPE_BS : desc,
			        ST_ofst(ret_st)+component_offset+xtra_BE_ofst, 
				ret_st,
				(field_id != 0 && component_ty_idx != 0) ?
				Get_TY (TREE_TYPE(exp)) : ret_ty_idx/* ty_idx */,
				field_id);

	    if (MTYPE_is_integral(TY_mtype(ret_ty_idx)) &&
		MTYPE_size_min(TY_mtype(ret_ty_idx)) > MTYPE_size_min(rtype)) {
	      /* We need an integer conversion */
	      wn1 = WN_Cvt(TY_mtype(ret_ty_idx), rtype, wn1);
	    }
	  }
          wn  = WN_CreateComma (OPR_COMMA, WN_rtype (wn1), MTYPE_V,
                                wn0, wn1);
        }
      }
      break;

    case COMPOUND_EXPR:
      {
        wn = WFE_Expand_Expr (TREE_OPERAND (exp, 0), FALSE);
	if (wn) {
	    wn = WN_CreateEval (wn);
	    WFE_Stmt_Append (wn, Get_Srcpos ());
	}
        wn = WFE_Expand_Expr (TREE_OPERAND (exp, 1));
      }
      break;

    case NON_LVALUE_EXPR:
      {
        wn = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
      }
      break;

    case SAVE_EXPR:
      {
	//DevWarn ("Encountered SAVE_EXPR at line %d", lineno);
        wn = WFE_Save_Expr (exp);
      }
      break;

    case UNSAVE_EXPR:
      {
	//DevWarn ("Encountered UNSAVE_EXPR at line %d", lineno);
        wn = WFE_Save_Expr (exp);
	WFE_Unsave_Expr (exp);
      }
      break;

    case TARGET_EXPR:
      {
	DevWarn ("Encountered TARGET_EXPR at line %d", lineno);
	wn0 = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
	wn1 = WFE_Expand_Expr (TREE_OPERAND (exp, 1));
      }
      break;

#ifdef GPLUSPLUS_FE
    case INIT_EXPR:
      {
	tree initiand    = TREE_OPERAND(exp, 0);
	tree initializer = TREE_OPERAND(exp, 1);
	tree init_expr;

	if (TREE_CODE(initiand) != VAR_DECL) {
	  DevWarn("expected VAR_DECL");
	  break;
	}

	if (TREE_CODE(initializer) != TARGET_EXPR) {
	  DevWarn("expected TARGET_EXPR");
	  break;
	}
	
	init_expr = TREE_OPERAND(initializer, 3);
	if (TREE_CODE(init_expr) != AGGR_INIT_EXPR) {
	  DevWarn("expected AGGR_INIT_EXPR");
	  break;
	}

	{
	  tree func = TREE_OPERAND(init_expr, 0);
	  tree args = TREE_OPERAND(init_expr, 1);
	  INT  num_args = 0;
	  INT  i;
	  tree list;
	  WN * call_wn;
	  ST * initiand_st = Get_ST(initiand);
	  WN * initiand_wn = WN_Lda(Pointer_Mtype,
				    ST_ofst(initiand_st), 
				    initiand_st);
	  WN * arg_wn;
	  
	  for (list = args; list; list = TREE_CHAIN(list))
	    ++num_args;
	  call_wn = WN_Create(OPR_CALL, MTYPE_V, MTYPE_V, num_args);
	  st = Get_ST(TREE_OPERAND(func, 0));
	  WN_st_idx(call_wn) = ST_st_idx(st);
	  WN_Set_Call_Default_Flags(call_wn);
	
	  
	  for (i = 0, list = TREE_OPERAND(init_expr, 1);
	       list;
	       ++i, list = TREE_CHAIN(list)) {
	    arg_wn = (i == 0) ? initiand_wn
			     : WFE_Expand_Expr(TREE_VALUE(list));
	    TY_IDX  arg_ty_idx = Get_TY(TREE_TYPE(TREE_VALUE(list)));
	    TYPE_ID arg_mtype  = TY_mtype(arg_ty_idx);
	    if (arg_mtype == MTYPE_M)
  	      arg_mtype = WN_rtype(arg_wn);
	    arg_wn = WN_CreateParm(Mtype_comparison(arg_mtype),
				   arg_wn,
				   arg_ty_idx,
				   WN_PARM_BY_VALUE);
	    WN_kid(call_wn, i) = arg_wn;
         	       WN_kid (call_wn, i) = arg_wn;
	  }

	  WFE_Stmt_Append(call_wn, Get_Srcpos());
	}       
        break;
      }

    case UNSAVE_EXPR:
      {
	wn = WFE_Expand_Expr( TREE_OPERAND (exp, 0));
	break;
      }     
    case NEW_EXPR:
      {
	wn = WFE_Expand_Expr(build_new_1(exp));
        break;
      }
    case RTL_EXPR:
      break;
#endif /* GPLUSPLUS_FE */

    case VA_ARG_EXPR:
      {
        // code swiped from builtins.c (std_expand_builtin_va_arg)
	INT64 align;
	INT64 rounded_size;
	tree type = TREE_TYPE (exp);
	TY_IDX ty_idx = Get_TY (type);
	TYPE_ID mtype = TY_mtype (ty_idx);

        /* Compute the rounded size of the type.  */
	align = PARM_BOUNDARY / BITS_PER_UNIT;
	rounded_size = (((int_size_in_bytes (type) + align - 1) / align) * align);

	/* Get AP.  */
	WN        *ap_load   = WFE_Expand_Expr (TREE_OPERAND (exp, 0));
        TY_IDX     ap_ty_idx = Get_TY (TREE_TYPE (TREE_OPERAND (exp, 0)));
        WN        *ap_addr;
        ST        *ap_st;
        WN_OFFSET  ap_offset;

        if (WN_operator(ap_load) == OPR_LDID) {
	  ap_st     = WN_st (ap_load);
          ap_offset = WN_offset (ap_load);
        }
        else
        if (WN_operator(ap_load) == OPR_ILOAD) {
          ap_st     = NULL;
          ap_offset = WN_offset (ap_load);
          ap_addr   = WN_COPY_Tree (WN_kid0 (ap_load));
          if (WN_has_side_effects (ap_addr))
            Fail_FmtAssertion ("VA_ARG_EXPR: ap address has side effects");
        }
        else
          Fail_FmtAssertion ("VA_ARG_EXPR: unknown operator for ap");

	if (Target_Byte_Sex == BIG_ENDIAN) {
	  Fail_FmtAssertion ("VA_ARG_EXPR not implemented for BIG_ENDIAN");
	  INT64 adj;
	  adj = TREE_INT_CST_LOW (TYPE_SIZE (type)) / BITS_PER_UNIT;
	  if (rounded_size > align)
	    adj = rounded_size;

	  wn = WN_Binary (OPR_ADD, Pointer_Mtype, wn,
			  WN_Intconst (Pointer_Mtype, rounded_size - adj));
	}

	/* Compute new value for AP.  */
	wn = WN_Binary (OPR_ADD, Pointer_Mtype, WN_COPY_Tree (ap_load),
			WN_Intconst (Pointer_Mtype, rounded_size));
        if (ap_st)
	  wn = WN_Stid (Pointer_Mtype, ap_offset, ap_st, ap_ty_idx, wn);
        else {
          wn = WN_CreateIstore (OPR_ISTORE, MTYPE_V, Pointer_Mtype, ap_offset,
                                ap_ty_idx, wn, ap_addr, 0);
        }
        WFE_Stmt_Append (wn, Get_Srcpos ());
        wn = WN_CreateIload (OPR_ILOAD, Widen_Mtype (mtype), mtype, -rounded_size,
			     ap_ty_idx, Make_Pointer_Type (ap_ty_idx, FALSE),
			     ap_load);
      }
      break;

    case ERROR_MARK:
      exit (RC_USER_ERROR);
      break;

    default:
      {
        Fail_FmtAssertion ("WFE_Expand_Expr: not implemented %s",
                           Operator_From_Tree [code].name);
      }
      break;
    }

#ifdef WFE_DEBUG
  if (wn)
    fdump_tree (stderr, wn);

  fprintf (stderr, // "{("
           ")} WFE_Expand_Expr: %s\n", Operator_From_Tree [code].name);
#endif /* WFE_DEBUG */

  if (need_result)
    FmtAssert (wn != 0 || code == CALL_EXPR || code == BIND_EXPR ||
               code == COMPOUND_EXPR ||
               code == INDIRECT_REF ||
               code == COMPONENT_REF ||
               ((code == COND_EXPR) && (TY_mtype (ty_idx) == MTYPE_V)),
	       ("WFE_Expand_Expr: NULL WHIRL tree for %s at line %d",
		Operator_From_Tree [code].name, lineno));

  return wn;
}

void WFE_One_Stmt (tree exp)
{
  WN *wn;
  wfe_save_expr_stack_last = -1; // to minimize searches
  if (get_expr_stmts_for_value ())
    wn = WFE_Expand_Expr_With_Sequence_Point (exp, TY_mtype (Get_TY (TREE_TYPE (exp))));
  else
    wn = WFE_Expand_Expr_With_Sequence_Point (exp, MTYPE_V);
  if (wn) {
    if (get_expr_stmts_for_value ()) {
      wn = WN_CreateEval (wn);
      WFE_Stmt_Append (wn, Get_Srcpos ());
      return;
    }
    for (;;) {
      if (WN_operator (wn) == OPR_COMMA) {
	WN *crwn = wn;
	if (WN_operator (WN_kid1 (wn)) == OPR_LDID                 &&
	    WN_st (WN_kid1 (wn)) == Return_Val_Preg                &&
	    (WN_operator (WN_last (WN_kid0 (wn))) == OPR_CALL   ||
	     WN_operator (WN_last (WN_kid0 (wn))) == OPR_ICALL)) {
	  WN_set_rtype (WN_last (WN_kid0 (wn)), MTYPE_V);
	  WFE_Stmt_Append (WN_kid0 (wn), Get_Srcpos ());
	  WN_Delete (crwn);
	  break;
	}
	else {
	  WFE_Stmt_Append (WN_kid0 (wn), Get_Srcpos ());
	  wn = WN_kid1 (wn);
	  WN_Delete (crwn);
	}
      }
      else {
        BOOL need_eval = TRUE;
	if (WN_operator (wn) == OPR_LDID &&
	    !TY_is_volatile (ST_type (WN_st (wn))))
          need_eval = FALSE;
        else
        if (WN_operator (wn) == OPR_ILOAD &&
            WN_has_side_effects (wn) == FALSE)
          need_eval = FALSE;
        if (need_eval) {
	  wn = WN_CreateEval (wn);
	  WFE_Stmt_Append (wn, Get_Srcpos ());
	}
	break;
      }
    }
  }
}

void WFE_Null_Return (void)
{
  WN *wn = WN_CreateReturn ();
  WFE_Stmt_Append (wn, Get_Srcpos());
}

UINT64
Get_Integer_Value (tree exp)
{

  //WEI: this may be the case that the block size is not a compile constant
  //Instead of crashing, we lie about its block size so the compiler 
  //can continue to run, since this error
  //should be reported in upc-share.c (set_upc_blocksize)
  //ditto for a null exp, which should've been caught elsewhere
  if (exp == NULL || TREE_CODE(exp) != INTEGER_CST) {
    return 1;
  }

#ifdef _LP64
	return TREE_INT_CST_LOW (exp);
#else
	UINT64 h = TREE_INT_CST_HIGH (exp);
	UINT64 l = TREE_INT_CST_LOW (exp);
	l = (l << 32) >> 32;	// zero-out high 32 bits
	h = h << 32;
	return (h | l);
#endif /* _LP64 */
}

void
WFE_Expr_Init (void)
{
  INT i;
  for (i = 0; i < LAST_AND_UNUSED_TREE_CODE; i++)
    FmtAssert (Operator_From_Tree [i].tree_code == i,
               ("Operator_From_Tree table incorrect at %d", i));
}

char *
WFE_Tree_Node_Name (tree op)
{
  return Operator_From_Tree [TREE_CODE (op)].name;
}
